/************************************************************************
 *   IRC - Internet Relay Chat, src/dbuf.c
 *   Copyright (C) 1990 Markku Savela
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * For documentation of the *global* functions implemented here,
 * see the header file (dbuf.h).
 *
 *
 * $Id: dbuf.c,v 1.20 2002/10/04 20:36:08 lucas Exp $
 */

#include "dbuf.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"

#include <stdio.h>
#include <stdlib.h>

#define DBUF_SIZE_CLIENT 1024
#define DBUF_SIZE_SERVER 65536

struct DBufBuffer {
   struct DBufBuffer *next;             /* Next data buffer, NULL if last */
   char              *start;            /* data starts here */
   char              *end;              /* data ends here */ 
   struct DBufBlock  *parent;		/* what block do I belog to? */
   char               data[1];          /* Actual data stored here -- do not change this! */
};

struct DBufBlock {
   struct DBufBuffer *buf;
   unsigned int block_seq;		/* the number of this block */
   int blocksize;			/* size of the buffer in each DBufBuffer in this block */
   int num;				/* how many buffers were there in this chunk? */
   int given_back;			/* how many have been 'given back' to me? */
   int is_freeing;			/* is this block being freed? */
   int (*callback) (int, int);		/* callback function to call when freed, calls with num and size */
   struct DBufBlock *next;
};

static int                 DBufUsedCountC = 0, DBufUsedCountS = 0;
static int                 DBufCountC = 0, DBufCountS = 0;
static struct DBufBuffer  *dbufFreeListC = NULL, *dbufFreeListS = NULL;
static struct DBufBlock   *dbuf_blocks = NULL;
static int                 dbuf_buffer_header_siz = 0;
static unsigned int        block_seq = 0;

/* pairs: 1,5 2,6 3,7 4,8 */
/* we want in s_debug: 6,2 5,1 8,4 7,3 */

void count_dbuf_memory(size_t* allocatedS, size_t* usedS, size_t* allocatedC, size_t* usedC, int *allocnS, int *usednS,
                       int *allocnC, int *usednC)
{
  *allocatedS = DBufCountS     * (dbuf_buffer_header_siz + DBUF_SIZE_SERVER);
  *usedS      = DBufUsedCountS * (dbuf_buffer_header_siz + DBUF_SIZE_SERVER);
  *allocatedC = DBufCountC     * (dbuf_buffer_header_siz + DBUF_SIZE_CLIENT);
  *usedC      = DBufUsedCountC * (dbuf_buffer_header_siz + DBUF_SIZE_CLIENT);

  *allocnS = DBufCountS;
  *allocnC = DBufCountC;
  *usednS = DBufUsedCountS;
  *usednC = DBufUsedCountC;
}

/* 
 * dbuf_init
 * initialize a stretch of memory as dbufs.
 */

void dbuf_allocblock(int num, int forserver)
{
   struct DBufBlock *dblock;
   struct DBufBuffer *dfree, *dbp;
   int i;
   int bufsize = forserver ? DBUF_SIZE_SERVER : DBUF_SIZE_CLIENT;

   dblock = (struct DBufBlock *) MyMalloc(sizeof(struct DBufBlock));  
   memset(dblock, 0, sizeof(struct DBufBlock));
   dblock->num = num;
   dblock->block_seq = block_seq++;
   dblock->next = dbuf_blocks;
   dblock->blocksize = bufsize;
   dbuf_blocks = dblock;

   dbp = (struct DBufBuffer *) MyMalloc((dbuf_buffer_header_siz + bufsize) * num);
   dblock->buf = dbp;

   dfree = dbp;

   for(i = 0; i < (num - 1); i++)
   {
     struct DBufBuffer *dbnext;
     dfree->parent = dblock;
     dbnext = (struct DBufBuffer *) (((char *) dfree) + bufsize + dbuf_buffer_header_siz);
     dfree->next = dbnext;
     dfree = dbnext;
   }

   dfree->parent = dblock;

   if(forserver)
   {
      dfree->next = dbufFreeListS;
      dbufFreeListS = dbp;
      DBufCountS += num;
   }
   else
   {
      dfree->next = dbufFreeListC;
      dbufFreeListC = dbp;
      DBufCountC += num;
   }
}

void dbuf_init()
{
   /* allocate one block of the initial size */

   dbuf_buffer_header_siz = sizeof(struct DBufBuffer) - sizeof(char[1]);

   dbuf_allocblock(INITIAL_CLIENT_DBUFS, 0);
   dbuf_allocblock(INITIAL_SERVER_DBUFS, 1);
}

/*
 * dbuf_alloc - allocates a struct DBufBuffer structure either from 
 * dbufFreeListC/S or create a new one.
 */
static struct DBufBuffer* dbuf_alloc(int forserver)
{
   struct DBufBuffer* db = forserver ? dbufFreeListS : dbufFreeListC;

   if (!db)
   {
     dbuf_allocblock(forserver ? INITIAL_SERVER_DBUFS : INITIAL_CLIENT_DBUFS, forserver);
     db = forserver ? dbufFreeListS : dbufFreeListC;
     if(!db) 
       return NULL;
   }

   if(forserver)
   {
      dbufFreeListS = dbufFreeListS->next;
      ++DBufUsedCountS;
   }
   else
   {
      dbufFreeListC = dbufFreeListC->next;
      ++DBufUsedCountC;
   }

   db->next  = NULL;
   db->start = db->end = db->data;

   return db;
}

void dbuf_block_free(struct DBufBlock *dbb)
{
   struct DBufBlock *db, *dbprev = NULL;
   int (*callback) (int, int);
   int mem, memsize;

   db = dbuf_blocks;
   while(db)
   {
      if(db == dbb)
      {
         if(dbprev)
            dbprev->next = db->next;
         else
            dbuf_blocks = db->next;

         mem = db->num;
         memsize = db->num * db->blocksize;
         callback = db->callback;
         free(db->buf);
         free(db);
         if(callback)
            callback(mem, memsize);
         return;
      }

      dbprev = db;
      db = db->next;
   }
}

/*
 * dbuf_free - return a struct DBufBuffer structure to the dbufFreeList
 */
static void dbuf_free(struct DBufBuffer* ptr)
{
   if(ptr->parent->is_freeing)
   {
      int forserver;

      if(ptr->parent->blocksize == DBUF_SIZE_CLIENT)
      {
         DBufUsedCountC--;
         forserver = 0;
      }
      else if(ptr->parent->blocksize == DBUF_SIZE_SERVER)
      {
         DBufUsedCountS--;
         forserver = 1;
      }
      else
         abort();

      ptr->parent->given_back++;
      if(ptr->parent->given_back == ptr->parent->num)
      {
         if(forserver)
            DBufCountS -= ptr->parent->num;
         else
            DBufCountC -= ptr->parent->num;
         dbuf_block_free(ptr->parent);
      }
      return;
   }

   if(ptr->parent->blocksize == DBUF_SIZE_CLIENT)
   {
      DBufUsedCountC--;
      ptr->next = dbufFreeListC;
      dbufFreeListC = ptr;
   }
   else if(ptr->parent->blocksize == DBUF_SIZE_SERVER)
   {
      DBufUsedCountS--;
      ptr->next = dbufFreeListS;
      dbufFreeListS = ptr;
   }
   else
      abort(); /* this should never happen, we have a memleak somewhere? */
}

/*
** This is called when malloc fails. Scrap the whole content
** of dynamic buffer and return -1. (malloc errors are FATAL,
** there is no reason to continue this buffer...). After this
** the "dbuf" has consistent EMPTY status... ;)
*/

static int dbuf_malloc_error(struct DBuf* dyn)
{
   struct DBufBuffer* db;

   dyn->length = 0;
   while ((db = dyn->head) != NULL) 
   {
      dyn->head = db->next;
      dbuf_free(db);
   }
   dyn->tail = NULL;

   return 0;
}

/*
 * dbuf_put - put a sequence of bytes in a dbuf
 */
int dbuf_put(struct DBuf* dyn, const char* buf, size_t length, int forserver)
{
   struct DBufBuffer** h;
   struct DBufBuffer* d;
   int chunk;

   /*
    * Locate the last non-empty buffer. If the last buffer is
    * full, the loop will terminate with 'd==NULL'. This loop
    * assumes that the 'dyn->length' field is correctly
    * maintained, as it should--no other check really needed.
    */
   if (dyn->length == 0)
      h = &(dyn->head);
   else
      h = &(dyn->tail);

   /* Append users data to buffer, allocating buffers as needed */

   dyn->length += length;

   for (; length > 0; h = &(d->next)) 
   {
      if ((d = *h) == NULL) 
      {
         if ((d = dbuf_alloc(forserver)) == NULL)
            return dbuf_malloc_error(dyn);

         dyn->tail = d;
         *h = d;        /* prev->next = d */
      }

      chunk = (d->data + d->parent->blocksize) - d->end;
      if (chunk) 
      {
         if (chunk > length)
            chunk = length;
       
         memcpy(d->end, buf, chunk);

         length -= chunk;
         buf += chunk;
         d->end += chunk;
      }
   }

   return 1;
}


char* dbuf_map(struct DBuf* dyn, size_t* length)
{
   if (dyn->length == 0)
   {
      *length = 0;
      return NULL;
   }

   *length = dyn->head->end - dyn->head->start;
   return dyn->head->start;
}

size_t dbuf_copy_map(struct DBuf* dyn, char *buf, int length)
{
  struct DBufBuffer *db;
  size_t chunk = 0, fulllen = 0;

  for(db = dyn->head; db; db = db->next)
  {
     chunk = db->end - db->start;
     if(!chunk) continue;

     if(fulllen + chunk > length)
        chunk = length - fulllen;

     memcpy(buf + fulllen, db->start, chunk);
     fulllen += chunk;

     if(fulllen >= length)
        break;
  }

  return fulllen;
}

void dbuf_delete(struct DBuf* dyn, size_t length)
{
   struct DBufBuffer* db;
   size_t chunk;

   if (length > dyn->length)
      length = dyn->length;

   while (length > 0) 
   {
      if ((db = dyn->head) == NULL)
         break;
      chunk = db->end - db->start; 
      if (chunk > length)
        chunk = length;

      length -= chunk;
      dyn->length -= chunk;
      db->start += chunk;

      if (db->start == db->end) 
      {
        dyn->head = db->next;
        dbuf_free(db);
      }
   }

   if (dyn->head == NULL) 
   {
      dyn->tail = 0;
      dyn->length = 0;
   }
}

size_t dbuf_get(struct DBuf *dyn, char *buf, size_t length)
{
   size_t moved = 0;
   size_t chunk;
   const char *b;

   while (length > 0 && (b = dbuf_map(dyn, &chunk)) != NULL) 
   {
      if (chunk > length)
         chunk = length;

      memcpy(buf, b, chunk);
      dbuf_delete(dyn, chunk);

      buf += chunk;
      length -= chunk;
      moved  += chunk;
   }
   return moved;
}

static size_t dbuf_flush(struct DBuf *dyn)
{
   struct DBufBuffer* db = dyn->head;
  
   if (db == NULL)
      return 0;

   /*
    * flush extra line terms
    */
   while (IsEol(*db->start)) 
   {
      if (++db->start == db->end) 
      {
         dyn->head = db->next;
         dbuf_free(db);
         if ((db = dyn->head) == NULL) 
         {
            dyn->tail = 0;
            dyn->length = 0;
            break;
         }
      }
      dyn->length--;
   }
   return dyn->length;
}

/*
 * dbuf_getmsg
 *
 * Check the buffers to see if there is a string which is terminated with
 * either a \r or \n present.  If so, copy as much as possible (determined by
 * length) into buf and return the amount copied - else return 0.
 *
 * There may be cr/lf pairs or leading garbage in the dbuf to start with
 *
 * case 1:
 *   data starts at buf + offset, entire message fits in a dbuf
 * case 2:
 *   data starts at buf + offset, more than one dbuf is spanned
 */
int dbuf_getmsg(struct DBuf* dyn, char* buf, size_t length)
{
   struct DBufBuffer* db;
   char *start;
   char *end;
   size_t count;
   size_t copied = 0;

   if (dbuf_flush(dyn) == 0)
     return 0;

   db = dyn->head;
   start = db->start;

   if (length > dyn->length)
      length = dyn->length;
   /*
    * might as well copy it while we're here
    */
   while (length > 0) 
   {
      end = IRCD_MIN(db->end, (start + length));
      while (start < end && !IsEol(*start))
         *buf++ = *start++;

      count = start - db->start;
      if (start < end) 
      {
         *buf = '\0';
         copied += count;
         dbuf_delete(dyn, copied);
         dbuf_flush(dyn);
         return copied;
      } 
      if ((db = db->next) == NULL)
         break;
      copied += count;
      length -= count;
      start = db->start;
   }
   return 0;  
}

int dbuf_startfree_block(int forserver, int doall, int (*callback) (int, int))
{
   struct DBufBlock *recentdb, *dbb;
   int block_count, numblocks = 0;

   /* to avoid too much fragmentation grossness, free the latest allocated buffers first. */

   while(1)
   {
      block_count = 0;
      recentdb = NULL;
      dbb = dbuf_blocks;

      while(dbb)
      {
         if((forserver && dbb->blocksize != DBUF_SIZE_SERVER) ||
            (!forserver && dbb->blocksize != DBUF_SIZE_CLIENT) ||
            dbb->is_freeing)
         {
            dbb = dbb->next;
            continue;
         }

         block_count++;

         if(!recentdb)
           recentdb = dbb;
         else
         {
            if(recentdb->block_seq < dbb->block_seq)
               recentdb = dbb;
         }
         dbb = dbb->next;
      }

      if(block_count <= 1)
        return numblocks;

      if(recentdb)
      {
         recentdb->callback = callback;
         recentdb->is_freeing = 1;
         numblocks++;
      }
      if(!doall)
         return numblocks;
   }

   return 0; /* never get here */
}
