/************************************************************************
 *   IRC - Internet Relay Chat, src/s_serv.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: s_serv.c,v 1.172 2002/10/09 00:07:47 lucas Exp $ */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"
#include "dh.h"
#include "zlink.h"
#include "userban.h"

#if defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
# include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
  /* older unices don't have strchr/strrchr .. help them out */
# include <strings.h>
# undef strchr
# define strchr index
#endif
#include "dich_conf.h"
#include "fdlist.h"
#include "throttle.h"

/* external variables */

extern fdlist serv_fdlist;
extern int lifesux;
static char buf[BUFSIZE];
extern int rehashed;
extern int zline_in_progress;

/* external functions */

#ifdef MAXBUFFERS
extern void reset_sock_opts();
#endif

extern char *smalldate(time_t);	/* defined in s_misc.c */
extern void s_die(void);	/* defined in ircd.c as VOIDSIG */
extern void report_conf_links(aClient *, aConfList *, int, char);
extern void show_opers(aClient *, char *);
extern void show_servers(aClient *, char *);
extern void count_memory(aClient *, char *);
extern void rehash_ip_hash();	/* defined in s_conf.c */
extern char *find_restartpass();
extern char *find_diepass();

/* Local function prototypes */

static int k_isnumber(char *);	/* return 0 if not, else return number */
static char *cluster(char *);

void make_capab_string(char *buf, unsigned int noflags);

int send_motd(aClient *, aClient *, int, char **);
void read_motd(char *);
#ifdef UNKLINE
static int flush_write(aClient *, int, char *, int, char *);
#endif

#ifdef SHORT_MOTD
void read_shortmotd(char *);
#endif

/* global variables */
#ifdef HIGHEST_CONNECTION
int max_connection_count = 1, max_client_count = 1;
#endif
char motd_last_changed_date[MAX_DATE_STRING]; /* enough room for date */	

#ifdef LOCKFILE
/* Shadowfax's lockfile code */
void        do_pending_klines(void);

struct pkl {
   char       *comment;		/* Kline Comment */
   char       *kline;		/* Actual Kline */
   struct pkl *next;		/* Next Pending Kline */
} *pending_klines = NULL;

time_t      pending_kline_time = 0;
#endif /* LOCKFILE */

/*
 * * m_functions execute protocol messages on this server: *
 * 
 * cptr: 
 ** always NON-NULL, pointing to a *LOCAL* client
 ** structure (with an open socket connected!). This 
 ** is the physical socket where the message originated (or
 ** which caused the m_function to be executed--some
 ** m_functions may call others...). 
 * 
 * sptr:
 ** the source of the message, defined by the
 ** prefix part of the message if present. If not or
 ** prefix not found, then sptr==cptr. 
 * 
 *      *Always* true (if 'parse' and others are working correct): 
 * 
 *      1)      sptr->from == cptr  (note: cptr->from == cptr) 
 * 
 *      2)      MyConnect(sptr) <=> sptr == cptr (e.g. sptr 
 * annot* be a local connection, unless it's actually
 * cptr!). 
 *
 * MyConnect(x) should probably  be defined as (x == x->from) --msa
 * 
 * parc:    
 ** number of variable parameter strings (if zero, 
 ** parv is allowed to be NULL)
 * 
 * parv:    
 ** a NULL terminated list of parameter pointers,
 *** parv[0], sender (prefix string), if not present his points to 
 *** an empty string.
 *
 ** [parc-1]:
 *** pointers to additional parameters 
 *** parv[parc] == NULL, *always* 
 * 
 * note:   it is guaranteed that parv[0]..parv[parc-1] are all
 *         non-NULL pointers.
 */

/*
 * m_version 
 * parv[0] = sender prefix 
 * parv[1] = remote server
 */
int m_version(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   extern char serveropts[];

   if (use_or_deliver(cptr, sptr, &CMD_VERSION, "%s", 1, parc, parv) == HUNTED_ISME)
	send_me_numeric(sptr, RPL_VERSION, version, debugmode, me.name, serveropts);
   return 0;
}

/*
 * m_squit
 * there are two types of squits: those going downstream (to the target server)
 * and those going back upstream (from the target server).
 * previously, it wasn't necessary to distinguish between these two types of 
 * squits because they neatly echoed back all of the QUIT messages during an squit.
 * This, however, is no longer practical.
 * 
 * To clarify here, DOWNSTREAM signifies an SQUIT heading towards the target server
 * UPSTREAM signifies an SQUIT which has successfully completed, heading out everywhere.
 *
 * acptr is the server being squitted.
 * a DOWNSTREAM squit is where the notice did not come from acptr->from.
 * an UPSTREAM squit is where the notice DID come from acptr->from.
 *
 *        parv[0] = sender prefix 
 *        parv[1] = server name 
 *	  parv[2] = comment
 */

int m_squit(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *server;
   aClient *acptr;
   char *comment = (parc > 2) ? parv[2] : sptr->name;

   if (!IsPrivileged(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }

   if (parc > 1) 
   {
      server = parv[1];
      /*
       * The following allows wild cards in SQUIT. Only useful when
       * the command is issued by an oper.
       */
      if(IsServer(cptr))
         acptr = find_server_target(server);
      else
      {
         for (acptr = client; (acptr = next_client(acptr, server)); acptr = acptr->next)
         {
	    if (IsServer(acptr) || IsMe(acptr))
	       break;
         }
      }

      if (acptr && IsMe(acptr)) 
      {
         acptr = cptr;
         server = cptr->sockhost;
      }
   }
   else 
   {
      /* This is actually protocol error. But, well, closing the
       * link is very proper answer to that...
       */
      server = cptr->sockhost;
      acptr = cptr;
   }

   if (!acptr) 
   {
      send_me_numeric(sptr, ERR_NOSUCHSERVER, server);
      return 0;
   }

   if (MyClient(sptr) && ((!OPCanGRoute(sptr) && !MyConnect(acptr)) ||	
		(!OPCanLRoute(sptr) && MyConnect(acptr)))) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }

   /* If the server is mine, we don't care about upstream or downstream,
      just kill it and do the notice. */

   if (MyConnect(acptr)) 
   {
      sendto_gnotice("from %s: Received SQUIT %s from %s (%s)",
		     me.name, acptr->name, get_client_name(sptr, HIDEME), comment);
      send_serv_message_butone(NULL, &me, NULL, &CMD_GNOTICE, ":Received SQUIT %s from %s (%s)",
		               server, get_client_name(sptr, HIDEME), comment);

#if defined(USE_SYSLOG) && defined(SYSLOG_SQUIT)
      syslog(LOG_DEBUG, "SQUIT From %s : %s (%s)",
	     parv[0], server, comment);
#endif
      /* I am originating this squit! Not cptr! */
      /* ack, but if cptr is squitting itself.. */
      if(cptr == sptr)
      {
         exit_client(&me, acptr, sptr, comment);
         return CLIENT_DIED; /* kludge */
      }
      return exit_client(&me, acptr, sptr, comment);
   }

   /* the server is not connected to me. Determine whether this is an upstream
      or downstream squit */

   if(sptr->from == acptr->from) /* upstream */
   {
      sendto_ops_lev(DEBUG_LEV, "Exiting server %s due to upstream squit by %s [%s]",
         acptr->name, sptr->name, comment);
      return exit_client(cptr, acptr, sptr, comment);
   }

   /* fallthrough: downstream */

   if(!(IsUnconnect(acptr->from))) /* downstream not unconnect capable */
   {
      sendto_ops_lev(DEBUG_LEV, "Exiting server %s due to non-unconnect server %s [%s]",
         acptr->name, acptr->from->name, comment);
      return exit_client(&me, acptr, sptr, comment);
   }

   sendto_ops_lev(DEBUG_LEV, "Passing along SQUIT for %s by %s [%s]",
         acptr->name, sptr->name, comment);
   sendto_one_client_real(acptr->from, acptr, sptr, &CMD_SQUIT, ":%s", comment);

   return 0;
}
/*
 * * ts_servcount *   returns the number of TS servers that are
 * connected to us
 */
int
ts_servcount()
{
int     i;
aClient *acptr;
int     r = 0;

   for (i = 0; i <= highest_fd; i++)
      if ((acptr = local[i]) && IsServer(acptr) && DoesTS(acptr))
	 r++;
   return r;
}
/*
 * * m_svinfo 
 *       parv[0] = sender prefix 
 *       parv[1] = TS_CURRENT for the server 
 *       parv[2] = TS_MIN for the server 
 *       parv[3] = server is standalone or connected to non-TS only 
 *       parv[4] = server's idea of UTC time
 */
int m_svinfo(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   time_t deltat, tmptime, theirtime;

   if (!IsServer(sptr) || !MyConnect(sptr))
      return 0;

   if(parc == 2 && mycmp(parv[1], "ZIP") == 0)
   {
      SetZipIn(sptr);
      sptr->serv->zip_in = zip_create_input_session();
      sendto_gnotice("from %s :Input from %s is now compressed",
                     me.name, get_client_name(sptr, HIDEME));
      send_serv_message_butone(sptr, &me, NULL, &CMD_GNOTICE, ":Input from %s is now compressed",
                     me.name, get_client_name(sptr, HIDEME));
      return ZIP_NEXT_BUFFER;
   }

   if(parc < 5 || !DoesTS(sptr))
      return 0;

   if (TS_CURRENT < atoi(parv[2]) || atoi(parv[1]) < TS_MIN) 
   {
       /* a server with the wrong TS version connected; since we're
       * TS_ONLY we can't fall back to the non-TS protocol so we drop
       * the link  -orabidoo
       */
      sendto_ops("Link %s dropped, wrong TS protocol version (%s,%s)",
		 get_client_name(sptr, HIDEME), parv[1], parv[2]);
      return exit_client(cptr, cptr, cptr, "Incompatible TS version");
   }

   tmptime = time(NULL);
   theirtime = atol(parv[4]);
   deltat = abs(theirtime - tmptime);

   if (deltat > TS_MAX_DELTA) 
   {
      sendto_gnotice("from %s: Link %s dropped, excessive TS delta (my TS=%d, their TS=%d, delta=%d)",
		me.name, get_client_name(sptr, HIDEME), tmptime, theirtime, deltat);
      send_serv_message_butone(sptr, &me, NULL, &CMD_GNOTICE, ":Link %s dropped, excessive TS delta (delta=%d)",
		get_client_name(sptr, HIDEME), deltat);
      return exit_client(cptr, cptr, cptr, "Excessive TS delta");
   }

   if (deltat > TS_WARN_DELTA) 
   {
	sendto_realops("Link %s notable TS delta (my TS=%d, their TS=%d, delta=%d)",
	     get_client_name(sptr, HIDEME), tmptime, theirtime, deltat);
   }

   return 0;
}

/* m_myid
 * parv[1] = base64 server id
 */
int m_myid(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *base64id;

   /*
    * only accept this from unknown connections
    * who have not yet sent it 
    */

   if(!(IsUnknown(cptr) || IsServer(cptr) || IsHandshake(cptr) || IsConnecting(cptr)) || 
      cptr->id.string[0] || parc < 2 || !IsIDCapable(cptr) || cptr != sptr)
      return 0;

   base64id = parv[1];
   if(base64id[0] != '!' || base64id[1] == '\0' || !valid_base64_server_id(base64id + 1))
      return 0;
   else
      base64id++;

   strncpy(cptr->id.string, base64id, 8);
   return 0;
}

/*
 * m_server 
 * old style / connect style with MYID:
 * parv[0] = sender prefix 
 * parv[1] = servername
 * parv[2] = hopcount 
 * parv[3] = serverinfo
 * new style (remote servers only):
 * parv[0] = sender prefix
 * parv[1] = servername
 * parv[2] = hopcount
 * parv[3] = server identity
 * parv[4] = serverinfo
 */
int m_server(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char info[REALLEN + 1], *inpath, *host;
   aClient *acptr, *bcptr;
   aConfItem *aconf;
   int hop;
   char *base64id = NULL;
   char nbuf[HOSTLEN * 2 + USERLEN + 5]; /* same size as in s_misc.c */

   if (IsPerson(cptr)) 
   {
      /* don't let users register as SERVERs */
      send_me_numeric(cptr, ERR_UNKNOWNCOMMAND, "SERVER");
      return 0;
   } 
	
   info[0] = '\0';
   inpath = get_client_name(cptr, HIDEME);

   if (parc < 4) 
   {
      sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":Not enough parameters");
      return 0;
   }

   host = parv[1];
   hop = atoi(parv[2]);

   if(parc > 4 && IsIDCapable(cptr))
   {
      base64id = parv[3];
      if(base64id[0] != '!' || base64id[1] == '\0' || !valid_base64_server_id(base64id + 1))
      {
         sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":Invalid identity");
         return 0;
      }
      strncpyzt(info, parv[4], REALLEN);
      base64id++;
   }
   else
      strncpyzt(info, parv[3], REALLEN);

   /*
    * July 5, 1997 
    * Rewritten to throw away server cruft from users, 
    * combined the hostname validity test with cleanup of host name,
    * so a cleaned up hostname can be returned as an error if
    * necessary. - Dianora
    */

   /* yes, the if(strlen) below is really needed!! */
   if (strlen(host) > HOSTLEN)
	  host[HOSTLEN] = '\0';

   /* hostile servername check */	
   {
      /*
       * Lets check for bogus names and clean them up we don't bother
       * cleaning up ones from users, becasuse we will never see them
       * any more - Dianora
       */
		
      int bogus_server = 0;
      int found_dot = 0;
      char clean_host[(2 * HOSTLEN) + 1];
      char *s;
      char *d;
      int n;

      s = host;
      d = clean_host;
      n = (2 * HOSTLEN) - 2;

      while (*s && n > 0) 
      {
         if ((unsigned char) *s < (unsigned char) ' ')   /* Is it a control character? */
         {
            bogus_server = 1;
            *d++ = '^';
            *d++ = (char) ((unsigned char) *s + 0x40);   /* turn it into a printable */
            n =- 2;
         }
         else if ((unsigned char) *s > (unsigned char) '~')
         {
            bogus_server = 1;
            *d++ = '.';
            n--;
         }
         else 
         {
            if (*s == '.')
               found_dot = 1;
            *d++ = *s;
            n--;
         }
         s++;
      }
      *d = '\0';

      if ((!found_dot) || bogus_server) 
      {
         sendto_one_client_nopostfix(sptr, NULL, &CMD_ERROR, ":Bogus server name (%s)", clean_host);
         return exit_client(cptr, cptr, cptr, "Bogus server name");
      }
   }
	
   /*
    * check to see this host even has an N line before bothering 
    * anyone about it. Its only a quick sanity test to stop the
    * conference room and win95 ircd dorks. Sure, it will be
    * redundantly checked again in m_server_estab() *sigh* yes there
    * will be wasted CPU as the conf list will be scanned twice. But
    * how often will this happen? - Dianora
    * 
    * This should (will be) be recoded to check the IP is valid as well,
    * with a pointer to the valid N line conf kept for later, saving an
    * extra lookup.. *sigh* - Dianora
    */

   if(!IsServer(cptr)) 
   {
      if (!(aconf = find_conf_name(host, CONF_NOCONNECT_SERVER))) 
      {
#ifdef WARN_NO_NLINE
         sendto_realops("Link %s dropped, no N: line", get_client_name(cptr, TRUE));
#endif
         return exit_client(cptr, cptr, cptr, "NO N line");
      }
   }
	
   if ((acptr = find_server(host, NULL))) 
   {
      /*
       * This link is trying feed me a server that I already have 
       * access through another path -- multiple paths not accepted 
       * currently, kill this link immediately!! 
       * 
       * Rather than KILL the link which introduced it, KILL the 
       * youngest of the two links. -avalon
       */

      bcptr = (cptr->firsttime > acptr->from->firsttime) ? cptr : acptr->from;
      sendto_one_client_nopostfix(bcptr, NULL, &CMD_ERROR, ":Server %s already exists", host);
      if (bcptr == cptr) 
      {
         sendto_gnotice("from %s: Link %s cancelled, server %s already exists",
                        me.name, get_client_name(bcptr, HIDEME), host);
         send_serv_message_butone(bcptr, &me, NULL, &CMD_GNOTICE, ":Link %s cancelled, server %s already exists",
                                  get_client_name(bcptr, HIDEME), host);
         return exit_client(bcptr, bcptr, &me, "Server Exists");
      }
      /* inform all those who care (set +n) -epi */
      strcpy(nbuf, get_client_name(bcptr, HIDEME));
      sendto_gnotice("from %s: Link %s cancelled, server %s reintroduced by %s",
                     me.name, nbuf, host, get_client_name(cptr, HIDEME));
      send_serv_message_butone(bcptr, &me, NULL, &CMD_GNOTICE, ":Link %s cancelled, server %s reintroduced by %s",
		               nbuf, host, get_client_name(cptr, HIDEME));
      exit_client(bcptr, bcptr, &me, "Server Exists");
   }

   if(base64id && (acptr = find_server_by_base64_id(base64id, NULL)))
   {
      /*
       * we have duplicate identities on the network.
       * this is most likely the result of some misconfiguration.
       * throw it out.
       */

      sendto_one(cptr, "ERROR :Duplicate identity!");

      sendto_gnotice("from %s: Link %s cancelled, identity %s already held by %s",
                     me.name, get_client_name(cptr, HIDEME), base64id, acptr->name);
      send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Link %s cancelled, identity %s already held by %s",
		               get_client_name(cptr, HIDEME), base64id, acptr->name);

      return exit_client(cptr, cptr, &me, "Duplicate identity");
   }

   /*
    * The following if statement would be nice to remove since user
    * nicks never have '.' in them and servers must always have '.' in
    * them. There should never be a server/nick name collision, but it
    * is possible a capricious server admin could deliberately do
    * something strange.
    * 
    * -Dianora
    */
	
   if ((acptr = find_client(host, NULL)) && acptr != cptr) 
   {
      /*
       * * Server trying to use the same name as a person. Would 
       * cause a fair bit of confusion. Enough to make it hellish for
       * a while and servers to send stuff to the wrong place.
       */
      sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":Nickname %s already exists!", host);
      strcpy(nbuf, get_client_name(cptr, HIDEME));
      sendto_gnotice("from %s: Link %s cancelled, servername/nick collision",
                     me.name, nbuf);
      send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Link %s cancelled, servername/nick collision", nbuf);
      return exit_client(cptr, cptr, cptr, "Nick as Server");
   }
	
   if (IsServer(cptr)) 
   {
      /*
       * * Server is informing about a new server behind this link.
       * Create REMOTE server structure, add it to list and propagate
       * word to my other server links...
       */
      if (parc == 1 || info[0] == '\0') 
      {
	  sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":No server info specified for %s", host);
	  return 0;
      }

      if (!(aconf = find_conf_host(cptr->confs, host, CONF_HUB)) ||
			 (aconf->port && (hop > aconf->port))) 
      {
          sendto_gnotice("from %s: Non-Hub link %s introduced %s(%s).",
		me.name, get_client_name(cptr, HIDEME), host,
		aconf ? (aconf->host ? aconf->host : "*") : "!");
	  send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Non-Hub link %s introduced %s(%s).",
                                   get_client_name(cptr, HIDEME), host, aconf ? (aconf->host ? aconf->host : "*") : "!");
	  sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":%s has no H: line for %s.",
                                      get_client_name(cptr, HIDEME), host);
	  return exit_client(cptr, cptr, cptr, "Too many servers");
      }
		
      acptr = make_client(cptr, sptr);
      (void) make_server(acptr);
      acptr->hopcount = hop;
      strncpyzt(acptr->name, host, sizeof(acptr->name));
      strncpyzt(acptr->info, info, REALLEN);
      acptr->serv->up = find_or_add(parv[0]);
		
      SetServer(acptr);

      /* if this server is behind a U-lined server, make it U-lined as well. - lucas */

      if (IsULine(sptr) || (find_uline(cptr->confs, acptr->name))) 
      {
          acptr->flags |= FLAGS_ULINE;
          sendto_ops("%s introducing U:lined server %s", cptr->name, acptr->name);
      }
		
      Count.server++;
		
      add_client_to_list(acptr);
      add_to_client_hash_table(acptr->name, acptr);

      if(base64id)
         add_base64_server(acptr, base64id);

      if(HasID(acptr))
      {
         /*
          * since token1 and ident were developed simultaneously, only allow for them
          * to either both be on or both be off.
          */

         /*
          * if(HasID(acptr)) is true, HasID(sptr) has to be true as well!
          */
    
         sendto_flag_serv_butone(cptr, 0, CAPAB_IDENT|CAPAB_TOKEN1, ":%s " MSG_SERVER " %s %d :%s",
                                 sptr->name, acptr->name, hop + 1, acptr->info);

         sendto_flag_serv_butone(cptr, CAPAB_IDENT|CAPAB_TOKEN1, 0, "!%s " TOK1_SERVER " %s %d !%s :%s",
                                 sptr->id.string, acptr->name, hop + 1, acptr->id.string, acptr->info);
      }
      else if(HasID(sptr))
      {
         sendto_flag_serv_butone(cptr, 0, CAPAB_IDENT|CAPAB_TOKEN1, ":%s " MSG_SERVER " %s %d :%s",
                                 sptr->name, acptr->name, hop + 1, acptr->info);

         sendto_flag_serv_butone(cptr, CAPAB_IDENT|CAPAB_TOKEN1, 0, "!%s " TOK1_SERVER " %s %d :%s",
                                 sptr->id.string, acptr->name, hop + 1, acptr->info);
      }
      else
      {
         sendto_flag_serv_butone(cptr, 0, CAPAB_IDENT|CAPAB_TOKEN1, ":%s " MSG_SERVER " %s %d :%s",
                                 sptr->name, acptr->name, hop + 1, acptr->info);

         sendto_flag_serv_butone(cptr, CAPAB_IDENT|CAPAB_TOKEN1, 0, ":%s " TOK1_SERVER " %s %d :%s",
                                 sptr->name, acptr->name, hop + 1, acptr->info);
      }

      return 0;
   }
	
   if (!IsUnknown(cptr) && !IsHandshake(cptr))
	  return 0;

   /*
    * * A local link that is still in undefined state wants to be a
    * SERVER. Check if this is allowed and change status
    * accordingly...
    */
   /*
    * * Reject a direct nonTS server connection if we're TS_ONLY
    * -orabidoo
    */
	
   strncpyzt(cptr->name, host, sizeof(cptr->name));
   strncpyzt(cptr->info, info[0] ? info : me.name, REALLEN);
   cptr->hopcount = hop;
	
   switch (check_server_init(cptr)) 
   {
	 case 0:
		return m_server_estab(cptr);
	 case 1:
		sendto_ops("Access check for %s in progress",
			  get_client_name(cptr, HIDEME));
		return 1;
	 default:
		ircstp->is_ref++;
		sendto_ops("Received unauthorized connection from %s.",
			  get_client_host(cptr));
		return exit_client(cptr, cptr, cptr, "No C/N conf lines");
   }
	
}

static void sendnick_TS(aClient *cptr, aClient *sptr)
{
   char ubuf[32], r_tsbuf[8], r_sidbuf[8], r_ipbuf[8];
   char *tsbuf, *sidbuf, *ipbuf;

   if (IsPerson(sptr)) 
   {
      send_umode(NULL, sptr, 0, SEND_UMODES, ubuf);
      if (!*ubuf)	/* trivial optimization - Dianora */
      {
	 ubuf[0] = '+';
	 ubuf[1] = '\0';
      }

      if(HasID(sptr))
      {
         /*
          * since token1 and ident were developed simultaneously, only allow for them
          * to either both be on or both be off.
          */
         if(IsIDCapable(cptr) && IsToken1(cptr))
         {
            tsbuf = base64enc_r(sptr->tsinfo, r_tsbuf);
            sidbuf = base64enc_r(sptr->user->servicestamp, r_sidbuf);
            ipbuf = base64enc_r(sptr->ip.s_addr, r_ipbuf);
            sendto_one(cptr, TOK1_NICK " %s %d !%s %s %s %s !%s:%s %s :%s",
                               sptr->name, sptr->hopcount + 1, tsbuf, ubuf,
                               sptr->user->username, sptr->user->host, sptr->id.string,
                               ipbuf, sidbuf, sptr->info);
         }
         else
            sendto_one(cptr, MSG_NICK " %s %d %ld %s %s %s %s %lu %lu :%s",
                               sptr->name, sptr->hopcount + 1, sptr->tsinfo, ubuf,
                               sptr->user->username, sptr->user->host, sptr->user->server,
                               sptr->user->servicestamp, htonl(sptr->ip.s_addr), 
			       sptr->info);
      }
      else
      {
         if(IsIDCapable(cptr) && IsToken1(cptr))
         {
            tsbuf = base64enc_r(sptr->tsinfo, r_tsbuf);
            sidbuf = base64enc_r(sptr->user->servicestamp, r_sidbuf);

            sendto_one(cptr, TOK1_NICK " %s %d !%s %s %s %s %s %s :%s",
                               sptr->name, sptr->hopcount + 1, tsbuf, ubuf,
                               sptr->user->username, sptr->user->host, sptr->user->server,
                               sidbuf, sptr->info);
         }
         else
            sendto_one(cptr, MSG_NICK " %s %d %ld %s %s %s %s %lu %lu :%s",
                               sptr->name, sptr->hopcount + 1, sptr->tsinfo, ubuf,
                               sptr->user->username, sptr->user->host, sptr->user->server,
                               sptr->user->servicestamp, htonl(sptr->ip.s_addr), sptr->info);
      }
   }
}

int do_server_estab(aClient *cptr)
{
   aClient *acptr;
   aConfItem *aconf;
   aChannel *chptr;
#ifndef HUB
   aConfItem **confp;
#endif
   char *inpath = get_client_name(cptr, HIDEME); /* inpath == host */

   SetServer(cptr);

   Count.server++;
   Count.myserver++;

   if(IsZipCapable(cptr) && DoZipThis(cptr))
   {
      sendto_one(cptr, "SVINFO ZIP");
      SetZipOut(cptr);
      cptr->serv->zip_out = zip_create_output_session();
   }

#ifdef MAXBUFFERS
  /* let's try to bump up server sock_opts... -Taner */
   reset_sock_opts(cptr->fd, 1);
#endif 

   /* adds to fdlist */
   addto_fdlist(cptr->fd, &serv_fdlist);

#ifndef NO_PRIORITY
   /* this causes the server to be marked as "busy" */
   check_fdlists();
#endif

   cptr->pingval = get_client_ping(cptr);
   cptr->sendqlen = get_sendq(cptr);

   /* check here again for DKEY reasons */
   if ((acptr = hash_find_server(cptr->name, NULL)))
   {
        char        nbuf[HOSTLEN * 2 + USERLEN + 5]; /* same size as in s_misc.c */
        aClient *bcptr;

      /*
       * Rather than KILL the link which introduced it, KILL the
       * youngest of the two links. -avalon
       */

      bcptr = (cptr->firsttime > acptr->from->firsttime) ? cptr : acptr->from;
      sendto_one_client_nopostfix(bcptr, NULL, &CMD_ERROR, ":Server %s already exists (final phase)", cptr->name);
      if (bcptr == cptr)
      {
         sendto_gnotice("from %s: Link %s cancelled, server %s already exists (final phase)",
                        me.name, get_client_name(bcptr, HIDEME), cptr->name);
         send_serv_message_butone(bcptr, &me, NULL, &CMD_GNOTICE, 
				":Link %s cancelled, server %s already exists (final phase)",
                                  get_client_name(bcptr, HIDEME), cptr->name);
         return exit_client(bcptr, bcptr, &me, "Server Exists (final phase)");
      }
      /* inform all those who care (set +n) -epi */
      strcpy(nbuf, get_client_name(bcptr, HIDEME));
      sendto_gnotice("from %s: Link %s cancelled, server %s reintroduced by %s",
                     me.name, nbuf, cptr->name, get_client_name(cptr, HIDEME));
      send_serv_message_butone(bcptr, &me, NULL, &CMD_GNOTICE, 
				":Link %s cancelled, server %s reintroduced by %s (final phase)",
                               nbuf, cptr->name, get_client_name(cptr, HIDEME));
      exit_client(bcptr, bcptr, &me, "Server Exists (final phase)");
   }


#ifdef PINGNAZI
   nextping = timeofday;
#endif

   if(cptr->id.string[0])
   {
      char b64id[8];

      /*
       * we need to do this or we strcpy on top of ourselves.
       * somewhat kludgy, but this is the best way to do things without
       * writing another function to handle "local" servers - lucas
       */
      strcpy(b64id, cptr->id.string);

      if(add_base64_server(cptr, b64id))
      {
         char *sname;

         acptr = find_server_by_base64_id(b64id, NULL);
         sname = acptr ? acptr->name : "*UNKNOWN*";

         sendto_gnotice("from %s: Link %s overridden, identity %s already held by %s",
                        me.name, get_client_name(cptr, HIDEME), b64id, sname);
         send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Link %s overridden, identity %s already held by %s",
                                  get_client_name(cptr, HIDEME), b64id, sname);

         return exit_client(cptr, cptr, &me, "Duplicate identity overridden");
      }
   }

  /* error, error, error! if a server is U:'d, and it connects to us, 
   * we need to figure that out! So, do it here. - lucas 
   */ 

   if ((find_uline(cptr->confs, cptr->name))) 
   {
      Count.myulined++;
      cptr->flags |= FLAGS_ULINE; 
   }

   cptr->serv->burstflags |= BFLAGS_SENDBURST;

   sendto_gnotice("from %s: Link with %s established:%s%s%s%s", me.name, inpath, 
                ZipOut(cptr) ? " output-compressed" : "",
                RC4EncLink(cptr) ? " encrypted" : "",
                IsULine(cptr) ? " ULined" : "", 
		DoesTS(cptr) ? " TS link" : " Non-TS!"); 

   /* Notify everyone of the fact that this has just linked: the entire network should get two
      of these, one explaining the link between me->serv and the other between serv->me */

   send_serv_message_butone(NULL, &me, NULL, &CMD_GNOTICE, 
                            ":Link with %s established:%s%s%s%s", 
                            inpath, 
                            ZipOut(cptr) ? " output-compressed" : "",
                            RC4EncLink(cptr) ? " encrypted" : "",
                            IsULine(cptr) ? " ULined" : "", 
		            DoesTS(cptr) ? " TS link" : " Non-TS!"); 

   add_to_client_hash_table(cptr->name, cptr);

   /* add it to scache */

   find_or_add(cptr->name);

   if(HasID(cptr))
   {
      sendto_flag_serv_butone(cptr, 0, CAPAB_IDENT|CAPAB_TOKEN1, ":%s " MSG_SERVER " %s %d :%s",
                              me.name, cptr->name, 2, cptr->info);

      sendto_flag_serv_butone(cptr, CAPAB_IDENT|CAPAB_TOKEN1, 0, "!%s " TOK1_SERVER " %s %d !%s :%s",
                              me.id.string, cptr->name, 2, cptr->id.string, cptr->info);
   }
   else
   {
      sendto_flag_serv_butone(cptr, 0, CAPAB_IDENT|CAPAB_TOKEN1, ":%s " MSG_SERVER " %s %d :%s",
                              me.name, cptr->name, 2, cptr->info);

      sendto_flag_serv_butone(cptr, CAPAB_IDENT|CAPAB_TOKEN1, 0, "!%s " TOK1_SERVER " %s %d :%s",
                              me.id.string, cptr->name, 2, cptr->info);
   }

   /*
    * Pass on my client information to the new server
    * 
    * First, pass only servers (idea is that if the link gets 
    * cancelled beacause the server was already there, there are no
    * NICK's to be cancelled...). Of course, if cancellation occurs,
    * all this info is sent anyway, and I guess the link dies when a
    * read is attempted...? --msa 
    * 
    * Note: Link cancellation to occur at this point means that at
    * least two servers from my fragment are building up connection
    * this other fragment at the same time, it's a race condition,
    * not the normal way of operation... 
    * 
    * ALSO NOTE: using the get_client_name for server names-- see
    * previous *WARNING*!!! (Also, original inpath is
    * destroyed...)
    */

   aconf = cptr->serv->nline;
   for (acptr = &me; acptr; acptr = acptr->prev) 
   {
      /* acptr->from == acptr for acptr == cptr */
      if (acptr->from == cptr)
		  continue;
      if (IsServer(acptr)) 
      {
         if(IsIDCapable(cptr) && IsToken1(cptr))
         {
            if(HasID(acptr->uplink))
            {
               if(HasID(acptr))
               {
                  sendto_one(cptr, "!%s " TOK1_SERVER " %s %d !%s :%s",
                             acptr->uplink->id.string, acptr->name, acptr->hopcount + 1, acptr->id.string, acptr->info);
               }
               else
               {
                  sendto_one(cptr, "!%s " TOK1_SERVER " %s %d :%s",
                             acptr->uplink->id.string, acptr->name, acptr->hopcount + 1, acptr->info);
               }
            }
            else
            {
               sendto_one(cptr, ":%s " TOK1_SERVER " %s %d :%s",
                          acptr->uplink->name, acptr->name, acptr->hopcount + 1, acptr->info);
            }
         }
         else
         {
            sendto_one(cptr, ":%s " MSG_SERVER " %s %d :%s",
                       acptr->serv->up, acptr->name, acptr->hopcount + 1, acptr->info);
         }
      }
   }

   /* send out our SQLINES too, if we're a hub */
   /* if we're a client server, clear out our sqlines in anticipation of 
    * sqlines from the hub. - lucas
    */
#ifdef HUB
   for(aconf = conf; aconf; aconf=aconf->next) 
   {
      if(aconf->status & CONF_SQLINE) 
         sendto_one_client_nopostfix(cptr, &me, &CMD_SQLINE, "%s :%s", aconf->name, aconf->passwd);
      else if(aconf->status & CONF_SGLINE) 
         sendto_one_client_nopostfix(cptr, &me, &CMD_SGLINE, "%d :%s:%s", strlen(aconf->name), 
                                     aconf->name, aconf->passwd);
   }
#else
   for(confp = &conf; (aconf = *confp);)
   {
      if(aconf->status & (CONF_SQLINE|CONF_SGLINE))
      {
         *confp = aconf->next;
         free_conf(aconf);
      }
      else
         confp = &aconf->next;
   }
#endif

   /* Bursts are about to start.. send a BURST */
   if (IsBurst(cptr))
     sendto_one_client_nopostfix(cptr, NULL, &CMD_BURST, "");     
	
   /*
    * Send it in the shortened format with the TS, if it's a TS
    * server; walk the list of channels, sending all the nicks that
    * haven't been sent yet for each channel, then send the channel
    * itself -- it's less obvious than sending all nicks first, but
    * on the receiving side memory will be allocated more nicely 
    * saving a few seconds in the handling of a split -orabidoo
    */
   {
      chanMember *cm;
      static char nickissent = 1;

      /* init send_channel_modes for cptr */   
      send_channel_modes(cptr, NULL);

      nickissent = 3 - nickissent;
      /*
       * flag used for each nick to check if we've sent it yet - must
       * be different each time and !=0, so we alternate between 1 and
       * 2 -orabidoo
       */
      for (chptr = channel; chptr; chptr = chptr->nextch) 
      {
	 for (cm = chptr->members; cm; cm = cm->nextuser) 
	 {
	    acptr = cm->cptr;
	    if (acptr->nicksent != nickissent) 
	    {
	       acptr->nicksent = nickissent;
	       if (acptr->from != cptr)
		  sendnick_TS(cptr, acptr);
	    }
	 }
	 send_channel_modes(cptr, chptr);
      }
      /* also send out those that are not on any channel */
      for (acptr = &me; acptr; acptr = acptr->prev)
	 if (acptr->nicksent != nickissent) 
	 {
	    acptr->nicksent = nickissent;
	    if (acptr->from != cptr)
	       sendnick_TS(cptr, acptr);
	 }
   }

   if(ZipOut(cptr))
   {
      unsigned long inb, outb;
      double rat;
  
      zip_out_get_stats(cptr->serv->zip_out, &inb, &outb, &rat);

      if(inb)
      {
         sendto_gnotice("from %s: Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
                        me.name, get_client_name(cptr, HIDEME), inb, outb, rat);
         send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, 
                       ":Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
                        me.name, get_client_name(cptr, HIDEME), inb, outb, rat);
      }
   }

   /* stuff a PING at the end of this burst so we can figure out when
      the other side has finished processing it. */
   cptr->flags |= FLAGS_PINGSENT;
   cptr->serv->burstflags |= BFLAGS_BURST;
   cptr->serv->burstflags &= ~BFLAGS_SENDBURST;
   if (IsBurst(cptr)) 
      cptr->serv->burstflags |= BFLAGS_SOBSENT;
   sendto_one_client_nopostfix(cptr, &me, &CMD_PING, ":%s", me.name);

   return 0;
}

int m_server_estab(aClient *cptr)
{
   aConfItem *aconf, *bconf;
   char *inpath, *host, *s, *encr;

   inpath = get_client_name(cptr, HIDEME);  /* "refresh" inpath with host  */
   host = cptr->name;

   if (!(aconf = find_conf(cptr->confs, host, CONF_NOCONNECT_SERVER))) 
   {
      ircstp->is_ref++;
      sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":Access denied. No N line for server %s", inpath);
      sendto_ops("Access denied. No N line for server %s", inpath);
      return exit_client(cptr, cptr, cptr, "No N line for server");
   }
   if (!(bconf = find_conf(cptr->confs, host, CONF_CONNECT_SERVER))) 
   {
      ircstp->is_ref++;
      sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":Only N (no C) field for server %s", inpath);
      sendto_ops("Only N (no C) field for server %s", inpath);
      return exit_client(cptr, cptr, cptr, "No C line for server");
   }

   encr = cptr->passwd;
   if (*aconf->passwd && !StrEq(aconf->passwd, encr)) 
   {
      ircstp->is_ref++;
      sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":No Access (passwd mismatch) %s", inpath);
      sendto_ops("Access denied (passwd mismatch) %s", inpath);
      return exit_client(cptr, cptr, cptr, "Bad Password");
   }
   memset(cptr->passwd, '\0', sizeof(cptr->passwd));

#ifndef	HUB
   if (serv_fdlist.last_entry) 
   {
      ircstp->is_ref++;
      sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":I'm a leaf not a hub");
      return exit_client(cptr, cptr, cptr, "I'm a leaf");
   }
#endif

   /* aconf->port is a CAPAB field, kind-of. kludge. mm, mm. */
   cptr->capabilities |= aconf->port;

   if (IsUnknown(cptr)) 
   {
      char capabbuf[512];
      unsigned int dontwantflags = 0;

      if (bconf->passwd[0])
	 sendto_one_client_nopostfix(cptr, NULL, &CMD_PASS, "%s :TS", bconf->passwd);
      /* Pass my info to the new server */

#ifdef HAVE_ENCRYPTION_ON
      dontwantflags |= WantDKEY(cptr) ? 0 : CAPAB_DKEY;
#endif      

      make_capab_string(capabbuf, dontwantflags);

      sendto_one_client_nopostfix(cptr, NULL, &CMD_CAPAB, "%s", capabbuf);
      sendto_one_client_nopostfix(cptr, NULL, &CMD_MYID, "!%s", me.id.string);
      sendto_one_client_nopostfix(cptr, NULL, &CMD_SERVER, "%s 1 :%s", me.name,
                                  (me.info[0]) ? (me.info) : "IRCers United");
   }
   else {
      s = (char *) strchr(aconf->host, '@');
      *s = '\0';	/* should never be NULL -- wanna bet? -Dianora */

      Debug((DEBUG_INFO, "Check Usernames [%s]vs[%s]", aconf->host, cptr->username));
      if (match(aconf->host, cptr->username)) 
      {
	 *s = '@';
	 ircstp->is_ref++;
	 sendto_ops("Username mismatch [%s]v[%s] : %s",
		    aconf->host, cptr->username,
		    get_client_name(cptr, HIDEME));
	 sendto_one_client_nopostfix(cptr, NULL, &CMD_ERROR, ":No Username Match");
	 return exit_client(cptr, cptr, cptr, "Bad User");
      }
      *s = '@';
   }

   /* send routing notice, this should never happen anymore */
   if (!DoesTS(cptr)) 
   {
      sendto_gnotice("from %s: Warning: %s linked, non-TS server",
                     me.name, get_client_name(cptr, TRUE));
      send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Warning: %s linked, non-TS server",
                               get_client_name(cptr, TRUE));
   }

   sendto_one_client_nopostfix(cptr, NULL, &CMD_SVINFO, "%d %d 0 :%ld", TS_CURRENT, TS_MIN, (ts_val) timeofday);

   /* sendto one(cptr, "CAPAB ...."); moved to after PASS but before SERVER
    * now in two places.. up above and in s_bsd.c. - lucas
    * This is to make sure we pass on our capabilities before we establish a server connection */

   det_confs_butmask(cptr, CONF_HUB | CONF_NOCONNECT_SERVER | CONF_ULINE);

   /* doesnt duplicate cptr->serv if allocated this struct already */
   make_server(cptr);
   cptr->serv->up = me.name;
   cptr->serv->nline = aconf;

#ifdef THROTTLE_ENABLE
   throttle_remove(inetntoa((char *)&cptr->ip));
#endif

   /* now fill out the servers info so nobody knows dink about it. */
   memset((char *)&cptr->ip, '\0', sizeof(struct in_addr));
   strcpy(cptr->hostip, "127.0.0.1");
   strcpy(cptr->sockhost, "localhost");

#ifdef HAVE_ENCRYPTION_ON
   if(!CanDoDKEY(cptr) || !WantDKEY(cptr))
      return do_server_estab(cptr);
   else
   {
      SetNegoServer(cptr);
      sendto_one(cptr, "DKEY START");
   }
#else
   return do_server_estab(cptr);
#endif

   return 0;
}
/* 
 * m_burst
 *      parv[0] = sender prefix
 *      parv[1] = SendQ if an EOB
*/
int m_burst(aClient *cptr, aClient *sptr, int parc, char *parv[]) {
  
  if (!IsServer(sptr) || sptr != cptr || parc > 2 || !IsBurst(sptr)) return 0;
  if (parc == 2) { /* This is an EOB */
    sptr->serv->burstflags &= ~(BFLAGS_EOBRECV);
    if (sptr->serv->burstflags & (BFLAGS_SOBSENT|BFLAGS_BURST)) return 0;

 /* we've already sent our EOB.. we synched first
  * no need to check IsBurst because we shouldn't receive a BURST if they're not BURST capab */

 #ifdef HTM_LOCK_ON_NETBURST
    HTMLOCK = NO;
 #endif
    sendto_gnotice("from %s: synch to %s in %d %s at %s sendq", me.name, *parv,
        (timeofday-sptr->firsttime), (timeofday-sptr->firsttime)==1?"sec":"secs", parv[1]);
    send_serv_message_butone(NULL, &me, NULL, &CMD_GNOTICE, ":synch to %s in %d %s at %s sendq",
        sptr->name, (timeofday-sptr->firsttime), (timeofday-sptr->firsttime)==1?"sec":"secs", parv[1]);

  }
  else { /* SOB.. lock HTM if defined by admin */
#ifdef HTM_LOCK_ON_NETBURST
    HTMLOCK = YES;
#endif
    sptr->serv->burstflags |= BFLAGS_EOBRECV;
  }
  return 0;
}

/*
 * m_info 
 * parv[0] = sender prefix 
 * parv[1] = servername
 */
int m_info(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char **text = infotext;
   char outstr[241];
   static time_t last_used = 0L;

   if (use_or_deliver(cptr, sptr, &CMD_INFO, "%s", 1, parc, parv) != HUNTED_ISME) 
      return 0;

   if(parc > 2 && IsAnOper(sptr))
   {
      if(mycmp(parv[2], "MD5") == 0)
      {
         char **tmp = checksums;

         while(*tmp)
            send_me_numeric(sptr, RPL_INFO, *tmp++);
         send_me_numericNA(sptr, RPL_ENDOFINFO); 
         return 0;
      }

      if(mycmp(parv[2], "ULIMIT") == 0)
      {
         char **tmp = ulimit_result;

         while(*tmp)
            send_me_numeric(sptr, RPL_INFO, *tmp++);
         send_me_numericNA(sptr, RPL_ENDOFINFO); 
         return 0;
      }
   }
   
   sendto_realops_lev(SPY_LEV, "INFO requested by %s (%s@%s) [%s]",
      sptr->name, sptr->user->username, sptr->user->host,
      sptr->user->server);
         
   if (!IsAnOper(sptr)) 
   {
      if (IsSquelch(sptr))
      {
         send_me_numericNA(sptr, RPL_ENDOFINFO);
         return 0;
      }
      if (!MyConnect(sptr))
          return 0;
      if ((last_used + MOTD_WAIT) > NOW) 
          return 0;
      else 
         last_used = NOW;
   }

   while (*text)
      send_me_numeric(sptr, RPL_INFO, *text++);
         
   send_me_numeric(sptr, RPL_INFO, "");

   /* I am -definately- going to come up with a replacement for this! -Rak */

   if (IsAnOper(sptr)) 
   {
#ifdef ANTI_NICK_FLOOD
    strcpy(outstr, " ANTI_NICK_FLOOD=1");
#else
    strcpy(outstr, " ANTI_NICK_FLOOD=0");
#endif
#ifdef ANTI_SPAMBOT
    strcat(outstr, " ANTI_SPAMBOT=1");
#else
    strcat(outstr, " ANTI_SPAMBOT=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);


#ifdef CLIENT_COUNT
    strcpy(outstr, " CLIENT_COUNT=1");
#else
    strcpy(outstr, " CLIENT_COUNT=0");
#endif
#ifdef CLIENT_FLOOD
    strcat(outstr, " CLIENT_FLOOD=1");
#else
    strcat(outstr, " CLIENT_FLOOD=0");
#endif
#ifdef CLIENT_SERVER
    strcat(outstr, " CLIENT_SERVER=1");
#else
    strcat(outstr, " CLIENT_SERVER=0");
#endif
#ifdef DNS_DEBUG
    strcat(outstr, " DNS_DEBUG=1");
#else
    strcat(outstr, " DNS_DEBUG=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);

#ifdef DO_IDENTD
    strcpy(outstr, " DO_IDENTD=1");
#else
    strcpy(outstr, " DO_IDENTD=0");
#endif
#ifdef E_LINES_OPER_ONLY
    strcat(outstr, " E_LINES_OPER_ONLY=1");
#else
    strcat(outstr, " E_LINES_OPER_ONLY=0");
#endif
#ifdef FAILED_OPER_NOTICE
    strcat(outstr, " FAILED_OPER_NOTICE=1");
#else
    strcat(outstr, " FAILED_OPER_NOTICE=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);


#ifdef FLUD
    strcpy(outstr, " FLUD=1");
#else
    strcpy(outstr, " FLUD=0");
#endif
#ifdef SHORT_MOTD
    strcat(outstr, " SHORT_MOTD=1");
#else
    strcat(outstr, " SHORT_MOTD=0");
#endif
#ifdef F_LINES_OPER_ONLY
    strcat(outstr, " F_LINES_OPER_ONLY=1");
#else
    strcat(outstr, " F_LINES_OPER_ONLY=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef HIGHEST_CONNECTION
    strcpy(outstr, " HIGHEST_CONNECTION=1");
#else
    strcpy(outstr, " HIGHEST_CONNECTION=0");
#endif
#ifdef HUB
    strcat(outstr, " HUB=1");
#else
    strcat(outstr, " HUB=0");
#endif
#ifdef IDENTD_COMPLAIN
    strcat(outstr, " IDENTD_COMPLAIN=1");
#else
    strcat(outstr, " IDENTD_COMPLAIN=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef IGNORE_FIRST_CHAR
    strcpy(outstr, " IGNORE_FIRST_CHAR=1");
#else
    strcpy(outstr, " IGNORE_FIRST_CHAR=0");
#endif
#ifdef KPATH
    strcat(outstr, " KPATH=1");
#else
    strcat(outstr, " KPATH=0");
#endif
#ifdef LOCKFILE
    strcat(outstr, " LOCKFILE=1");
#else
    strcat(outstr, " LOCKFILE=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef MAXBUFFERS
    strcpy(outstr, " MAXBUFFERS=1");
#else
    strcpy(outstr, " MAXBUFFERS=0");
#endif
#ifdef NON_REDUNDANT_KLINES
    strcat(outstr, " NON_REDUNDANT_KLINES=1");
#else
    strcat(outstr, " NON_REDUNDANT_KLINES=0");
#endif
#ifdef NO_CHANOPS_WHEN_SPLIT
    strcat(outstr, " NO_CHANOPS_WHEN_SPLIT=1");
#else
    strcat(outstr, " NO_CHANOPS_WHEN_SPLIT=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef NO_DEFAULT_INVISIBLE
    strcpy(outstr, " NO_DEFAULT_INVISIBLE=1");
#else
    strcpy(outstr, " NO_DEFAULT_INVISIBLE=0");
#endif
#ifdef NO_MIXED_CASE
    strcat(outstr, " NO_MIXED_CASE=1");
#else
    strcat(outstr, " NO_MIXED_CASE=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef NO_OPER_FLOOD
    strcpy(outstr, " NO_OPER_FLOOD=1");
#else
    strcpy(outstr, " NO_OPER_FLOOD=0");
#endif
#ifdef NO_PRIORITY
    strcat(outstr, " NO_PRIORITY=1");
#else
    strcat(outstr, " NO_PRIORITY=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef OLD_Y_LIMIT
    strcpy(outstr, " OLD_Y_LIMIT=1");
#else
    strcpy(outstr, " OLD_Y_LIMIT=0");
#endif
#ifdef REJECT_IPHONE
    strcat(outstr, " REJECT_IPHONE=1");
#else
    strcat(outstr, " REJECT_IPHONE=0");
#endif
#ifdef RFC1035_ANAL
    strcat(outstr, " RFC1035_ANAL=1");
#else
    strcat(outstr, " RFC1035_ANAL=0");
#endif
#ifdef SHORT_MOTD
    strcat(outstr, " SHORT_MOTD=1");
#else
    strcat(outstr, " SHORT_MOTD=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);


#ifdef SHOW_HEADERS
    strcpy(outstr, " SHOW_HEADERS=1");
#else
    strcpy(outstr, " SHOW_HEADERS=0");
#endif
#ifdef SHOW_INVISIBLE_LUSERS
    strcat(outstr, " SHOW_INVISIBLE_LUSERS=1");
#else
    strcat(outstr, " SHOW_INVISIBLE_LUSERS=0");
#endif
#ifdef SHOW_UH
    strcat(outstr, " SHOW_UH=1");
#else
    strcat(outstr, " SHOW_UH=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef STATS_NOTICE
    strcpy(outstr, " STATS_NOTICE=1");
#else
    strcpy(outstr, " STATS_NOTICE=0");
#endif
#ifdef SUNDBE
    strcat(outstr, " SUNDBE=1");
#else
    strcat(outstr, " SUNDBE=0");
#endif
#ifdef UNKLINE
    strcat(outstr, " UNKLINE=1");
#else
    strcat(outstr, " UNKLINE=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);



#ifdef USERNAMES_IN_TRACE
    strcpy(outstr, " USERNAMES_IN_TRACE=1");
#else
    strcpy(outstr, " USERNAMES_IN_TRACE=0");
#endif
#ifdef USE_FAST_FD_ISSET
    strcat(outstr, " USE_FAST_FD_ISSET=1");
#else
    strcat(outstr, " USE_FAST_FD_ISSET=0");
#endif
#ifdef USE_SYSLOG
    strcat(outstr, " USE_SYSLOG=1");
#else
    strcat(outstr, " USE_SYSLOG=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);


#ifdef WARN_NO_NLINE
    strcpy(outstr, " WARN_NO_NLINE=1");
#else
    strcpy(outstr, " WARN_NO_NLINE=0");
#endif
    send_me_numeric(sptr, RPL_INFO, outstr);


#ifdef FLUD
    ircsprintf(outstr, "FLUD_BLOCK=%d FLUD_NUM=%d FLUD_TIME=%d",
          FLUD_BLOCK, FLUD_NUM, FLUD_TIME);
    send_me_numeric(sptr, RPL_INFO, outstr);
#endif


    ircsprintf(outstr, " HARD_FDLIMIT_=%d HYBRID_SOMAXCONN=%d", HARD_FDLIMIT_, HYBRID_SOMAXCONN);
    send_me_numeric(sptr, RPL_INFO, outstr);
    ircsprintf(outstr, " MOTD_WAIT=%d MOTD_MAX=%d MAXIMUM_LINKS=%d", MOTD_WAIT, MOTD_MAX, MAXIMUM_LINKS);
    send_me_numeric(sptr, RPL_INFO, outstr);
    ircsprintf(outstr, " MAX_NICK_CHANGES=%d MAX_NICK_TIME=%d", MAX_NICK_CHANGES, MAX_NICK_TIME);
    send_me_numeric(sptr, RPL_INFO, outstr);
    ircsprintf(outstr, " NICKNAMEHISTORYLENGTH=%d NOISY_HTM=%d", NICKNAMEHISTORYLENGTH, NOISY_HTM);
    send_me_numeric(sptr, RPL_INFO, outstr);
    ircsprintf(outstr, " MAXSENDQLENGTH=%d BUFFERPOOL=%d INIT_MAXCLIENTS=%d", MAXSENDQLENGTH, BUFFERPOOL, INIT_MAXCLIENTS);
    send_me_numeric(sptr, RPL_INFO, outstr);
    ircsprintf(outstr, " TS_MAX_DELTA=%d TS_WARN_DELTA=%d", TS_MAX_DELTA, TS_WARN_DELTA);
    send_me_numeric(sptr, RPL_INFO, outstr);
   }

   ircsprintf(outstr, "Genesis date: %s", genesis);
   send_me_numeric(sptr, RPL_INFO, outstr);
   ircsprintf(outstr, "Birth Date: %s, compile # %s", creation, generation);
   send_me_numeric(sptr, RPL_INFO, outstr);
   ircsprintf(outstr, "On-line since %s", myctime(me.firsttime));
   send_me_numeric(sptr, RPL_INFO, outstr);
   send_me_numericNA(sptr, RPL_ENDOFINFO);

   return 0;
}

/*
 * m_links 
 * parv[0] = sender prefix 
 * parv[1] = servername mask 
 * or 
 * parv[0] = sender prefix 
 * parv[1] = server to query 
 * parv[2] = servername mask
 */
int m_links(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char       *mask;
   aClient    *acptr;
   char        clean_mask[(2 * HOSTLEN) + 1];
   char       *s;
   char       *d;
   int         n;

   if (parc > 2) 
   {
      if (use_or_deliver(cptr, sptr, &CMD_LINKS, "%s :%s", 1, parc, parv) != HUNTED_ISME)
         return 0;
      mask = parv[2];
   }
   else
      mask = parc < 2 ? NULL : parv[1];

   /* reject non-local requests */
   if (!IsAnOper(sptr) && !MyConnect(sptr))
      return 0;

   /*
    * * sigh* Before the kiddies find this new and exciting way of
    * * annoying opers, lets clean up what is sent to all opers
    * * -Dianora
    */

   if (mask) 
   {      /* only necessary if there is a mask */
      s = mask;
      d = clean_mask;
      n = (2 * HOSTLEN) - 2;
      while (*s && n > 0) 
      {
         if ((unsigned char) *s < (unsigned char) ' ') /* Is it a control character? */
         {
            *d++ = '^';
            *d++ = (char) ((unsigned char)*s + 0x40);   /* turn it into a printable */
            s++;
            n -= 2;
         }
         else if ((unsigned char) *s > (unsigned char) '~') 
         {
            *d++ = '.';
            s++;
            n--;
         }
         else 
         {
            *d++ = *s++;
            n--;
         }
      }
      *d = '\0';
   }

   if (MyConnect(sptr))
      sendto_realops_lev(SPY_LEV,
          "LINKS %s requested by %s (%s@%s) [%s]",
          mask ? clean_mask : "*",
          sptr->name, sptr->user->username,
          sptr->user->host, sptr->user->server);

   for (acptr = client, (void) collapse(mask); acptr; acptr = acptr->next) 
   {
      if (!IsServer(acptr) && !IsMe(acptr))
         continue;
      if (!BadPtr(mask) && match(mask, acptr->name))
         continue;
#ifdef HIDEULINEDSERVS
      if (!IsOper(sptr) && IsULine(acptr))
         continue;
#endif
      send_me_numeric(sptr, RPL_LINKS, acptr->name, acptr->serv->up,
            acptr->hopcount, (acptr->info[0] ? acptr->info : "(Unknown Location)"));
   }

   send_me_numeric(sptr, RPL_ENDOFLINKS, BadPtr(mask) ? "*" : clean_mask);
   return 0;
}

static int  report_array[12][3] =
{
   {CONF_CONNECT_SERVER, RPL_STATSCLINE, 'C'},
   {CONF_NOCONNECT_SERVER, RPL_STATSNLINE, 'N'},
   {CONF_CLIENT, RPL_STATSILINE, 'I'},
   {CONF_GLINE, RPL_STATSGLINE, 'G'}, 
   {CONF_KILL, RPL_STATSKLINE, 'K'},
   {CONF_OPERATOR, RPL_STATSOLINE, 'O'},
   {CONF_HUB, RPL_STATSHLINE, 'H'},
   {CONF_LOCOP, RPL_STATSOLINE, 'o'},
   {CONF_QUARANTINE, RPL_STATSQLINE, 'Q'},
   {CONF_ULINE, RPL_STATSULINE, 'U'},
   {0, 0}
};

static void report_configured_links(aClient *sptr, int mask)
{
   static char null[] = "<NULL>";
   aConfItem  *tmp;
   int        *p, port;
   char        c, *host, *pass, *name;
   
   for (tmp = conf; tmp; tmp = tmp->next)
   {
     if (tmp->status & mask) 
     {
        for (p = &report_array[0][0]; *p; p += 3)
          if (*p&tmp->status)
            break;
        if (!*p)
          continue;
        c = (char) *(p + 2);
        host = BadPtr(tmp->host) ? null : tmp->host;
        pass = BadPtr(tmp->passwd) ? null : tmp->passwd;
        name = BadPtr(tmp->name) ? null : tmp->name;
        port = (int) tmp->port;

        /* block out IP addresses of U: lined servers in 
         * a C/N request - lucas/xpsycho
         * modified on request by taz, hide all
         * ips to other than routing staff or admins
         * -epi
         */

        if(tmp->status & CONF_SERVER_MASK)
	   host = "*";

        if (tmp->status & CONF_QUARANTINE)
           send_me_numeric(sptr, p[1], c, pass, name, port, get_conf_class(tmp));
        else if (tmp->status & CONF_GLINE)
           send_me_numeric(sptr, p[1], (tmp->status & CONF_SGLINE) ? tolowertab[(u_char)c] :
                           touppertab[(u_char)c], pass, name);
        else 
        {
           if(!IsAnOper(sptr) && !IsULine(sptr))
              send_me_numeric(sptr, p[1], c, "*", name, port, get_conf_class(tmp));
           else
              send_me_numeric(sptr, p[1], c, host, name, port, get_conf_class(tmp));
        }
     }
   }
   return;
}

/*
 * 
 * m_stats 
 *      parv[0] = sender prefix 
 *      parv[1] = statistics selector (defaults to Message frequency) 
 *      parv[2] = server name (current server defaulted, if omitted) 
 */

/*
 * 
 * m_stats/stats_conf 
 *    Report N/C-configuration lines from this server. This could 
 *    report other configuration lines too, but converting the 
 *    status back to "char" is a bit akward--not worth the code 
 *    it needs...
 * 
 *    Note:   The info is reported in the order the server uses 
 *    it--not reversed as in ircd.conf!
 */

/* create_numericmsg(&mp, numericbuf, 1234) */;
#define create_numericmsg(x, z, y) ircsprintf(z, "%d", y); (x)->token = (x)->message = z;

int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   static char Lformat[] = "%s %u %u %u %u %u %u %u %s";
   static char Sformat[] = "Name SendQ SendM SendBytes RcveM RcveBytes Open_since Idle TS";

   static char LLformat[] = "%s %u %u %u %u %u %u %u";
   static char LSformat[] = "Name Clients TotalClients SendM SendBytes RcveM RcveBytes Idle";

   extern aListener *listen_list;

   struct Message *mptr;
   aClient *acptr;
   aListener *lptr;
   char stat = parc > 1 ? parv[1][0] : '\0';
   int i;
   int doall = 0, wilds = 0;
   char *name;
   time_t sincetime;
   struct msg_pointer mp;
   char numericbuf[16];
   static time_t last_used = 0L;

   if (IsSquelch(sptr))
       return 0;

   if (use_or_deliver(cptr, sptr, &CMD_STATS, "%s %s", 2, parc, parv) != HUNTED_ISME)
       return 0;

   if (!IsAnOper(sptr) && !IsULine(sptr)) 
   {
      /* allow remote stats p l ? u */
      if (!((stat == 'p') || (stat == 'P') || (stat=='?') || (stat=='u') ||
          (stat=='l') || (stat=='L')) && !MyConnect(sptr))
        return 0;

      /* Penalize this user for doing a /stats command */
      if(MyClient(sptr))
         sptr->since += 5;

      if ((last_used + MOTD_WAIT) > NOW)
        return 0;
      else
        last_used = NOW;
   }
   
   if (parc > 2) 
   {
      name = parv[2];
      if (!mycmp(name, me.name))
        doall = 2;
      else if (match(name, me.name) == 0)
        doall = 1;
      if (strchr(name, '*') || strchr(name, '?'))
        wilds = 1;
   }
   else
     name = me.name;
   
   if (stat != (char) 0 && !IsULine(sptr))
     sendto_realops_lev(SPY_LEV, "STATS %c requested by %s (%s@%s) [%s]", stat,
         sptr->name, sptr->user->username, sptr->user->host,
         sptr->user->server);

   switch (stat) 
   {
    case 'L':
    case 'l':
      /* changed behavior totally.  This is what we do now:
       * #1: if the user is not opered, never return ips for anything
       * #2: we DON'T deny /stats l for opers.  Ever heard of /sping?
       *     it's easy to see if you're lagging a server, why restrict
       *     something used 99% of the time for good when you're not
       *     doing any harm?
       * #3: NEVER return all users on a server, UGH, just like
       *     /trace, this was fiercely obnoxious.  If you don't
       *     add an argument, you get all SERVER links.
       */
      create_numericmsg(&mp, numericbuf, RPL_STATSLINKINFO)
      sendto_one_client(sptr, &me, &mp, Sformat);

      if ((parc > 2) && !(doall || wilds))
      {         /* Single client lookup */
         if (!IsAnOper(sptr))
         {
            send_me_numericNA(sptr, ERR_NOPRIVILEGES);
            break;
         }
         if (!(acptr = find_person(name, NULL)) || !MyConnect(acptr))
            break;
         /*
          * sincetime might be greater than timeofday,
          * store a new value here to avoid sending 
          * negative since-times. -Rak
          */
         sincetime = (acptr->since > timeofday) ? 0 :
         timeofday - acptr->since;
         sendto_one_client(sptr, &me, &mp, Lformat, 
                           get_client_name(acptr, TRUE),
                           (int) DBufLength(&acptr->sendQ),
                           (int) acptr->sendM, (int) acptr->sendK,
                           (int) acptr->receiveM, (int) acptr->receiveK,
                           timeofday - acptr->firsttime,
                           sincetime,
                           IsServer(acptr) ? (DoesTS(acptr) ? "TS" : "NoTS") : "-");
      }
      else
      {
         for (i = 0; i <= highest_fd; i++) 
         {
            if (!(acptr = local[i]))
              continue;
            if(!IsServer(acptr))
              continue; /* nothing but servers */
#ifdef HIDEULINEDSERVS
            if(IsULine(acptr) && !IsAnOper(sptr))
              continue;
#endif
            sincetime = (acptr->since > timeofday) ? 0 :
            timeofday - acptr->since;
            sendto_one_client(sptr, &me, &mp, Lformat, 
                              get_client_name(acptr, HIDEME),
                              (int) DBufLength(&acptr->sendQ),
                              (int) acptr->sendM, (int) acptr->sendK,
                              (int) acptr->receiveM, (int) acptr->receiveK,
                              timeofday - acptr->firsttime,
                              sincetime,
                              IsServer(acptr) ? (DoesTS(acptr) ? "TS" : "NoTS") : "-");
         }
      }
      break;
    case 'C':
    case 'c':
#ifdef HIDEULINEDSERVS
      if (!IsAnOper(sptr))
          send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      else
#endif
         report_configured_links(sptr, CONF_CONNECT_SERVER | CONF_NOCONNECT_SERVER);
      break;
      
    case 'D':
      break;

    case 'E':
    case 'e':
#ifdef E_LINES_OPER_ONLY
      if (!IsAnOper(sptr))
        break;
#endif
      report_conf_links(sptr, &EList1, RPL_STATSELINE, 'E');
      report_conf_links(sptr, &EList2, RPL_STATSELINE, 'E');
      report_conf_links(sptr, &EList3, RPL_STATSELINE, 'E');
      break;
      
    case 'F':
    case 'f':
#ifdef F_LINES_OPER_ONLY
      if (!IsAnOper(sptr))
        break;
#endif
      report_conf_links(sptr, &FList1, RPL_STATSFLINE, 'F');
      report_conf_links(sptr, &FList2, RPL_STATSFLINE, 'F');
      report_conf_links(sptr, &FList3, RPL_STATSFLINE, 'F');
      break;
      
    case 'G':
    case 'g':
      if (!IsAnOper(sptr))
          sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
      else
         report_configured_links(sptr, CONF_GLINE);
      break;

    case 'H':
    case 'h':
#ifdef HIDEULINEDSERVS
      if (!IsOper(sptr))
          sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
      else
#endif
      report_configured_links(sptr, CONF_HUB);
      break;
      
    case 'I':
    case 'i':
      report_configured_links(sptr, CONF_CLIENT);
      break;

    case 'J':
    case 'j':
      if (!IsAnOper(sptr))
      {
         sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
         break;
      }
      create_numericmsg(&mp, numericbuf, RPL_STATSLINKINFO)
      sendto_one_client(sptr, &me, &mp, LSformat);
      for (lptr = listen_list; lptr; lptr = lptr->next) 
      {
         sendto_one_client(sptr, &me, &mp, LLformat,
            get_listener_name(lptr),
            lptr->confs->value.aconf->clients, lptr->ccount,
            (int) lptr->sendM, (int) lptr->sendK,
            (int) lptr->receiveM, (int) lptr->receiveK,
            timeofday - lptr->lasttime);
      }
      break;
      
    case 'k':
      if(IsAnOper(sptr))
         report_userbans_match_flags(cptr, UBAN_TEMPORARY|UBAN_LOCAL, 0);
      break;

    case 'K':
      if(IsAnOper(sptr))
         report_userbans_match_flags(cptr, UBAN_LOCAL, UBAN_TEMPORARY);
      break;  

    case 'A':
    case 'a':
      if(IsAnOper(sptr))
         report_userbans_match_flags(cptr, UBAN_NETWORK, 0);
      break;  
      
    case 'M':
      for (mptr = server_msgtab; mptr->cmd; mptr++)
        send_me_numeric(sptr, RPL_STATSCOMMANDS, mptr->cmd, mptr->count, mptr->bytes);
      break;

    case 'm':
      /*
       * original behaviour was not to report the command, if
       * the command hadn't been used. I'm going to always
       * report the command instead -Dianora
       */
      for (mptr = user_msgtab; mptr->cmd; mptr++)
        send_me_numeric(sptr, RPL_STATSCOMMANDS, mptr->cmd, mptr->count, mptr->bytes);
      break;
      
    case 'N':
    case 'n':
      send_me_numeric(sptr, RPL_STATSCOUNT, "User Connects Today: ", Count.today);
      send_me_numeric(sptr, RPL_STATSCOUNT, "User Connects past week: ", Count.weekly);
      send_me_numeric(sptr, RPL_STATSCOUNT, "User Connects past month: ", Count.monthly);
      send_me_numeric(sptr, RPL_STATSCOUNT, "User Connects past year: ", Count.yearly);
      break;
    case 'o':
    case 'O':
      report_configured_links(sptr, CONF_OPS);
      break;
      
    case 'p':
    case 'P':
      show_opers(sptr, parv[0]);
      break;
      
    case 'Q':
    case 'q':
      report_configured_links(sptr, CONF_QUARANTINE);
      break;
      
    case 'R':
    case 'r':
#ifdef DEBUGMODE
      send_usage(sptr, parv[0]);
#endif
      break;
      
    case 'S':
    case 's':
      if (IsAnOper(sptr))
        list_scache(cptr, sptr, parc, parv);
      else
        send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      break;
      
    case 'T':
	if (!IsAnOper(sptr))
	{
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		break;
	}
	throttle_stats(sptr, parv[0]);
	break;

    case 't':
      if (!IsAnOper(sptr)) 
      {
         send_me_numericNA(sptr, ERR_NOPRIVILEGES);
         break;
      }
      tstats(sptr, parv[0]);
      break;
      
    case 'U':
#ifdef HIDEULINEDSERVS
      if (!IsOper(sptr))
          send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      else
#endif
      report_configured_links(sptr, CONF_ULINE);
      break;
      
    case 'u':
        {
           time_t now;
           
           now = timeofday - me.since;
           send_me_numeric(sptr, RPL_STATSUPTIME, now / 86400, (now / 3600) % 24, (now / 60) % 60, now % 60);
#ifdef HIGHEST_CONNECTION
           send_me_numeric(sptr, RPL_STATSCONN, max_connection_count, max_client_count);
#endif
           break;
        }
      
    case 'v':
    case 'V':
      show_servers(sptr, parv[0]);
      break;
      
    case 'Y':
    case 'y':
      report_classes(sptr);
      break;
      
    case 'Z':
    case 'z':
      if (IsAnOper(sptr)) 
         count_memory(sptr, parv[0]);
      else
        send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      break;
      
    case '?':
      serv_info(sptr, parv[0]);
      
    default:
      stat = '*';
      break;
   }
   send_me_numeric(sptr, RPL_ENDOFSTATS, stat);
   return 0;
}
/*
 * * m_users 
 *        parv[0] = sender prefix 
 *        parv[1] = servername
 */
int
m_users(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   if (use_or_deliver(cptr, sptr, &CMD_USERS, ":%s", 1, parc, parv) == HUNTED_ISME) 
   {
      send_me_numeric(sptr, RPL_LOCALUSERS, Count.local, Count.max_loc);
      send_me_numeric(sptr, RPL_GLOBALUSERS, Count.total, Count.max_tot);
   }
   return 0;
}
/*
 * Note: At least at protocol level ERROR has only one parameter, 
 * although this is called internally from other functions  --msa 
 *
 *      parv[0] = sender prefix 
 *      parv[*] = parameters
 */
int m_error(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char   *para;

   para = (parc > 1 && *parv[1] != '\0') ? parv[1] : "no error";

   Debug((DEBUG_ERROR, "Received ERROR message from %s: %s",
	  sptr->name, para));

   if (!(IsServer(sptr) || IsConnecting(sptr) || IsHandshake(sptr)))
      return 0;

   if (cptr == sptr)
   {
      sendto_gnotice("from %s: ERROR from %s: %s",
                     me.name, get_client_name(cptr, HIDEME), para);
      send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, ":ERROR from %s: %s",
                               get_client_name(cptr, HIDEME), para);
   }
   else
   {
      sendto_gnotice("from %s: ERROR from %s via %s: %s",
                     me.name, sptr->name, get_client_name(cptr, HIDEME), para);
      send_serv_message_butone(cptr, &me, NULL, &CMD_GNOTICE, ":ERROR from %s via %s: %s",
                               sptr->name, get_client_name(cptr, HIDEME), para);
   }
   return 0;
}
/*
 * m_help 
 * parv[0] = sender prefix
 */
int m_help(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   int i;
   aMotd *helpfile_ptr;
   static time_t last_used = 0L;

   if (!IsAnOper(sptr)) 
   {
      /* reject non local requests */
      if ((last_used + MOTD_WAIT) > NOW)
	 return 0;
      else
	 last_used = NOW;
   }

   if (!IsAnOper(sptr) || (helpfile == (aMotd *) NULL)) 
   {
      for (i = 0; user_msgtab[i].cmd; i++)
	 send_me_notice(sptr, ":%s", user_msgtab[i].cmd);
      return 0;
   }

   helpfile_ptr = helpfile;
   while (helpfile_ptr) 
   {
      send_me_notice(sptr, ":%s", helpfile_ptr->line);
      helpfile_ptr = helpfile_ptr->next;
   }

   return 0;
}
/*
 * parv[0] = sender parv[1] = host/server mask. 
 * parv[2] = server to query
 * 
 * 199970918 JRL hacked to ignore parv[1] completely and require parc > 3
 * to cause a force
 */
int m_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   int send_lusers(aClient *, aClient *, int, char **);

   if (parc > 2) 
   {
      if (use_or_deliver(cptr, sptr, &CMD_LUSERS, "%s %s", 2, parc, parv) != HUNTED_ISME)
         return 0;
   }		
   return send_lusers(cptr,sptr,parc,parv);
}
/*
 * send_lusers
 *     parv[0] = sender
 *     parv[1] = host/server mask.
 *     parv[2] = server to query
 */
int send_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
#define LUSERS_CACHE_TIME 180
   static long 		 last_time=0;
   static int   	 s_count = 0, c_count = 0, u_count = 0, i_count = 0;
   static int   	 o_count = 0, m_client = 0, m_server = 0, m_ulined = 0;
   int 			 forced;
   aClient 		*acptr;

   forced = (IsAnOper(sptr) && (parc > 3));

   Count.unknown = 0;
   m_server = Count.myserver;
   m_ulined = Count.myulined;
   m_client = Count.local;
   i_count = Count.invisi;
   u_count = Count.unknown;
   c_count = Count.total - Count.invisi;
   s_count = Count.server;
   o_count = Count.oper;
   if (forced || (timeofday > last_time + LUSERS_CACHE_TIME)) 
   {
      last_time = timeofday;
      /* only recount if more than a second has passed since last request
       * use LUSERS_CACHE_TIME instead... 
       */
      s_count = 0;
      c_count = 0;
      u_count = 0;
      i_count = 0;
      o_count = 0;
      m_client = 0;
      m_server = 0;
      m_ulined = 0;

      for (acptr = client; acptr; acptr = acptr->next) 
      {
	 switch (acptr->status) 
	 {
	    case STAT_SERVER:
	       if (MyConnect(acptr))
               {
		  m_server++;
                  if(IsULine(acptr))
                     m_ulined++;
               }
	    case STAT_ME:
	       s_count++;
	       break;
	    case STAT_CLIENT:
	       if (IsAnOper(acptr))
		  o_count++;
#ifdef	SHOW_INVISIBLE_LUSERS
	       if (MyConnect(acptr))
		  m_client++;
	       if (!IsInvisible(acptr))
		  c_count++;
	       else
		  i_count++;
#else
	       if (MyConnect(acptr)) 
	       {
		  if (IsInvisible(acptr)) 
		  {
		     if (IsAnOper(sptr))
			m_client++;
		  }
		  else
		     m_client++;
	       }
	       if (!IsInvisible(acptr))
		  c_count++;
	       else
		  i_count++;
#endif
	       break;
	    default:
	       u_count++;
	       break;
	 }
      }
      /*
       * We only want to reassign the global counts if the recount time
       * has expired, and NOT when it was forced, since someone may
       * supply a mask which will only count part of the userbase
       * -Taner
       */
      if (!forced) 
      {
	 if (m_server != Count.myserver) 
	 {
	    sendto_realops_lev(DEBUG_LEV,
			       "Local server count off by %d",
			       Count.myserver - m_server);
	    Count.myserver = m_server;
	 }
	 if (m_ulined != Count.myulined) 
	 {
	    sendto_realops_lev(DEBUG_LEV,
			       "Local ulinedserver count off by %d",
			       Count.myulined - m_ulined);
	    Count.myulined = m_ulined;
	 }
	 if (s_count != Count.server) 
	 {
	    sendto_realops_lev(DEBUG_LEV,
			       "Server count off by %d",
			       Count.server - s_count);
	    Count.server = s_count;
	 }
	 if (i_count != Count.invisi) 
	 {
	    sendto_realops_lev(DEBUG_LEV,
			       "Invisible client count off by %d",
			       Count.invisi - i_count);
	    Count.invisi = i_count;
	 }
	 if ((c_count + i_count) != Count.total) 
	 {
	    sendto_realops_lev(DEBUG_LEV, "Total client count off by %d",
			       Count.total - (c_count + i_count));
	    Count.total = c_count + i_count;
	 }
	 if (m_client != Count.local) 
	 {
	    sendto_realops_lev(DEBUG_LEV,
			       "Local client count off by %d",
			       Count.local - m_client);
	    Count.local = m_client;
	 }
	 if (o_count != Count.oper) 
	 {
	    sendto_realops_lev(DEBUG_LEV,
			"Oper count off by %d", Count.oper - o_count);
	    Count.oper = o_count;
	 }
	 Count.unknown = u_count;
      }				/* Complain & reset loop */
   }				/* Recount loop */

#ifdef SAVE_MAXCLIENT_STATS
   if ((timeofday - last_stat_save) > SAVE_MAXCLIENT_STATS_TIME) { /* save stats */
     FILE *fp;

     last_stat_save = timeofday;

     fp=fopen(DPATH "/.maxclients", "w");
     if (fp!=NULL) {
       fprintf(fp, "%d %d %li %li %li %ld %ld %ld %ld", Count.max_loc,
               Count.max_tot, Count.weekly, Count.monthly,
               Count.yearly, Count.start, Count.week, Count.month,
               Count.year);
       fclose(fp);
       sendto_realops_lev(DEBUG_LEV, "Saved maxclient statistics");
     }
   }
#endif
     

#ifndef	SHOW_INVISIBLE_LUSERS
   if (IsAnOper(sptr) && i_count)
#endif
      send_me_numeric(sptr, RPL_LUSERCLIENT, c_count, i_count, s_count);
#ifndef	SHOW_INVISIBLE_LUSERS
   else
      send_me_numeric(sptr, RPL_LUSERCLIENT, c_count, 0, s_count);
#endif
   if (o_count)
      send_me_numeric(sptr, RPL_LUSEROP, o_count);
   if (u_count > 0)
      send_me_numeric(sptr, RPL_LUSERUNKNOWN, u_count);

   /* This should be ok */
   if (Count.chan > 0)
      send_me_numeric(sptr, RPL_LUSERCHANNELS, Count.chan);

#ifdef HIDEULINEDSERVS
   send_me_numeric(sptr, RPL_LUSERME, m_client, IsAnOper(sptr) ? m_server : m_server - m_ulined);
#else
   send_me_numeric(sptr, RPL_LUSERME, m_client, m_server);
#endif

#ifdef CLIENT_COUNT
   send_me_numeric(sptr, RPL_LOCALUSERS, Count.local, Count.max_loc);
   send_me_numeric(sptr, RPL_GLOBALUSERS, Count.total, Count.max_tot);
#else
# ifdef HIGHEST_CONNECTION
   send_me_numeric(sptr, RPL_STATSCONN, max_connection_count, max_client_count);
   if (m_client > max_client_count)
      max_client_count = m_client;
   if ((m_client + m_server) > max_connection_count) 
   {
      max_connection_count = m_client + m_server;
      if (max_connection_count % 10 == 0)
	 sendto_ops("New highest connections: %d (%d clients)",
                    max_connection_count, max_client_count);
   }
# endif /* HIGHEST_CONNECTION */
#endif /* CLIENT_COUNT */
   return 0;
}

/***********************************************************************
 * m_connect() - Added by Jto 11 Feb 1989
 ***********************************************************************/
/*
 * * m_connect 
 *      parv[0] = sender prefix 
 *      parv[1] = servername 
 * 	parv[2] = port number 
 * 	parv[3] = remote server
 */
int m_connect(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   int         port, tmpport, retval;
   aConfItem  *aconf;
   aClient    *acptr;

   if (!IsPrivileged(sptr))
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return -1;
   }

   if ((MyClient(sptr) && !OPCanGRoute(sptr) && parc > 3) ||
       (MyClient(sptr) && !OPCanLRoute(sptr) && parc <= 3))
   {
       send_me_numericNA(sptr, ERR_NOPRIVILEGES);
       return 0;
   }

   if (use_or_deliver(cptr, sptr, &CMD_CONNECT, "%s %s %s", 3, parc, parv) != HUNTED_ISME)
      return 0;

   if (parc < 2 || *parv[1] == '\0') 
   {
      send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "CONNECT");
      return -1;
   }

   if ((acptr = find_server(parv[1], NULL)))
   {
      send_me_notice(sptr, ":Connect: Server %s already exists from %s.", parv[1], acptr->from->name);
      return 0;
   }

   for (aconf = conf; aconf; aconf = aconf->next)
      if (aconf->status == CONF_CONNECT_SERVER && match(parv[1], aconf->name) == 0)
	 break;

   /* Checked first servernames, then try hostnames. */
   if (!aconf)
      for (aconf = conf; aconf; aconf = aconf->next)
	 if (aconf->status == CONF_CONNECT_SERVER &&
	     (match(parv[1], aconf->host) == 0 ||
	      match(parv[1], strchr(aconf->host, '@') + 1) == 0))
	    break;

   if (!aconf) 
   {
      send_me_notice(sptr, ":Connect: No C line found for %s.", parv[1]);
      return 0;
   }
   /*
    * * Get port number from user, if given. If not specified, use
    * the default form configuration structure. If missing from
    * there, then use the precompiled default.
    */
   tmpport = port = aconf->port;
   if (parc > 2 && !BadPtr(parv[2])) 
   {
      if ((port = atoi(parv[2])) <= 0) 
      {
	 send_me_noticeNA(sptr, ":Connect: Illegal port number");
	 return 0;
      }
   }
   else if (port <= 0 && (port = SERV_PORTNUM) <= 0) 
   {
      send_me_noticeNA(sptr, ":Connect: missing port number");
      return 0;
   }

   /*
    * Notify all operators about remote connect requests
    * Let's notify about local connects, too. - lucas
    * sendto_ops_butone -> sendto_serv_butone(), like in df. -mjs
    */
      sendto_gnotice("from %s: %s CONNECT %s %s from %s",
		me.name,  IsAnOper(cptr) ? "Local" : "Remote", 
		parv[1], parv[2] ? parv[2] : "",
		get_client_name(sptr, HIDEME));
      send_serv_message_butone(NULL, &me, NULL, &CMD_GNOTICE,":%s CONNECT %s %s from %s", 
		IsAnOper(cptr) ? "Local" : "Remote",
		parv[1], parv[2] ? parv[2] : "",
		get_client_name(sptr, HIDEME));

#if defined(USE_SYSLOG) && defined(SYSLOG_CONNECT)
   syslog(LOG_DEBUG, "CONNECT From %s : %s %s", parv[0], parv[1], parv[2] ? parv[2] : "");
#endif

   aconf->port = port;
   switch (retval = connect_server(aconf, sptr, NULL))
   {
      case 0:
	 send_me_notice(sptr, ":*** Connecting to %s.", aconf->name);
	 break;

      case -1:
	 send_me_notice(sptr, ":*** Couldn't connect to %s.", aconf->name);
	 break;

      case -2:
	 send_me_notice(sptr, ":*** Host for %s is unknown.", aconf->name);
	 break;

      default:
	 send_me_notice(sptr, ":*** Connection to %s failed: %s", aconf->name, strerror(retval));
   }
   aconf->port = tmpport;
   return 0;
}

/*
 * m_wallops (write to *all* opers currently online) 
 * parv[0] = sender prefix 
 * parv[1] = message text
 */
int m_wallops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char       *message = parc > 1 ? parv[1] : NULL;

   if (BadPtr(message)) 
   {
      send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "WALLOPS");
      return 0;
   }

   if (!IsServer(sptr) && MyConnect(sptr) && !OPCanWallOps(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }
   
   sendto_wallops(sptr, ":%s WALLOPS :%s", parv[0], message);
   send_serv_message_butone(cptr, sptr, NULL, &CMD_WALLOPS, ":%s", message);
   return 0;
}

/*
 * m_locops (write to *all* local opers currently online) 
 * parv[0] = sender prefix 
 * parv[1] = message text
 */
int m_locops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *message = parc > 1 ? parv[1] : NULL;

   if (BadPtr(message)) 
   {
      send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "LOCOPS");
      return 0;
   }

   if (!IsServer(sptr) && MyConnect(sptr) && !OPCanLocOps(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }
   sendto_locops("from %s: %s", parv[0], message);
   return (0);
}

/*
 * m_goper  (Russell) sort of like wallop, but only to ALL +o clients
 * on every server. 
 * parv[0] = sender prefix 
 * parv[1] = message text 
 * Taken from df465, ported to hybrid. -mjs
 */
int m_goper(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *message = parc > 1 ? parv[1] : NULL;

   if (BadPtr(message)) 
   {
      send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "GOPER");
      return 0;
   }

   if (!IsServer(sptr) || !IsULine(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }

   send_serv_message_butone(cptr, sptr, NULL, &CMD_GOPER, ":%s", message);
   sendto_ops("from %s: %s", parv[0], message);
   return 0;
}

/*
 * m_gnotice  (Russell) sort of like wallop, but only to +g clients on *
 * this server. 
 * parv[0] = sender prefix 
 * parv[1] = message text 
 * ported from df465 to hybrid -mjs
 *
 * This function itself doesnt need any changes for the move to +n routing
 * notices, to sendto takes care of it.  Now only sends to +n clients -epi
 */
int m_gnotice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *message = parc > 1 ? parv[1] : NULL;

   if (BadPtr(message)) 
   {
      send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "GNOTICE");
      return 0;
   }

   if (!IsServer(sptr) && MyConnect(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }

   send_serv_message_butone(cptr, sptr, NULL, &CMD_GNOTICE, ":%s", message);
   sendto_gnotice("from %s: %s", parv[0], message);
   return 0;
}

int m_globops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *message = parc > 1 ? parv[1] : NULL;

   /* a few changes, servers weren't able to globop -mjs */

   if (BadPtr(message)) 
   {
      send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "GLOBOPS");
      return 0;
   }

   /* must be a client, must be an oper or a Ulined server -mjs */

   if (MyClient(sptr) && !OPCanGlobOps(sptr) && !IsULine(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }

   send_serv_message_butone(cptr, sptr, NULL, &CMD_GLOBOPS, ":%s", message);
   send_globops("from %s: %s", parv[0], message);
   return 0;
}

int m_chatops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *message = parc > 1 ? parv[1] : NULL;

   if (BadPtr(message)) 
   {
      send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "CHATOPS");
      return 0;
   }

   if (MyClient(sptr) && (!IsAnOper(sptr) || !SendChatops(sptr)) && !IsULine(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }

   send_serv_message_butone(cptr, sptr, NULL, &CMD_CHATOPS, ":%s", message);
   send_chatops("from %s: %s", parv[0], message);
   return 0;
}

/*
 * m_time 
 * parv[0] = sender prefix 
 * parv[1] = servername
 */

int m_time(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   if (use_or_deliver(cptr, sptr, &CMD_TIME, "%s", 1, parc, parv) == HUNTED_ISME)
      send_me_numeric(sptr, RPL_TIME, me.name, date((long) 0));
   return 0;
}

/*
 * m_admin
 * parv[0] = sender prefix 
 * parv[1] = servername
 */
int m_admin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aConfItem  *aconf;

   if (use_or_deliver(cptr, sptr, &CMD_ADMIN, "%s", 1, parc, parv) != HUNTED_ISME)
      return 0;

   if (IsPerson(sptr))
      sendto_realops_lev(SPY_LEV, "ADMIN requested by %s (%s@%s) [%s]", sptr->name,
	  sptr->user->username, sptr->user->host, sptr->user->server);

   if ((aconf = find_admin())) 
   {
      send_me_numeric(sptr, RPL_ADMINME, me.name);
      send_me_numeric(sptr, RPL_ADMINLOC1, aconf->host ? aconf->host : "");
      send_me_numeric(sptr, RPL_ADMINLOC2, aconf->passwd ? aconf->passwd : "");
      send_me_numeric(sptr, RPL_ADMINEMAIL, aconf->name ? aconf->name : "");
   }
   else
      send_me_numeric(sptr, ERR_NOADMININFO, me.name);

   return 0;
}

/* Shadowfax's server side, anti flood code */
#ifdef FLUD
extern int  flud_num;
extern int  flud_time;
extern int  flud_block;
#endif

#ifdef ANTI_SPAMBOT
extern int  spam_num;
extern int  spam_time;
#endif

#ifdef NO_CHANOPS_WHEN_SPLIT
extern int  server_split_recovery_time;
#endif

/*
 * m_set - set options while running
 */
int m_set(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char       *command;

   if (!MyClient(sptr) || !IsOper(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }

   if (parc > 1) 
   {
      command = parv[1];
      if (!strncasecmp(command, "MAX", 3)) 
      {
	 if (parc > 2) 
         {
  	    int new_value = atoi(parv[2]);

	    if (new_value > MASTER_MAX) 
	    {
	       send_me_notice(sptr, ":You cannot set MAXCLIENTS to > MASTER_MAX (%d)", MASTER_MAX);
	       return 0;
	    }
	    if (new_value < 32) 
	    {
	       send_me_notice(sptr, ":You cannot set MAXCLIENTS to < 32 (%d:%d)", MAXCLIENTS, highest_fd);
	       return 0;
	    }

	    MAXCLIENTS = new_value;
	    send_me_notice(sptr, ":NEW MAXCLIENTS = %d (Current = %d)", MAXCLIENTS, Count.local);
	    sendto_realops("%s!%s@%s set new MAXCLIENTS to %d (%d current)",
			   parv[0], sptr->user->username, sptr->sockhost, MAXCLIENTS, Count.local);
	    return 0;
	 }
         send_me_notice(sptr, ":Current Maxclients = %d (%d)", MAXCLIENTS, Count.local);
	 return 0;
      }
      else if (!strncasecmp(command, "THROTTLE", 8))
      {
           /* several values available:
            * ENABLE [on|off] to enable the code
            * COUNT [n] to set a max count, must be > 1
            * TIME [n] to set a max time before expiry, must be > 5
            * RECORDTIME [n] to set a time for the throttle records to expire
            * HASH [n] to set the size of the hash table, must be bigger than
            * the default */

		char *changed = NULL;
		char *to = NULL;
		
		if (parc > 3)
		{
			if (!strcasecmp(parv[2], "ENABLE"))
			{
				changed = "ENABLE";
				if (tolower(*parv[3]) == 'y' || !strcasecmp(parv[3], "on"))
				{
					throttle_enable = 1;
					to = "ON";
				}
				else if (tolower(*parv[3]) == 'n' || !strcasecmp(parv[3], "off"))
				{
					throttle_enable = 0;
					to = "OFF";
				}
			}
			else if (!strcasecmp(parv[2], "COUNT"))
			{
				int cnt;
				changed = "COUNT";
				cnt = atoi(parv[3]);
				if (cnt > 1)
				{
					throttle_tcount = cnt;
					to = parv[3];
				}
			}
			else if (!strcasecmp(parv[2], "TIME"))
			{
				int cnt;
				changed = "TIME";
				cnt = atoi(parv[3]);
				if (cnt >= 5)
				{
					throttle_ttime = cnt;
					to = parv[3];
				}
			}
			else if (!strcasecmp(parv[2], "RECORDTIME"))
			{
				int cnt;
				changed = "RECORDTIME";
				cnt = atoi(parv[3]);
				if (cnt >= 30)
				{
					throttle_rtime = cnt;
					to = parv[3];
				}
			}
			else if (!strcasecmp(parv[2], "HASH"))
			{
				int cnt;
				changed = "HASH";
				cnt = atoi(parv[3]);
				if (cnt >= THROTTLE_HASHSIZE)
				{
#ifdef THROTTLE_ENABLE
					throttle_resize(cnt);
#endif
					to = parv[3];
				}
			}

			if (to != NULL)
			{
				sendto_ops("%s has changed throttle %s to %s",
					   parv[0], changed, to);
				/* there used to be a send_me_notice here, but its
				 * frickin redundant, so i cut it.  -epi
				 */
			}
		}
		else
		{
			/* they arent changing anything, so report general stuff. */
			send_me_notice(sptr, ":THROTTLE %s", throttle_enable ? "enabled" : "disabled");
			send_me_notice(sptr, ":THROTTLE COUNT=%d", throttle_tcount);
			send_me_notice(sptr, ":THROTTLE TIME=%d sec", throttle_ttime);
			send_me_notice(sptr, ":THROTTLE RECORDTIME=%d sec", throttle_rtime);
		}
	}
#ifdef FLUD
      else if (!strncasecmp(command, "FLUDNUM", 7)) 
      {
	 if (parc > 2) 
 	 {
            int newval = atoi(parv[2]);
	    if (newval <= 0) 
 	    {
	       send_me_noticeNA(sptr, ":flud NUM must be > 0");
	       return 0;
	    }
	    flud_num = newval;
	    sendto_ops("%s has changed flud NUM to %i", parv[0], flud_num);
	    send_me_notice(sptr, ":flud NUM is now set to %i", flud_num);
	    return 0;
	 }
	 else 
         {
	    send_me_notice(sptr, ":flud NUM is currently %i", flud_num);            
	    return 0;
	 }
      }
      else if (!strncasecmp(command, "FLUDTIME", 8)) 
      {
	 if (parc > 2) 
	 {
  	    int newval = atoi(parv[2]);
	    if (newval <= 0) 
	    {
	       send_me_noticeNA(sptr, ":flud TIME must be > 0");
	       return 0;
	    }
	    flud_time = newval;
	    sendto_ops("%s has changed flud TIME to %i", parv[0], flud_time);
	    send_me_notice(sptr, ":flud TIME is now set to %i", flud_time);
	    return 0;
	 }
	 else 
         {
	    send_me_notice(sptr, ":flud TIME is now currently %i", flud_time);
	    return 0;
	 }
      }
      else if (!strncasecmp(command, "FLUDBLOCK", 9)) 
      {
	 if (parc > 2) 
	 {
  	    int newval = atoi(parv[2]);
	    if (newval < 0) 
	    {
	       send_me_noticeNA(sptr, ":flud BLOCK must be >= 0");
	       return 0;
	    }
	    flud_block = newval;
	    if (flud_block == 0) 
            {
	       sendto_ops("%s has disabled flud detection/protection",
			  parv[0]);
	       send_me_noticeNA(sptr, ":flud detection disabled");
	    }
	    else {
	       sendto_ops("%s has changed flud BLOCK to %i",
			  parv[0], flud_block);
	       send_me_notice(sptr, ":flud BLOCK is now set to %i", flud_block);
	    }
	    return 0;
	 }
	 else 
         {
	    send_me_notice(sptr, ":flud BLOCK is currently %i", flud_block);
	    return 0;
	 }
      }
#endif
#ifdef NO_CHANOPS_WHEN_SPLIT
      else if (!strncasecmp(command, "SPLITDELAY", 10)) 
      {
	 if (parc > 2) 
	 {
  	    int newval = atoi(parv[2]);
	    if (newval < 0) 
	    {
	       send_me_noticeNA(sptr, ":split delay must be >= 0");
	       return 0;
	    }
	    if (newval > MAX_SERVER_SPLIT_RECOVERY_TIME)
	       newval = MAX_SERVER_SPLIT_RECOVERY_TIME;
	    sendto_ops("%s has changed spam SERVER SPLIT RECOVERY TIME  to %i", parv[0], newval);
	    send_me_notice(sptr, ":SERVER SPLIT RECOVERY TIME is now set to %i", newval);
	    server_split_recovery_time = (newval * 60);
	    return 0;
	 }
	 else 
         {
	    send_me_notice(sptr, ":SERVER SPLIT RECOVERY TIME is currenly %i", server_split_recovery_time / 60);
	 }
      }
#endif
#ifdef ANTI_SPAMBOT
      /* int spam_time = MIN_JOIN_LEAVE_TIME; 
       * int spam_num = MAX_JOIN_LEAVE_COUNT;
       */
      else if (!strncasecmp(command, "SPAMNUM", 7)) 
      {
	 if (parc > 2) 
	 {
   	    int newval = atoi(parv[2]);
	    if (newval <= 0) 
	    {
	       send_me_noticeNA(sptr, ":spam NUM must be > 0");
	       return 0;
	    }
	    if (newval < MIN_SPAM_NUM)
	       spam_num = MIN_SPAM_NUM;
	    else
	       spam_num = newval;
	    sendto_ops("%s has changed spam NUM to %i", parv[0], spam_num);
	    send_me_notice(sptr, ":spam NUM is now set to %i", spam_num);
	    return 0;
	 }
	 else 
         {
	    send_me_notice(sptr, ":spam NUM is currently %i", spam_num);
	    return 0;
	 }
      }
      else if (!strncasecmp(command, "SPAMTIME", 8)) 
      {
	 if (parc > 2) 
	 {
   	    int newval = atoi(parv[2]);
	    if (newval <= 0) 
	    {
	       send_me_noticeNA(sptr, ":spam TIME must be > 0");
	       return 0;
	    }
	    if (newval < MIN_SPAM_TIME)
	       spam_time = MIN_SPAM_TIME;
	    else
	       spam_time = newval;
	    sendto_ops("%s has changed spam TIME to %i", parv[0], spam_time);
	    send_me_notice(sptr, ":SPAM TIME is now set to %i", spam_time);
	    return 0;
	 }
	 else 
         {
	    send_me_notice(sptr, ":spam TIME is currently %i", spam_time);
	    return 0;
	 }
      }

#endif
   }
   else 
   {
      send_me_noticeNA(sptr, ":Options: MAX");
      send_me_noticeNA(sptr, ":Options: THROTTLE <ENABLE|COUNT|TIME|RECORDTIME|HASH> [setting]");
#ifdef FLUD
      send_me_noticeNA(sptr, ":Options: FLUDNUM, FLUDTIME, FLUDBLOCK");
#endif

#ifdef ANTI_SPAMBOT
      send_me_noticeNA(sptr, ":Options: SPAMNUM, SPAMTIME");
#endif

#ifdef NO_CHANOPS_WHEN_SPLIT
      send_me_noticeNA(sptr, ":Options: SPLITDELAY");
#endif

   }
   return 0;
}
/*
 * m_htm - high traffic mode info
 */
int m_htm(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#define LOADCFREQ 5
   char       *command;

   extern int  lifesux, LRV, LCF, noisy_htm;	/* in ircd.c */
   extern int currlife;

   if (!MyClient(sptr) || !IsOper(sptr)) 
   {
      send_me_numericNA(sptr, ERR_NOPRIVILEGES);
      return 0;
   }
   send_me_notice(sptr, ":HTM is %s(%d), %s. Max rate = %dk/s. Current = %dk/s",
	          lifesux ? "ON" : "OFF", lifesux,
	          noisy_htm ? "NOISY" : "QUIET",
	          LRV, currlife);
   if (parc > 1) 
   {
      command = parv[1];
      if (!strcasecmp(command, "TO")) 
      {
	 if (parc > 2) 
	 {
   	    int new_value = atoi(parv[2]);

	    if (new_value < 10) 
	       send_me_noticeNA(sptr, ":\002Cannot set LRV < 10!\002");
	    else
	       LRV = new_value;

	    send_me_notice(sptr, ":NEW Max rate = %dk/s. Current = %dk/s", LRV, currlife);
	    sendto_realops("%s!%s@%s set new HTM rate to %dk/s (%dk/s current)",
			parv[0], sptr->user->username, sptr->sockhost,
			   LRV, currlife);
	 }
	 else
	    send_me_noticeNA(sptr, ":LRV command needs an integer parameter");
      }
      else 
      {
	 if (!strcasecmp(command, "ON")) 
	 {
	    lifesux = 1;
	    send_me_noticeNA(sptr, ":HTM is now ON.");
	    sendto_ops("Entering high-traffic mode: Forced by %s!%s@%s",
		       sptr->name, sptr->user->username, sptr->sockhost);
	    LCF = 30;		/* 30s */
	 }
	 else if (!strcasecmp(command, "OFF")) 
	 {
#ifdef HTM_LOCK_ON_NETBURST
            if (HTMLOCK == YES) 
            {
               if (!IsAdmin(sptr)) 
               {
                 send_me_noticeNA(sptr, ":Cannot change HTM - Currently LOCKED");
                 return 0;
               }
               else 
               {
                 send_me_noticeNA(sptr, ":Removing HTM LOCK");
                 sendto_ops("Resuming standard operation through HTM LOCK: Forced by %s!%s@%s",
                       parv[0], sptr->user->username, sptr->sockhost);
                  lifesux = 0;
                  LCF = LOADCFREQ;
                 return 0;
               }
            }
#endif 

	    lifesux = 0;
	    LCF = LOADCFREQ;
	    send_me_noticeNA(sptr, ":HTM is now OFF.");
	    sendto_ops("Resuming standard operation: Forced by %s!%s@%s",
		       parv[0], sptr->user->username, sptr->sockhost);
	 }
	 else if (!strcasecmp(command, "QUIET")) 
	 {
	    sendto_ops("HTM is now QUIET");
	    noisy_htm = NO;
	 }
	 else if (!strcasecmp(command, "NOISY")) 
	 {
	    sendto_ops("HTM is now NOISY");
	    noisy_htm = YES;
	 }
	 else
	    send_me_noticeNA(sptr, ":Commands are:HTM [ON] [OFF] [TO int] [QUIET] [NOISY]");
      }
   }
   return 0;
}

/*
 * cluster() input            
 * - pointer to a hostname output 
 * pointer to a static of the hostname masked for use in a kline. side 
 * effects - NONE
 * 
 * reworked a tad -Dianora
 */

static char *
cluster(char *hostname)
{
   static char result[HOSTLEN + 1];	/* result to return */
   char        temphost[HOSTLEN + 1];	/* workplace */
   char       *ipp;		/* used to find if host is ip # only */
   char       *host_mask;	/* used to find host mask portion to '*' */
   char       *zap_point = (char *) NULL; /* used to zap last nnn portion of an ip # */
   char       *tld;		/* Top Level Domain */
   int         is_ip_number;	/* flag if its an IP # */
   int         number_of_dots;	/* count # of dots for ip# and domain klines */

   if (!hostname)
      return (char *) NULL;	/* EEK! */

   /*
    * If a '@' is found in the hostname, this is bogus and must have
    * been introduced by server that doesn't check for bogus domains
    * (dns spoof) very well. *sigh* just return it... I could also
    * legitimately return (char *)NULL as above.
    * 
    * -Dianora
    */

   if (strchr(hostname, '@')) 
   {
      strncpyzt(result, hostname, HOSTLEN);
      return (result);
   }

   strncpyzt(temphost, hostname, HOSTLEN);

   is_ip_number = YES;		/* assume its an IP# */
   ipp = temphost;
   number_of_dots = 0;

   while (*ipp) 
   {
      if (*ipp == '.') 
      {
	 number_of_dots++;
	 if (number_of_dots == 3)
	    zap_point = ipp;
	 ipp++;
      }
      else if (!isdigit(*ipp)) 
      {
	 is_ip_number = NO;
	 break;
      }
      ipp++;
   }

   if (is_ip_number && (number_of_dots == 3)) 
   {
      zap_point++;
      *zap_point++ = '*';	/* turn 111.222.333.444 into ... */
      *zap_point = '\0';	/* 111.222.333.* */
      strncpy(result, temphost, HOSTLEN);
      return (result);
   }
   else {
      tld = strrchr(temphost, '.');
      if (tld) 
      {
	 number_of_dots = 2;
	 if (tld[3])	 		/* its at least a 3 letter tld */
	    number_of_dots = 1;
	 if (tld != temphost)		/* in these days of dns spoofers ... */
	    host_mask = tld - 1;	/* Look for host portion to '*' */
	 else
	    host_mask = tld;		/* degenerate case hostname is '.com' ect. */

	 while (host_mask != temphost) 
	 {
	    if (*host_mask == '.')
	       number_of_dots--;
	    if (number_of_dots == 0) 
	    {
	       result[0] = '*';
	       strncpy(result + 1, host_mask, HOSTLEN - 1);
	       return (result);
	    }
	    host_mask--;
	 }
	 result[0] = '*';	/* foo.com => *foo.com */
	 strncpy(result + 1, temphost, HOSTLEN);
      }
      else {		/*  no tld found oops. just return it as is */
	 strncpy(result, temphost, HOSTLEN);
	 return (result);
      }
   }

   return (result);
}

int dbuf_callback(int mem, int memsize)
{
   sendto_ops("%d dbufs have been freed (a total of %d bytes of memory)", memsize);
   return 0;
}

/*
 * * m_rehash *
 * 
 */
int m_rehash(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   if (!OPCanRehash(sptr)) 
   {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
   }
	
   if (parc > 1) 
   {
      if (mycmp(parv[1], "DNS") == 0) 
      {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "DNS");
	flush_cache();		/* flush the dns cache */
	res_init();		/* re-read /etc/resolv.conf */
	sendto_ops("%s is rehashing DNS while whistling innocently", parv[0]);
	return 0;
      }
      else if (mycmp(parv[1], "TKLINES") == 0)
      {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "temp local-bans");
	sendto_ops("%s is clearing temp local-bans while whistling innocently", parv[0]);
        remove_userbans_match_flags(UBAN_LOCAL|UBAN_TEMPORARY, 0);
	return 0;
      }
      else if (mycmp(parv[1], "THROTTLES") == 0)
      {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "throttles");
	throttle_rehash();
	sendto_ops("%s is rehashing throttles", parv[0]);
	return 0;
      }
      else if (mycmp(parv[1], "GC") == 0) 
      {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "garbage collecting");
	block_garbage_collect();
	sendto_ops("%s is garbage collecting while whistling innocently", parv[0]);
	return 0;
      }
      else if (mycmp(parv[1], "MOTD") == 0) 
      {
	sendto_ops("%s is forcing re-reading of MOTD file", parv[0]);
	read_motd(MOTD);
#ifdef SHORT_MOTD
	read_shortmotd(SHORTMOTD);
#endif
	return (0);
      }
      else if (mycmp(parv[1], "SDEBUF") == 0) 
      {
         int doall = 0, ret;

         if(parc > 2 && mycmp(parv[2], "ALL") == 0)
            doall = 1;

         ret = dbuf_startfree_block(1, doall, dbuf_callback);
 	 sendto_ops("%s is trying to free%sserver dbufs while whistling innocently -- %d marked to free",
                    parv[0], doall ? " ALL " : "", ret);
      }
      else if (mycmp(parv[1], "CDEBUF") == 0) 
      {
         int doall = 0, ret;

         if(parc > 2 && mycmp(parv[2], "ALL") == 0)
            doall = 1;

         ret = dbuf_startfree_block(0, doall, dbuf_callback);
	 sendto_ops("%s is trying to free%sclient dbufs while whistling innocently -- %d marked to free", 
                    parv[0], doall ? " ALL " : "", ret);
      }
      else if (mycmp(parv[1], "IP") == 0) 
      {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "ip hash");
	rehash_ip_hash();
	sendto_ops("%s is rehashing iphash while whistling innocently", parv[0]);
      }
      else if(mycmp(parv[1], "AKILLS") == 0) 
      {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "akills");
	sendto_ops("%s is rehashing network-bans", parv[0]);
        remove_userbans_match_flags(UBAN_NETWORK, 0);
	return 0;
       }
   }
   else {
      sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], configfile);
      sendto_ops("%s is rehashing Server config file while whistling innocently", parv[0]);
# ifdef USE_SYSLOG
      syslog(LOG_INFO, "REHASH From %s\n", get_client_name(sptr, FALSE));
# endif
      return rehash(cptr, sptr, (parc > 1) ? ((*parv[1] == 'q') ? 2 : 0) : 0);
   }
   return 0;			/* shouldn't ever get here */
}

/*
 * * m_restart *
 * 
 */
int
m_restart(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char       *pass = NULL;

   if (!OPCanRestart(sptr))
   {
      sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
   }
   /*
    * m_restart is now password protected as in df465 only change --
    * this one doesn't allow a reason to be specified. future changes:
    * crypt()ing of password, reason to be re-added -mjs
    */
   if ((pass = (char *) find_restartpass())) 
   {
      if (parc < 2) 
      {
	 sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "RESTART");
	 return 0;
      }
      if (strcmp(pass, parv[1])) 
      {
	 sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
	 return 0;
      }
   }

#ifdef USE_SYSLOG
   syslog(LOG_WARNING, "Server RESTART by %s\n",
	  get_client_name(sptr, FALSE));
#endif
   sprintf(buf, "Server RESTART by %s", get_client_name(sptr, TRUE));
   restart(buf);
   return 0;			/* NOT REACHED */
}
/*
 * * m_trace 
 *        parv[0] = sender prefix 
 *        parv[1] = servername
 */
int
m_trace(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   int      i;
   aClient *acptr=NULL;
   aClass      *cltmp;
   char        *tname;
   int          doall, link_s[MAXCONNECTIONS + 1], link_u[MAXCONNECTIONS + 1];
   int          cnt = 0, wilds = 0, dow = 0;
	
   tname = (parc > 1) ? parv[1] : me.name;

#ifdef HIDEULINEDSERVS
   if((acptr = next_client_double(client, tname)))
   {
      if (!(IsAnOper(sptr)) && IsULine(acptr))
      {
         sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
         return 0;
      }
      acptr = NULL; /* shrug, we borrowed it, reset it just in case */
   }
#endif

   if (parc > 2)
	if (hunt_server(cptr, sptr, ":%s TRACE %s :%s", 2, parc, parv))
		 return 0;

   switch (hunt_server(cptr, sptr, ":%s TRACE :%s", 1, parc, parv)) {
	 case HUNTED_PASS:	/*  note: gets here only if parv[1] exists */
		  {
			  aClient    *ac2ptr = next_client_double(client, tname);
			  if (ac2ptr)
				sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name, parv[0],
						version, debugmode, tname, 
						ac2ptr->from->name);
			  else
				sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name, parv[0],
						version, debugmode, tname, 
						"ac2ptr_is_NULL!!");
			  return 0;
		  }
	 case HUNTED_ISME:
		break;
	 default:
		return 0;
   }
	if(!IsAnOper(sptr)) 
	{
		if (parv[1] && !strchr(parv[1],'.') && (strchr(parv[1], '*')
			|| strchr(parv[1], '?'))) 
		  /* bzzzt, no wildcard nicks for nonopers */
		  {
			  sendto_one(sptr, rpl_str(RPL_ENDOFTRACE),me.name,
							 parv[0], parv[1]);
			  return 0;        
		  }
	}
	sendto_realops_lev(SPY_LEV, "TRACE requested by %s (%s@%s) [%s]",
			 sptr->name, sptr->user->username, sptr->user->host,
			 sptr->user->server);
	
   doall = (parv[1] && (parc > 1)) ? !match(tname, me.name) : TRUE;
   wilds = !parv[1] || strchr(tname, '*') || strchr(tname, '?');
   dow = wilds || doall;
	if(!IsAnOper(sptr) || !dow) /* non-oper traces must be full nicks */
	  /* lets also do this for opers tracing nicks */
	  {
		  char      *name;
		  int       class;
		  acptr = hash_find_client(tname,(aClient *)NULL);
		  if(!acptr || !IsPerson(acptr)) 
		  {
			  /* this should only be reached if the matching
				target is this server */
			  sendto_one(sptr, rpl_str(RPL_ENDOFTRACE),me.name,
							 parv[0], tname);
			  return 0;
			  
		  }
		  name = get_client_name(acptr,FALSE);
		  class = get_client_class(acptr);
		  if (IsAnOper(acptr)) 
		  {
			  sendto_one(sptr, rpl_str(RPL_TRACEOPERATOR),
					 me.name, parv[0], class, name,
					 MyConnect(acptr) ? timeofday - acptr->lasttime : 0);
		  } else {
			  sendto_one(sptr,rpl_str(RPL_TRACEUSER),
					 me.name, parv[0], class, name,
					 MyConnect(acptr) ? timeofday - acptr->lasttime : 0);
		  }
		  sendto_one(sptr, rpl_str(RPL_ENDOFTRACE),me.name,
						 parv[0], tname);
		  return 0;	 
	  }
   if (dow && lifesux && !IsOper(sptr)) 
   {
      sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
      return 0;
   }
	
   memset((char *) link_s, '\0', sizeof(link_s));
   memset((char *) link_u, '\0', sizeof(link_u));
   /*
    * Count up all the servers and clients in a downlink.
    */
   if (doall)
	  for (acptr = client; acptr; acptr = acptr->next) 
	  {
	      if (IsPerson(acptr) && (!IsInvisible(acptr) || IsAnOper(sptr)))
	      {
		 if(acptr->from->fd == me.fd)
		    link_u[MAXCONNECTIONS]++;
		 else
		    link_u[acptr->from->fd]++;
	      }
	      else if (IsServer(acptr))
	      {
#ifdef HIDEULINEDSERVS
		 if (IsOper(sptr) || !IsULine(acptr))
		 {
#endif
		    if(acptr->from->fd == me.fd)
		       link_s[MAXCONNECTIONS]++;
		    else
		       link_s[acptr->from->fd]++;
#ifdef HIDEULINEDSERVS
		 }
#endif

	      }
	  }
		
	
   /*
    * report all direct connections 
    */
	
   for (i = 0; i <= highest_fd; i++) 
   {
		char       *name;
		int         class;
		
      if (!(acptr = local[i]))	/* Local Connection? */
		  continue;
#ifdef HIDEULINEDSERVS
      if (!IsOper(sptr) && IsULine(acptr))
		continue;
#endif
      if (IsInvisible(acptr) && dow &&
			 !(MyConnect(sptr) && IsAnOper(sptr)) &&
			 !IsAnOper(acptr) && (acptr != sptr))
		  continue;
      if (!doall && wilds && match(tname, acptr->name))
		  continue;
      if (!dow && mycmp(tname, acptr->name))
		  continue;
      if (IsAnOper(sptr))
		  name = get_client_name(acptr, FALSE);
      else
		  name = get_client_name(acptr, HIDEME);
      class = get_client_class(acptr);
		
      switch (acptr->status) 
      {
		 case STAT_CONNECTING:
			sendto_one(sptr, rpl_str(RPL_TRACECONNECTING), me.name,
						  parv[0], class, name);
			cnt++;
			break;
		 case STAT_HANDSHAKE:
			sendto_one(sptr, rpl_str(RPL_TRACEHANDSHAKE), me.name,
						  parv[0], class, name);
			cnt++;
			break;
		 case STAT_ME:
			break;
		 case STAT_UNKNOWN:
			/* added time -Taner */
			sendto_one(sptr, rpl_str(RPL_TRACEUNKNOWN),
				  me.name, parv[0], class, name,
				  acptr->firsttime ? timeofday - acptr->firsttime : -1);
			cnt++;
			break;
		 case STAT_CLIENT:
			/*
			 * Only opers see users if there is a wildcard but
			 * anyone can see all the opers.
			 */
			if (((IsAnOper(sptr) &&
				(MyClient(sptr))) || !(dow && IsInvisible(acptr)))
				 || !dow || IsAnOper(acptr)) 
		        {
				if (IsAnOper(acptr))
				  sendto_one(sptr,
					 rpl_str(RPL_TRACEOPERATOR),
					 me.name, parv[0], class, name,
					 timeofday - acptr->lasttime);
			}
			break;
		 case STAT_SERVER:
			sendto_one(sptr, rpl_str(RPL_TRACESERVER),
							 me.name, parv[0], class, link_s[i],
							 link_u[i], name, 
							 *(acptr->serv->bynick) ? acptr->serv->bynick : "*", 
							 *(acptr->serv->byuser) ? acptr->serv->byuser : "*", 
							 *(acptr->serv->byhost) ? acptr->serv->byhost : me.name);
			cnt++;
			break;
		 case STAT_LOG:
			sendto_one(sptr, rpl_str(RPL_TRACELOG), me.name,
						  parv[0], LOGFILE, acptr->port);
			cnt++;
			break;
		 default:		/*
					 * ...we actually shouldn't come
					 * * here... --msa 
					 */
			sendto_one(sptr, rpl_str(RPL_TRACENEWTYPE), me.name,
						  parv[0], name);
			cnt++;
			break;
      }
   }
   /*
    * Add these lines to summarize the above which can get rather long
    * and messy when done remotely - Avalon
    */
   if (!SendWallops(sptr) || !cnt) 
   {
      if (cnt) 
      {
			sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name,
						  parv[0], tname);
			return 0;
      }
      /* let the user have some idea that its at the end of the trace */
      sendto_one(sptr, rpl_str(RPL_TRACESERVER),
					  me.name, parv[0], 0, link_s[MAXCONNECTIONS],
					  link_u[MAXCONNECTIONS], me.name, "*", "*", me.name,
					  MyConnect(acptr) ? timeofday - acptr->lasttime : 0);
      sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name,
					  parv[0], tname);
      return 0;
   }
#ifdef HIDEULINEDSERVS
   if (IsOper(sptr))
#endif
   for (cltmp = FirstClass(); doall && cltmp; cltmp = NextClass(cltmp))
	  if (Links(cltmp) > 0)
		 sendto_one(sptr, rpl_str(RPL_TRACECLASS), me.name,
				parv[0], Class (cltmp), Links(cltmp));
	
   sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
   return 0;
}
/*
 * * m_motd 
 * 	 parv[0] = sender prefix 
 *       parv[1] = servername
 */
int
m_motd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   static time_t last_used = 0L;

   if (hunt_server(cptr, sptr, ":%s MOTD :%s", 1, parc, parv) != HUNTED_ISME)
		 return 0;

   if(!IsAnOper(sptr)) 
   {
      if (IsSquelch(sptr))
      {
         send_me_numericNA(sptr, RPL_ENDOFMOTD);
         return 0;
      }
      if ((last_used + MOTD_WAIT) > NOW)
         return 0;
      else
         last_used = NOW;
   }

   sendto_realops_lev(SPY_LEV, "MOTD requested by %s (%s@%s) [%s]",
		   sptr->name, sptr->user->username, sptr->user->host,
		      sptr->user->server);
   send_motd(cptr, sptr, parc, parv);
	return 0;
}

/*
 ** send_motd
 **  parv[0] = sender prefix
 **  parv[1] = servername
 **
 ** This function split off so a server notice could be generated on a
 ** user requested motd, but not on each connecting client.
 ** -Dianora
 */
int 
send_motd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aMotd *temp;
   struct tm  *tm;

   tm = motd_tm;
   if (motd == (aMotd *) NULL) 
   {
      sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0]);
      return 0;
   }
   sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name);

   if (tm)
      sendto_one(sptr,
		 ":%s %d %s :- %d/%d/%d %d:%02d", me.name, RPL_MOTD,
	     parv[0], tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year,
		 tm->tm_hour, tm->tm_min);

   temp = motd;
   while (temp) 
   {
      sendto_one(sptr, rpl_str(RPL_MOTD),  me.name, parv[0], temp->line);
      temp = temp->next;
   }
   sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0]);
   return 0;
}
/*
 * read_motd() - From CoMSTuD, added Aug 29, 1996
 */
void read_motd(char *filename)
{
   aMotd *temp, *last;
   struct stat sb;
   char        buffer[MOTDLINELEN], *tmp;
   int         fd;

   /* Clear out the old MOTD */

   while (motd) 
   {
      temp = motd->next;
      MyFree(motd);
      motd = temp;
   }
   fd = open(filename, O_RDONLY);
   if (fd == -1)
      return;
   fstat(fd, &sb);
   motd_tm = localtime(&sb.st_mtime);
   last = (aMotd *) NULL;

   while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) 
   {
      if ((tmp = (char *) strchr(buffer, '\n')))
	 *tmp = '\0';
      if ((tmp = (char *) strchr(buffer, '\r')))
	 *tmp = '\0';
      temp = (aMotd *) MyMalloc(sizeof(aMotd));

      strncpyzt(temp->line, buffer, MOTDLINELEN);
      temp->next = (aMotd *) NULL;
      if (!motd)
	 motd = temp;
      else
	 last->next = temp;
      last = temp;
   }
   close(fd);

   if (motd_tm)
      (void) sprintf(motd_last_changed_date,
		     "%d/%d/%d %d:%02d",
		     motd_tm->tm_mday,
		     motd_tm->tm_mon + 1,
		     1900 + motd_tm->tm_year,
		     motd_tm->tm_hour,
		     motd_tm->tm_min);
}

#ifdef SHORT_MOTD
void read_shortmotd(char *filename)
{
   aMotd *temp, *last;
   char        buffer[MOTDLINELEN], *tmp;
   int         fd;

   /* Clear out the old MOTD */

   while (shortmotd)
   {
      temp = shortmotd->next;
      MyFree(shortmotd);
      shortmotd = temp;
   }
   fd = open(filename, O_RDONLY);
   if (fd == -1)
      return;

   last = (aMotd *) NULL;

   while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) 
   {
      if ((tmp = (char *) strchr(buffer, '\n')))
	 *tmp = '\0';
      if ((tmp = (char *) strchr(buffer, '\r')))
	 *tmp = '\0';
      temp = (aMotd *) MyMalloc(sizeof(aMotd));

      strncpyzt(temp->line, buffer, MOTDLINELEN);
      temp->next = (aMotd *) NULL;
      if (!shortmotd)
	 shortmotd = temp;
      else
	 last->next = temp;
      last = temp;
   }
   close(fd);
}
#endif

/*
 * read_help() - modified from from CoMSTuD's read_motd added Aug 29,
 * 1996 modifed  Aug 31 1997 - Dianora
 * 
 * Use the same idea for the oper helpfile
 */
void
read_help(char *filename)
{
   aMotd *temp, *last;
   char        buffer[MOTDLINELEN], *tmp;
   int         fd;

   /* Clear out the old HELPFILE */

   while (helpfile) 
   {
      temp = helpfile->next;
      MyFree(helpfile);
      helpfile = temp;
   }

   fd = open(filename, O_RDONLY);
   if (fd == -1)
      return;

   last = (aMotd *) NULL;

   while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) 
   {
      if ((tmp = (char *) strchr(buffer, '\n')))
	 *tmp = '\0';
      if ((tmp = (char *) strchr(buffer, '\r')))
	 *tmp = '\0';
      temp = (aMotd *) MyMalloc(sizeof(aMotd));

      strncpyzt(temp->line, buffer, MOTDLINELEN);
      temp->next = (aMotd *) NULL;
      if (!helpfile)
	 helpfile = temp;
      else
	 last->next = temp;
      last = temp;
   }
   close(fd);
}
/*
 * * m_close - added by Darren Reed Jul 13 1992.
 */
int
m_close(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient *acptr;
   int     i;
   int         closed = 0;

   if (!MyOper(sptr)) 
   {
      sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
   }

   for (i = highest_fd; i; i--) 
   {
      if (!(acptr = local[i]))
	 continue;
      if (!IsUnknown(acptr) && !IsConnecting(acptr) &&
	  !IsHandshake(acptr))
	 continue;
      sendto_one(sptr, rpl_str(RPL_CLOSING), me.name, parv[0],
		 get_client_name(acptr, TRUE), acptr->status);
      (void) exit_client(acptr, acptr, acptr, "Oper Closing");
      closed++;
   }
   sendto_one(sptr, rpl_str(RPL_CLOSEEND), me.name, parv[0], closed);
   return 0;
}
int
m_die(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient *acptr;
   int     i;
   char       *pass = NULL;

   if (!OPCanDie(sptr))
   {
      sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
   }
   /* X line -mjs */

   if ((pass = (char *) find_diepass())) 
   {
      if (parc < 2) 
      {
	 sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "DIE");
	 return 0;
      }
      if (strcmp(pass, parv[1])) 
      {
	 sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
	 return 0;
      }
   }

   for (i = 0; i <= highest_fd; i++) 
   {
      if (!(acptr = local[i]))
	 continue;
      if (IsClient(acptr))
	 sendto_one(acptr,
		    ":%s NOTICE %s :Server Terminating. %s",
		    me.name, acptr->name,
		    get_client_name(sptr, FALSE));
      else if (IsServer(acptr))
	 sendto_one(acptr, ":%s ERROR :Terminated by %s",
		    me.name, get_client_name(sptr, TRUE));
   }
   (void) s_die();
   return 0;
}

struct capablist {
   char *name;
   unsigned int flag;
};

struct capablist my_capabs[] = {
   { "TS3", 		CAPAB_TS3 },
   { "NOQUIT", 		CAPAB_NOQUIT },
   { "SSJOIN", 		CAPAB_NSJOIN },
   { "BURST", 		CAPAB_BURST },
   { "UNCONNECT", 	CAPAB_UNCONN },
   { "ZIP", 		CAPAB_ZIP },
#ifdef HAVE_ENCRYPTION_ON
   { "DKEY", 		CAPAB_DKEY },
#endif
   { "IDENTITY", 	CAPAB_IDENT },
   { "NICKIP",		CAPAB_NICKIP },
   { "TOKEN1", 		CAPAB_TOKEN1 },
   { "TSMODE", 		CAPAB_TSMODE },
   { NULL, 0 }
};

void make_capab_string(char *buf, unsigned int noflags)
{  
   int i;
   int c;

   *buf = '\0';

   for(i = c = 0; my_capabs[i].name; i++)
   {
      if(!(my_capabs[i].flag & noflags))
      {
         if(c)
            strcat(buf, " ");
         strcat(buf, my_capabs[i].name);
         c++;
      }
   }
}


/*
 * m_capab 
 * Communicate what I can do to another server 
 * This has to be able to be sent and understood while
 * the client is UNREGISTERED. Therefore, we
 * absolutely positively must not check to see if
 * this is a server or a client. It's probably an unknown!
 */
int 
m_capab(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   int         i, j;

   /* If it's not local, or it has already set capabilities, silently ignore it. */

   if(cptr != sptr || cptr->capabilities)
      return 0;

   for (i = 1; i < parc; i++) 
   {
      for(j = 0; my_capabs[j].name; j++)
      {
         if(strcmp(parv[i], my_capabs[j].name) == 0)
            cptr->capabilities |= my_capabs[j].flag;
      }
   }

   if((IsToken1(cptr) && !IsIDCapable(cptr)) || (!IsToken1(cptr) && IsIDCapable(cptr)))
   { 
      return exit_client(cptr, cptr, &me, "(!(IDENTITY && TOKEN1))");
   }

   if(IsNickIP(cptr) && IsToken1(cptr))
	UnSetNickIP(cptr);

   return 0;
}

/*
 * Shadowfax's LOCKFILE code 
 */
#ifdef LOCKFILE

int
lock_kline_file()
{
int         fd;

   /* Create Lockfile */

   if ((fd = open(LOCKFILE, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) 
   {
      sendto_realops("%s is locked, klines pending", klinefile);
      pending_kline_time = time(NULL);
      return (-1);
   }
   (void) close(fd);
   return (1);
}

void
do_pending_klines()
{
int         fd;
char        s[20];
struct pkl *k, *ok;

   if (!pending_klines)
      return;

   /* Create Lockfile */
   if ((fd = open(LOCKFILE, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) 
   {
      sendto_realops("%s is locked, klines pending", klinefile);
      pending_kline_time = time(NULL);
      return;
   }
   (void) ircsprintf(s, "%d\n", getpid());
   (void) write(fd, s, strlen(s));
   close(fd);

   /* Open klinefile */
   if ((fd = open(klinefile, O_WRONLY | O_APPEND)) == -1) 
   {
      sendto_realops("Pending klines cannot be written, cannot open %s",
		     klinefile);
      unlink(LOCKFILE);
      return;
   }

   /* Add the Pending Klines */

   k = pending_klines;
   while (k) 
   {
      write(fd, k->comment, strlen(k->comment));
      write(fd, k->kline, strlen(k->kline));
      free(k->comment);
      free(k->kline);
      ok = k;
      k = k->next;
      free(ok);
   }
   pending_klines = NULL;
   pending_kline_time = 0;

   close(fd);

   /* Delete the Lockfile */
   unlink(LOCKFILE);
}
#endif
	 
/* m_svskill - Just about the same as outta df
 *  - Raistlin
 * parv[0] = servername
 * parv[1] = client
 * parv[2] = nick stamp
 * parv[3] = kill message
 */
	 
int m_svskill(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
   aClient *acptr;
   char *comment;
   char reason[TOPICLEN + 1];
   ts_val ts = 0;

   if (parc > 3) 
   {
      comment = parv[3] ? parv[3] : parv[0];
      ts = atol(parv[2]);
   }
   else
      comment = (parc > 2 && parv[2]) ? parv[2] : parv[0];
      
   if(!IsULine(sptr)) return -1;
   if((acptr = find_person_target(parv[1])) && (!ts || ts == acptr->tsinfo))
   {
      if(MyClient(acptr))
      {
         strcpy(reason, "SVSKilled: ");
         strncpy(reason + 11, comment, TOPICLEN - 11);
         reason[TOPICLEN] = '\0';
         exit_client(acptr, acptr, sptr, reason);
         return (acptr == cptr) ? CLIENT_DIED : 0;
      }

      if(!ts)
         ts = acptr->tsinfo;

      if(acptr->from == cptr)
      {
         sendto_ops_lev(DEBUG_LEV, "Received wrong-direction SVSKILL for %s (behind %s) from %s",
            acptr->name, cptr->name, get_client_name(sptr, HIDEME));
         return 0;
      }
      else if(ts && !IsIDCapable(acptr->from))
         sendto_one_client_nopostfix(acptr, sptr, &CMD_SVSKILL, "%s %ld :%s", parv[1], ts, comment);
      else
         sendto_one_client_real(acptr->from, acptr, sptr, &CMD_SVSKILL, ":%s", comment);
   }
   return 0;
}
	 
/* m_akill -
 * Parse AKILL command
 * parv[1]=host, parv[2]=user, parv[3]=length, parv[4]=akiller, parv[5]=time set, parv[6]=reason
 */      
int m_akill(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aClient *acptr;
    char *user, *host, *reason, *akiller, buffer[1024], *current_date,
        fbuf[512];
    time_t length=0, timeset=0;
    int i;
    struct userBan *ban, *oban;

    if(!IsServer(sptr) || (parc < 6))
        return 0;

    if(!IsULine(sptr))
    {
        send_serv_message_butone(NULL, &me, NULL, &CMD_GLOBOPS, ":Non-ULined server %s trying to AKILL!",
                                 get_client_name(sptr, HIDEME));
        send_globops("From %s: Non-ULined server %s trying to AKILL!", me.name,
                     get_client_name(sptr, HIDEME));
        return 0;
    }

    host=parv[1];
    user=parv[2];
    akiller=parv[4];
    length=atoi(parv[3]);
    timeset=atoi(parv[5]);
    reason=(parv[6] ? parv[6] : "<no reason>");

    if(length == 0) /* a "permanent" akill? */
       length = (86400 * 7); /* hold it for a week */

    /* is this an old bogus akill? */
    if(timeset + length <= NOW)
       return 0;

    current_date=smalldate((time_t)timeset);
    /* cut reason down a little, eh? */
    /* 250 chars max */
    if(strlen(reason)>250)
        reason[251]=0;

    ban = make_hostbased_ban(user, host);
    if(!ban)
    {
       sendto_realops_lev(DEBUG_LEV, "make_hostbased_ban(%s, %s) failed on akill", user, host);
       return 0;
    }

    /* if it already exists, pass it on */
    oban = find_userban_exact(ban, 0);
    if(oban)
    {
        /* pass along the akill anyways */
       send_serv_message_butone(cptr, sptr, NULL, &CMD_AKILL, "%s %s %d %s %d :%s",
                                host, user, length, akiller, timeset, reason);
       userban_free(ban);
       return 0;
    }

    ircsprintf(buffer, "%s (%s)", reason, current_date);
    ban->flags |= (UBAN_NETWORK|UBAN_TEMPORARY);
    ban->reason = (char *) MyMalloc(strlen(buffer) + 1);
    strcpy(ban->reason, buffer);
    ban->timeset = timeset;
    ban->duration = length;

    add_hostbased_userban(ban);

    /* send it off to any other servers! */
    send_serv_message_butone(cptr, sptr, NULL, &CMD_AKILL, "%s %s %d %s %d :%s",
                             host, user, length, akiller, timeset, reason);

    /* Check local users against it */
    for (i = 0; i <= highest_fd; i++)
    {
        if (!(acptr = local[i]) || IsMe(acptr) || IsLog(acptr))
            continue;
        if (IsPerson(acptr) && user_match_ban(acptr, ban))
        {
            sendto_ops("Network-ban active for %s",
                       get_client_name(acptr, FALSE));
            ircsprintf(fbuf, "Network-banned: %s", reason);
            exit_client(acptr, acptr, &me, fbuf);
            i--;
        }
    }

    return 0;
}

int m_rakill(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    struct userBan *ban, *oban;

    if(!IsServer(sptr))
        return 0;

    /* just quickly find the akill and be rid of it! */
    if(parc<3)
    {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
                   "RAKILL");
        return 0;
    }

    if(!IsULine(sptr))
    {
        send_serv_message_butone(NULL, &me, NULL, &CMD_GLOBOPS, ":Non-ULined server %s trying to RAKILL!",
                                 get_client_name(sptr, HIDEME));
        send_globops("From %s: Non-ULined server %s trying to RAKILL!", me.name,
                     get_client_name(sptr, HIDEME));
        return 0;
    }

    ban = make_hostbased_ban(parv[2], parv[1]);
    if(!ban)
       return 0;

    ban->flags |= UBAN_NETWORK;
    oban = find_userban_exact(ban, UBAN_NETWORK);
    if(oban)
    {
       remove_userban(oban);
       userban_free(oban);
    }

    userban_free(ban);
    send_serv_message_butone(cptr, sptr, NULL, &CMD_RAKILL, "%s %s", parv[1], parv[2]);

    return 0;
}

/*
 * RPL_NOWON   - Online at the moment (Succesfully added to WATCH-list)
 * RPL_NOWOFF  - Offline at the moement (Succesfully added to WATCH-list)
 * RPL_WATCHOFF   - Succesfully removed from WATCH-list.
 * ERR_TOOMANYWATCH - Take a guess :>  Too many WATCH entries.
 */
static void
show_watch(aClient *cptr, char *name, int rpl1, int rpl2) 
{
	aClient *acptr;	
	
	if ((acptr = find_person(name, NULL)))
	  sendto_one(cptr, rpl_str(rpl1), me.name, cptr->name,
					 acptr->name, acptr->user->username,
					 acptr->user->host, MyConnect(acptr) ? acptr->lasttime : 0);
	else
	  sendto_one(cptr, rpl_str(rpl2), me.name, cptr->name,
					 name, "*", "*", 0);
}
	
/*
 * m_watch
 */
int   m_watch(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
	aClient  *acptr;
	char  *s, *p, *user;
	char def[2] = "l";

        if(!IsPerson(sptr)) /* watch is only for people! */
           return 0;
	
	if (parc < 2) 
	{
		/* Default to 'l' - list who's currently online */
		parc = 2;
		parv[1] = def;
	}
	
	for (p = NULL, s = strtoken(&p, parv[1], ", "); s; s = strtoken(&p, NULL, ", ")) 
	{
		if ((user = (char *)strchr(s, '!')))
		  *user++ = '\0'; /* Not used */
		
		/*
		 * Prefix of "+", they want to add a name to their WATCH
		 * list. 
		 */
		if (*s == '+') 
		{
			if (*(s+1)) 
			{
				if (sptr->user->watches >= MAXWATCH) 
				{
					sendto_one(sptr, err_str(ERR_TOOMANYWATCH),
								  me.name, cptr->name, s+1);					
					continue;
				}				
				add_to_watch_hash_table(s+1, sptr);
			}
			show_watch(sptr, s+1, RPL_NOWON, RPL_NOWOFF);
			continue;
		}
		
		/*
		 * Prefix of "-", coward wants to remove somebody from their
		 * WATCH list.  So do it. :-)
		 */
		if (*s == '-') 
		{
			del_from_watch_hash_table(s+1, sptr);
			show_watch(sptr, s+1, RPL_WATCHOFF, RPL_WATCHOFF);
			continue;
		}
					
		/*
		 * Fancy "C" or "c", they want to nuke their WATCH list and start
		 * over, so be it.
		 */
		if (*s == 'C' || *s == 'c') 
		{
			hash_del_watch_list(sptr);
			continue;
		}
		
		/*
		 * Now comes the fun stuff, "S" or "s" returns a status report of
		 * their WATCH list.  I imagine this could be CPU intensive if its
		 * done alot, perhaps an auto-lag on this?
		 */
		if (*s == 'S' || *s == 's') 
		{
			Link *lp;
			aWatch *anptr;
			int  count = 0;
							
			/*
			 * Send a list of how many users they have on their WATCH list
			 * and how many WATCH lists they are on.
			 */
			anptr = hash_get_watch(sptr->name);
			if (anptr)
			  for (lp = anptr->watch, count = 1; (lp = lp->next); count++);
			sendto_one(sptr, rpl_str(RPL_WATCHSTAT), me.name, parv[0],
						  sptr->user->watches, count);
			
			/*
			 * Send a list of everybody in their WATCH list. Be careful
			 * not to buffer overflow.
			 */
			if ((lp = sptr->user->watch) == NULL) 
			{
				sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0],
							  *s);
				continue;
			}
			*buf = '\0';
			strcpy(buf, lp->value.wptr->nick);
			count = strlen(parv[0])+strlen(me.name)+10+strlen(buf);
			while ((lp = lp->next)) 
			{
				if (count+strlen(lp->value.wptr->nick)+1 > BUFSIZE - 2) 
				{
					sendto_one(sptr, rpl_str(RPL_WATCHLIST), me.name,
								  parv[0], buf);
					*buf = '\0';
					count = strlen(parv[0])+strlen(me.name)+10;
				}
				strcat(buf, " ");
				strcat(buf, lp->value.wptr->nick);
				count += (strlen(lp->value.wptr->nick)+1);
			}
			sendto_one(sptr, rpl_str(RPL_WATCHLIST), me.name, parv[0], buf);
			sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0], *s);
			continue;
		}
		
		/*
		 * Well that was fun, NOT.  Now they want a list of everybody in
		 * their WATCH list AND if they are online or offline? Sheesh,
		 * greedy arn't we?
		 */
		if (*s == 'L' || *s == 'l') 
		{
			Link *lp = sptr->user->watch;
			
			while (lp) 
			{
				if ((acptr = find_person(lp->value.wptr->nick, NULL)))
				  sendto_one(sptr, rpl_str(RPL_NOWON), me.name, parv[0],
								 acptr->name, acptr->user->username,
								 acptr->user->host, acptr->tsinfo);
				/*
				 * But actually, only show them offline if its a capital
				 * 'L' (full list wanted).
				 */
				else if (isupper(*s))
				  sendto_one(sptr, rpl_str(RPL_NOWOFF), me.name, parv[0],
								 lp->value.wptr->nick, "*", "*",
								 lp->value.wptr->lasttime);
				lp = lp->next;
			}
			
			sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0],  *s);
			continue;
		}
		/* Hmm.. unknown prefix character.. Ignore it. :-) */
	}
	
	return 0;
}


#define DKEY_GOTIN  0x01
#define DKEY_GOTOUT 0x02

#define DKEY_DONE(x) (((x) & (DKEY_GOTIN|DKEY_GOTOUT)) == (DKEY_GOTIN|DKEY_GOTOUT))

int m_dkey(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   if(!(IsNegoServer(sptr) && parc > 1))
   {
      if(IsPerson(sptr))
         return 0;
      return exit_client(sptr, sptr, sptr, "Not negotiating now");
   }
#ifdef HAVE_ENCRYPTION_ON
   if(mycmp(parv[1], "START") == 0)
   {
      char keybuf[1024];

      if(parc != 2)
         return exit_client(sptr, sptr, sptr, "DKEY START failure");

      if(sptr->serv->sessioninfo_in != NULL && sptr->serv->sessioninfo_out != NULL)
         return exit_client(sptr, sptr, sptr, "DKEY START duplicate?!");

      sptr->serv->sessioninfo_in = dh_start_session();
      sptr->serv->sessioninfo_out = dh_start_session();

      sendto_realops("Initiating diffie-hellman key exchange with %s", sptr->name);

      dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_in);
      sendto_one(sptr, "DKEY PUB I %s", keybuf);

      dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_out);
      sendto_one(sptr, "DKEY PUB O %s", keybuf);
      return 0;
   }

   if(mycmp(parv[1], "PUB") == 0)
   {
      char keybuf[1024];
      int keylen;

      if(parc != 4 || !sptr->serv->sessioninfo_in || !sptr->serv->sessioninfo_out)
         return exit_client(sptr, sptr, sptr, "DKEY PUB failure");

      if(mycmp(parv[2], "O") == 0) /* their out is my in! */
      {
         if(!dh_generate_shared(sptr->serv->sessioninfo_in, parv[3]))
            return exit_client(sptr, sptr, sptr, "DKEY PUB O invalid");
         sptr->serv->dkey_flags |= DKEY_GOTOUT;
      }
      else if(mycmp(parv[2], "I") == 0) /* their out is my in! */
      {
         if(!dh_generate_shared(sptr->serv->sessioninfo_out, parv[3]))
            return exit_client(sptr, sptr, sptr, "DKEY PUB I invalid");
         sptr->serv->dkey_flags |= DKEY_GOTIN;
      }
      else
         return exit_client(sptr, sptr, sptr, "DKEY PUB bad option");

      if(DKEY_DONE(sptr->serv->dkey_flags))
      {

         sendto_one(sptr, "DKEY DONE");
         SetRC4OUT(sptr);

         keylen = 1024;
         if(!dh_get_s_shared(keybuf, &keylen, sptr->serv->sessioninfo_in))
            return exit_client(sptr, sptr, sptr, "Could not setup encrypted session");
         sptr->serv->rc4_in = rc4_initstate(keybuf, keylen);

         keylen = 1024;
         if(!dh_get_s_shared(keybuf, &keylen, sptr->serv->sessioninfo_out))
            return exit_client(sptr, sptr, sptr, "Could not setup encrypted session");
         sptr->serv->rc4_out = rc4_initstate(keybuf, keylen);

         dh_end_session(sptr->serv->sessioninfo_in);
         dh_end_session(sptr->serv->sessioninfo_out);

         sptr->serv->sessioninfo_in = sptr->serv->sessioninfo_out = NULL;
         return 0;
      }

      return 0;
   }

   if(mycmp(parv[1], "DONE") == 0)
   {
      if(!((sptr->serv->sessioninfo_in == NULL && sptr->serv->sessioninfo_out == NULL) &&
           (sptr->serv->rc4_in != NULL && sptr->serv->rc4_out != NULL)))
         return exit_client(sptr, sptr, sptr, "DKEY DONE when not done!");
      SetRC4IN(sptr);
      sendto_realops("Diffie-Hellman exchange with %s complete, connection encrypted.", sptr->name);
      sendto_one(sptr, "DKEY EXIT");
      return RC4_NEXT_BUFFER;
   }

   if(mycmp(parv[1], "EXIT") == 0)
   {
      if(!(IsRC4IN(sptr) && IsRC4OUT(sptr)))
         return exit_client(sptr, sptr, sptr, "DKEY EXIT when not in proper stage");
      ClearNegoServer(sptr);
      return do_server_estab(sptr);
   }
#endif

   return 0;
}

int m_kline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    struct userBan *ban, *oban;
#if defined (LOCKFILE)
    struct pkl *k;
#else
    int         out;
#endif

    char        buffer[1024];

    char       *filename;       /* filename to use for kline */
    char       *user, *host;
    char       *reason;
    char       *current_date;
    aClient    *acptr;
    char        tempuser[USERLEN + 2];
    char        temphost[HOSTLEN + 1];
    int         temporary_kline_time = 0;       /* -Dianora */
    time_t      temporary_kline_time_seconds = 0;
    int         time_specified = 0;
    char       *argv;
    int         i;
    char       fbuf[512];

    if (!MyClient(sptr) || !IsAnOper(sptr) || !OPCanKline(sptr))
    {
        send_me_numericNA(sptr, ERR_NOPRIVILEGES);
        return 0;
    }

    if (parc < 2)
    {
        send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "KLINE");
        return 0;
    }

    argv = parv[1];

    if ((temporary_kline_time = k_isnumber(argv)) >= 0)
    {
        if (parc < 3)
        {
            send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "KLINE");
            return 0;
        }
        if (temporary_kline_time > (24 * 60 * 7))
            temporary_kline_time = (24 * 60 * 7);       /*  Max it at 1 week */

        temporary_kline_time_seconds =
            (time_t) temporary_kline_time *(time_t) 60;

        /* turn it into minutes */
        argv = parv[2];
        parc--;
        time_specified = 1;
    }
    else
    {
        temporary_kline_time = 0; /* -1 minute klines are bad... :) - lucas */
    }

    if(strchr(argv, ' '))
    {
        send_me_noticeNA(sptr,
                   ":Poorly formatted hostname "
                   "(contains spaces). Be sure you are using the form: "
                   "/quote KLINE [time] <user@host/nick> :<reason>");
        return 0;
    }


    if ((host = strchr(argv, '@')) || *argv == '*')
    {
        /* Explicit user@host mask given */

        if (host)               /* Found user@host */
        {
            user = argv;                /* here is user part */
            *(host++) = '\0';   /* and now here is host */
        }
        else
        {
            user = "*";         /* no @ found, assume its *@somehost */
            host = argv;
        }

        if (!*host)             /* duh. no host found, assume its '*' host */
            host = "*";
        strncpyzt(tempuser, user, USERLEN + 2); /* allow for '*' in front */
        strncpyzt(temphost, host, HOSTLEN);
        user = tempuser;
        host = temphost;
    }
    else
    {
        /* Try to find user@host mask from nick */

        if (!(acptr = find_chasing(sptr, argv, NULL, "KLINE")))
            return 0;

        if (!acptr->user)
            return 0;

        if (IsServer(acptr))
        {
            send_me_noticeNA(sptr, ":Can't KLINE a server, use @'s where appropriate");
            return 0;
        }
        /*
         * turn the "user" bit into "*user", blow away '~' if found in
         * original user name (non-idented)
         */

        tempuser[0] = '*';
        if (*acptr->user->username == '~')
            strcpy(tempuser + 1, (char *) acptr->user->username + 1);
        else
            strcpy(tempuser + 1, acptr->user->username);
        user = tempuser;
        host = cluster(acptr->user->host);
    }

    if (time_specified)
        argv = parv[3];
    else
        argv = parv[2];

#ifdef DEFAULT_KLINE_TIME
    if (time_specified == 0)
    {
        temporary_kline_time = DEFAULT_KLINE_TIME;
        temporary_kline_time_seconds =
            (time_t) temporary_kline_time *(time_t) 60;
    }
#endif

    if (parc > 2)
    {
        if (*argv)
            reason = argv;
        else
            reason = "No reason";
    }
    else
        reason = "No reason";

    if (!match(user, "akjhfkahfasfjd") &&
        !match(host, "ldksjfl.kss...kdjfd.jfklsjf"))
    {
        send_me_noticeNA(sptr, ":Can't ban *@*");
        return 0;
    }

    /* we can put whatever we want in temp K: lines */
    if (temporary_kline_time == 0 && strchr(reason, ':'))
    {
        send_me_noticeNA(sptr, ":Invalid character ':' in comment");
        return 0;
    }

    if (temporary_kline_time == 0 && strchr(reason, '#'))
    {
        send_me_noticeNA(sptr, ":Invalid character '#' in comment");
        return 0;
    }

    ban = make_hostbased_ban(user, host);
    if(!ban)
    {
        send_me_notice(sptr, ":Malformed ban %s@%s", user, host);
        return 0;
    }

    if ((oban = find_userban_exact(ban, 0)))
    {
        char *ktype = (oban->flags & UBAN_LOCAL) ? "local-banned" : "network-banned";

        send_me_notice(sptr, ":[%s@%s] already %s for %s",
                       user, host, ktype, oban->reason ? oban->reason : "<No Reason>");

        userban_free(ban);
        return 0;
    }

    current_date = smalldate((time_t) 0);
    ircsprintf(buffer, "%s (%s)", reason, current_date);

    ban->flags |= UBAN_LOCAL;
    ban->reason = (char *) MyMalloc(strlen(buffer) + 1);
    strcpy(ban->reason, buffer);

    if (temporary_kline_time)
    {
        ban->flags |= UBAN_TEMPORARY;
        ban->timeset = timeofday;
        ban->duration = temporary_kline_time_seconds;
    }

    if(user_match_ban(sptr, ban))
    {
        send_me_notice(sptr, ":You attempted to add a ban [%s@%s] which would affect yourself. Aborted.",
                       user, host);
        userban_free(ban);
        return 0;
    }

    add_hostbased_userban(ban);

    /* Check local users against it */
    for (i = 0; i <= highest_fd; i++)
    {
        if (!(acptr = local[i]) || IsMe(acptr) || IsLog(acptr))
            continue;

        if (IsPerson(acptr) && user_match_ban(acptr, ban))
        {
            sendto_ops("Local-ban active for %s",
                       get_client_name(acptr, FALSE));
            ircsprintf(fbuf, "Local-banned: %s", reason);
            exit_client(acptr, acptr, &me, fbuf);
            i--;
        }
    }

    host = get_userban_host(ban, fbuf, 512);

    if(temporary_kline_time)
    {
        sendto_realops("%s added temporary %d min. local-ban for [%s@%s] [%s]",
                       parv[0], temporary_kline_time, user, host, reason);
        return 0;
    }

    filename = klinefile;

#ifdef KPATH
    send_me_notice(sptr, ":Added local-ban [%s@%s] to server klinefile", user, host);
#else
    send_me_notice(sptr, ":Added local-ban [%s@%s] to server configfile", user, host);
#endif /* KPATH */

    sendto_realops("%s added local-ban for [%s@%s] [%s]",
                   parv[0], user, host, reason);

#if defined(LOCKFILE)
    if ((k = (struct pkl *) malloc(sizeof(struct pkl))) == NULL)
    {
        send_me_noticeNA(sptr, ":Problem allocating memory");
        return (0);
    }

    (void) ircsprintf(buffer, "#%s!%s@%s K'd: %s@%s:%s\n",
                      sptr->name, sptr->user->username,
                      sptr->user->host, user, host,
                      reason);

    if ((k->comment = strdup(buffer)) == NULL)
    {
        free(k);
        send_me_noticeNA(sptr, ":Problem allocating memory");
        return (0);
    }

    ircsprintf(buffer, "K:%s:%s (%s):%s\n",
               host,
               reason,
               current_date,
               user);

    if ((k->kline = strdup(buffer)) == NULL)
    {
        free(k->comment);
        free(k);
        send_me_noticeNA(sptr, ":Problem allocating memory");
        return (0);
    }
    k->next = pending_klines;
    pending_klines = k;

    do_pending_klines();
    return (0);

#else /*  LOCKFILE - MDP */

    if ((out = open(filename, O_RDWR | O_APPEND | O_CREAT)) == -1)
    {
        send_me_notice(sptr, ":Problem opening %s", filename);
        return 0;
    }

    ircsprintf(buffer, "#%s!%s@%s K'd: %s@%s:%s\n",
               sptr->name, sptr->user->username,
               sptr->user->host, user, host,
               reason);

    if (write(out, buffer, strlen(buffer)) <= 0)
    {
        send_me_notice(sptr, ":Problem writing to %s", filename);
        close(out);
        return 0;
    }

    ircsprintf(buffer, "K:%s:%s (%s):%s\n", host, reason,
               current_date, user);

    if (write(out, buffer, strlen(buffer)) <= 0)
    {
        send_me_notice(sptr, ":Problem writing to %s", filename);
        close(out);
        return 0;
    }

    close(out);

#ifdef USE_SYSLOG
    syslog(LOG_NOTICE, "%s added local-ban for [%s@%s] [%s]", parv[0],
           user, host, reason);
#endif

    return 0;
#endif /* LOCKFILE */
}

#ifdef UNKLINE
/*
 * m_unkline
 *
 *      parv[0] = sender
 *      parv[1] = address to remove
 *
 */
int m_unkline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    struct userBan *ban;
    int         in, out;
    int         pairme = NO;
    char        buf[256], buff[256];
    char        temppath[256];
    char       *filename;       /* filename to use for unkline */
    char       *user, *host;
    char       *p;
    int         nread;
    int         error_on_write = NO;
    struct stat oldfilestat;
    mode_t      oldumask;

    ircsprintf(temppath, "%s.tmp", klinefile);

    if (!IsAnOper(sptr) || !OPCanUnKline(sptr))
    {
        send_me_numericNA(sptr, ERR_NOPRIVILEGES);
        return 0;
    }

    if (parc < 2)
    {
        send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "UNKLINE");
        return 0;
    }

    if ((host = strchr(parv[1], '@')) || *parv[1] == '*') {
        /* Explicit user@host mask given */

        if (host)       /* Found user@host */
        {
            user = parv[1];     /* here is user part */
            *(host++) = '\0';   /* and now here is host */
        }
        else
        {
            user = "*";         /* no @ found, assume its *@somehost */
            host = parv[1];
        }
    }
    else
    {
        send_me_noticeNA(sptr, ":Invalid parameters");
        return 0;
    }

    if ((user[0] == '*') && (user[1] == '\0') && (host[0] == '*') &&
        (host[1] == '\0'))
    {
        send_me_noticeNA(sptr, ":Cannot unban everyone");
        return 0;
    }

    ban = make_hostbased_ban(user, host);
    if(ban)
    {
        struct userBan *oban;

        ban->flags |= (UBAN_LOCAL|UBAN_TEMPORARY);
        if((oban = find_userban_exact(ban, UBAN_LOCAL|UBAN_TEMPORARY)))
        {
            char tmp[512];

            host = get_userban_host(oban, tmp, 512);

            remove_userban(oban);
            userban_free(oban);
            userban_free(ban);

            send_me_notice(sptr, ":Temporary local-ban for [%s@%s] is removed", user, host);
            sendto_ops("%s has removed the local-ban for: [%s@%s] (%d matches)",
                       parv[0], user, host, 1);

            return 0;
        }

        userban_free(ban);
    }

# if defined(LOCKFILE)
    if (lock_kline_file() < 0)
    {
        send_me_notice(sptr, ":%s is locked. Try again in a few minutes", klinefile);
        return 0;
    }
# endif

    filename = klinefile;

    if ((in = open(filename, O_RDONLY)) == -1)
    {
        send_me_notice(sptr, ":Cannot open %s", filename);
# if defined(LOCKFILE)
        unlink(LOCKFILE);
# endif
        return 0;
    }

    if (fstat(in, &oldfilestat) < 0)    /*  Save the old file mode */
        oldfilestat.st_mode = 0644;

    oldumask = umask(0);                /* ircd is normally too paranoid */

    if ((out = open(temppath, O_WRONLY | O_CREAT, oldfilestat.st_mode)) == -1)
    {
        send_me_notice(sptr, ":Cannot open %s", temppath);
        close(in);
# if defined (LOCKFILE)
        unlink(LOCKFILE);
# endif
        umask(oldumask);                /* Restore the old umask */
        return 0;
    }
    umask(oldumask);            /* Restore the old umask */

    /*
     * #Dianora!db@ts2-11.ottawa.net K'd: foo@bar:No reason K:bar:No
     * reason (1997/08/30 14.56):foo
     */

    remove_userbans_match_flags(UBAN_LOCAL, UBAN_TEMPORARY);

    while ((nread = dgets(in, buf, sizeof(buf))) > 0)
    {
        buf[nread] = '\0';

        if ((buf[1] == ':') && ((buf[0] == 'k') || (buf[0] == 'K')))
        {
            /* its a K: line */

            char       *found_host;
            char       *found_user;
            char       *found_comment;

            strcpy(buff, buf);

            p = strchr(buff, '\n');
            if (p)
                *p = '\0';
            p = strchr(buff, '\r');
            if (p)
                *p = '\0';

            found_host = buff + 2;      /* point past the K: */

            p = strchr(found_host, ':');
            if (p == (char *) NULL)
            {
                send_me_noticeNA(sptr, ":local-ban file corrupted, can't find 'host'");
                if (!error_on_write)
                    error_on_write = flush_write(sptr,
                                                 out, buf, strlen(buf),
                                                 temppath);
                continue;               /* This K line is corrupted ignore */
            }
            *p = '\0';
            p++;

            found_comment = p;
            p = strchr(found_comment, ':');
            if (p == (char *) NULL)
            {
                send_me_noticeNA(sptr, ":local-ban file corrupted, can't find 'comment'");
                if (!error_on_write)
                    error_on_write = flush_write(sptr,
                                                 out, buf, strlen(buf),
                                                 temppath);
                continue;               /* This K line is corrupted ignore */
            }
            *p = '\0';
            p++;

            found_user = p;
            /*
             * Ok, if its not an exact match on either the user or the
             * host then, write the K: line out, and I add it back to the
             * K line tree
             */
            if (strcasecmp(host, found_host) || strcasecmp(user, found_user))
            {
                struct userBan *tban;
                char *ub_u, *ub_r;

                if (!error_on_write)
                    error_on_write = flush_write(sptr,
                                                 out, buf, strlen(buf),
                                                 temppath);

                if(BadPtr(found_host))
                    continue;

                ub_u = BadPtr(found_user) ? "*" : found_user;
                ub_r = BadPtr(found_comment) ? "<No Reason>" : found_comment;

                if(!(tban = make_hostbased_ban(ub_u, found_host)))
                    continue;

                tban->flags |= UBAN_LOCAL;
                tban->reason = (char *) MyMalloc(strlen(ub_r) + 1);
                strcpy(tban->reason, ub_r);
                tban->timeset = NOW;

                add_hostbased_userban(tban);
            }
            else
                pairme++;
        }
        else if (buf[0] == '#')
        {
            char       *userathost;
            char       *found_user;
            char       *found_host;

            strcpy(buff, buf);
            /*
             * #Dianora!db@ts2-11.ottawa.net K'd: foo@bar:No reason
             * K:bar:No reason (1997/08/30 14.56):foo
             *
             * If its a comment coment line, i.e. #ignore this line Then just
             * ignore the line
             */
            p = strchr(buff, ':');
            if (p == (char *) NULL)
            {
                if (!error_on_write)
                    error_on_write = flush_write(sptr,
                                                 out, buf, strlen(buf),
                                                 temppath);
                continue;
            }
            *p = '\0';
            p++;

            userathost = p;
            p = strchr(userathost, ':');

            if (p == (char *) NULL)
            {
                if (!error_on_write)
                    error_on_write = flush_write(sptr,
                                                 out, buf, strlen(buf),
                                                 temppath);
                continue;
            }
            *p = '\0';

            while (*userathost == ' ')
                userathost++;

            found_user = userathost;
            p = strchr(found_user, '@');
            if (p == (char *) NULL)
            {
                if (!error_on_write)
                    error_on_write = flush_write(sptr, 
                                                 out, buf, strlen(buf),
                                                 temppath);
                continue;
            }
            *p = '\0';
            found_host = p;
            found_host++;

            if ((strcasecmp(found_host, host)) ||
                (strcasecmp(found_user, user)))
            {
                if (!error_on_write)
                    error_on_write = flush_write(sptr, 
                                                 out, buf, strlen(buf),
                                                 temppath);
            }
        }
        else
        {   /* its the ircd.conf file, and not a Kline or comment */
            if (!error_on_write)
                error_on_write = flush_write(sptr, 
                                             out, buf, strlen(buf), temppath);
        }
    }

    close(in);

    /* The result of the rename should be checked too... oh well */
    /*
     * If there was an error on a write above, then its been reported
     * and I am not going to trash the original kline /conf file
     * -Dianora
     */
    if ((!error_on_write) && (close(out) >= 0))
        rename(temppath, filename);
    else {
        send_me_noticeNA(sptr, ":Couldn't write temp kline file, aborted.");
# if defined (LOCKFILE)
        unlink(LOCKFILE);
# endif
        return 0;
    }

# if defined (LOCKFILE)
    unlink(LOCKFILE);
# endif

    if (pairme == NO)
    {
        send_me_notice(sptr, ":%s NOTICE %s :No local-ban for %s@%s", user, host);
        return 0;
    }

    send_me_notice(sptr, ":local-ban for [%s@%s] is removed", user, host);
    sendto_ops("%s has removed the local-ban for: [%s@%s] (%d matches)",
               parv[0], user, host, pairme);
    return 0;
}

/*
 * flush_write()
 *
 * inputs
 * - pointer to client structure of oper requesting unkline
 * - out is the file descriptor
 * - buf is the buffer to write
 * - ntowrite is the expected number of character to be written
 * - temppath is the temporary file name to be written output
 *   YES for error on write
 *   NO for success side effects
 * - if successful, the buf is written to output file if a write failure
 *   happesn, and the file pointed to by temppath, if its non NULL, is
 *   removed.
 *
 * The idea here is, to be as robust as possible when writing to the kline
 * file.
 *
 * Yes, I could have dug the opernick out of sptr. I didn't feel like it.
 * so sue me.
 *
 * -Dianora
 */
static int flush_write(aClient *sptr, int out, char *buf,
                       int ntowrite, char *temppath)
{
    int         nwritten;
    int         error_on_write = NO;

    nwritten = write(out, buf, ntowrite);
    if (nwritten != ntowrite)
    {
        send_me_notice(sptr, ":Unable to write to %s", temppath);
        error_on_write = YES;
        (void) close(out);
        if (temppath != (char *) NULL)
            (void) unlink(temppath);
    }
    return (error_on_write);
}
#endif /* UNKLINE */


/*
 * k_isnumber()
 *
 * inputs
 * - pointer to ascii string in output
 * - 0 if not an integer number, else the number side effects
 * - none return -1 if not an integer.
 * (if someone types in maxint, oh well..) - lucas
 */

static int k_isnumber(char *p)
{
    int         result = 0;

    while (*p)
    {
        if (isdigit(*p))
        {
            result *= 10;
            result += ((*p) & 0xF);
            p++;
        }
        else
            return (-1);
    }
    /*
     * in the degenerate case where oper does a /quote kline 0 user@host
     * :reason i.e. they specifically use 0, I am going to return 1
     * instead as a return value of non-zero is used to flag it as a
     * temporary kline
     */

    /*
     * er, no. we return 0 because 0 means that it's a permanent kline. -lucas
     * oh, and we only do this if DEFAULT_KLINE_TIME is specified.
     */

#ifndef DEFAULT_KLINE_TIME
    if(result == 0)
        result = 1;
#endif

    return (result);
}

