/************************************************************************
 *   IRC - Internet Relay Chat, src/send.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *		      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: send.c,v 1.86 2002/10/04 20:36:08 lucas Exp $ */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "msg.h"
#include <stdio.h>
#include "numeric.h"
#include "zlink.h"
#include "dh.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

static char sendservbuf[4096];
static char sendbuf[2048];
static char remotebuf[2048];
/* 
 * WARNING
 * rc4buf in send.c must be the same size as zipOutBuf in zlink.c
 */
static char rc4buf[16384];
static int sentalong[MAXCONNECTIONS];
static int sent_serial;

static void sendto_fdlist_preformatted(fdlist *listp, char *buf, int len);
static int send_message(aClient *, char *, int);

void init_send()
{
   memset(sentalong, 0, sizeof(int) * MAXCONNECTIONS);
   sent_serial = 0;
}

/* This routine increments our serial number so it will
 * be unique from anything in sentalong, no need for a memset
 * except for every MAXINT calls - lucas
 */

/* This should work on any OS where an int is 32 bit, I hope.. */

#define HIGHEST_SERIAL INT_MAX

#define INC_SERIAL() if(sent_serial == HIGHEST_SERIAL) \
   { memset(sentalong, 0, sizeof(sentalong)); sent_serial = 0; } \
   sent_serial++;


/*
 * dead_link
 *
 * somewhere along the lines of sending out, there was an error.
 * we can't close it from the send loop, so mark it as dead
 * and close it from the main loop.
 *
 * if this link is a server, tell routing people.
 */

static int dead_link(aClient *to, char *notice) 
{
   int errtmp = errno;  /* so we don't munge this later */

   to->flags |= FLAGS_DEADSOCKET;
   /*
    * If because of BUFFERPOOL problem then clean dbuf's now so that
    * notices don't hurt operators below.
    */
   DBufClear(&to->recvQ);
   DBufClear(&to->sendQ);
   /*
    * Ok, if the link we're dropping is a server, send a routing
    * notice..
    */
   if (IsServer(to) && !(to->flags & FLAGS_CLOSING))
   {
      char fbuf[512];

      ircsprintf(fbuf, "from %s: %s", me.name, notice);
      sendto_gnotice(fbuf, get_client_name(to, HIDEME), strerror(errtmp));
      ircsprintf(fbuf, ":%s GNOTICE :%s", me.name, notice);
      sendto_serv_butone(to, fbuf, get_client_name(to, HIDEME), strerror(errtmp));
   }  
 
   return -1;
}

/*
 * send_message 
 * Internal utility which delivers one message buffer to the socket. 
 * Takes care of the error handling and buffering, if needed.
 */
static int send_message(aClient *to, char *msg, int len) 
{
   static int  SQinK;
   int forserver;
   
#ifdef DUMP_DEBUG
   fprintf(dumpfp, "-> %s: %s"\n, (to->name ? to->name : "*"), msg);
#endif
   if (to->from)
     to = to->from;   /* shouldn't be necessary */

   if(IsServer(to) || IsNegoServer(to))
   {
      forserver = 1;
      if(len > 510) 
      {
         msg[511]='\n';
         msg[512]='\0';
         len = 512;
      }
      else 
      {
         msg[len++] = '\n';
         msg[len] = '\0';
      }
   }
   else
   {
      forserver = 0;
      if(len > 509) 
      {
         msg[510]='\r';
         msg[511]='\n';
         msg[512]='\0';
         len = 512;
      }
      else 
      {
         msg[len++] = '\r';
         msg[len++] = '\n';
         msg[len] = '\0';
      }
   }

   if (IsMe(to)) 
   {
      sendto_ops("Trying to send to myself! [%s]", msg);
      return 0;
   }
   
   if (IsDead(to))
     return 0;

   if (DBufLength(&to->sendQ) > to->sendqlen) 
   {
      /* this would be a duplicate notice, but it contains some useful information that
         would be spamming the rest of the network. Kept in. - lucas */

      if (IsServer(to)) 
	sendto_ops("Max SendQ limit exceeded for %s: %d > %d",
		get_client_name(to, HIDEME), DBufLength(&to->sendQ), to->sendqlen);
      to->flags |= FLAGS_SENDQEX;
      return dead_link(to, "Max Sendq exceeded for %s, closing link");
   }

   /*
    * Update statistics. The following is slightly incorrect
    * because it counts messages even if queued, but bytes only
    * really sent. Queued bytes get updated in SendQueued.
    */
   to->sendM++;
   me.sendM++;
   if(to->lstn)
      to->lstn->sendM++;

   if(forserver) /* don't try any of this for clients */
   {
      if(ZipOut(to))
      {
         int ldata = (to->serv->burstflags & BFLAGS_SENDBURST); 
         /* ldata == 1 if we know data is coming in large quantities */

         msg = zip_output(to->serv->zip_out, msg, &len, 0, &ldata);

         if(len == -1) /* error */
         {
            sendto_realops("Zipout error for %s: (%d) %s", to->name, ldata, msg);
            return dead_link(to, "Zipout error for %s");
         }

         if(len == 0)
            return 0;
      }

#ifdef HAVE_ENCRYPTION_ON
      if(IsRC4OUT(to))
      {
         rc4_process_stream_to_buf(to->serv->rc4_out, msg, rc4buf, len);
         msg = rc4buf;
      }
#endif
   }   

   if (dbuf_put(&to->sendQ, msg, len, forserver) < 0)
     return dead_link(to, "Buffer allocation error for %s, closing link");

   /*
    * This little bit is to stop the sendQ from growing too large
    * when there is no need for it to. Thus we call send_queued()
    * every time 2k has been added to the queue since the last
    * non-fatal write. Also stops us from deliberately building a
    * large sendQ and then trying to flood that link with data
    * (possible during the net relinking done by servers with a large
    * load).
    */

   if (to->flags & FLAGS_BLOCKED)
      return 0;

#ifdef ALWAYS_SEND_DURING_SPLIT
   if (currently_processing_netsplit)
   {
      send_queued(to);
      return 0;
   }
#endif

   SQinK = (DBufLength(&to->sendQ) >> 10);
   if (IsServer(to)) 
   {
      if ((to->serv->burstflags & BFLAGS_SENDBURST))
      {
        if(SQinK > (to->lastsq + 32))
           send_queued(to); /* during a burst, send_queued every 32k (1/2 size of a server dbuf) */
      }
      else if (SQinK > (to->lastsq + 8)) /* every 8k for servers.. */
	send_queued(to);
   }
   else 
   {
      if (SQinK > (to->lastsq + 4)) /* every 4k for clients.. */
	send_queued(to);
   }
   return 0;
}


/*
 * send_queued 
 * This function is called from the main select-loop (or whatever) 
 * when there is a chance the some output would be possible. This 
 * attempts to empty the send queue as far as possible...
 */
int send_queued(aClient *to) 
{
   static char clisendbuf[4096];
   char *msg;
   int len, rlen, more_data = 0;
   int forserver = IsServer(to);
	
   /*
    * Once socket is marked dead, we cannot start writing to it,
    * even if the error is removed...
    */

   if (IsDead(to)) 
   {
      /*
       * Actually, we should *NEVER* get here--something is not
       * working correct if send_queued is called for a dead
       * socket... --msa
       */

      sendto_realops_lev(DEBUG_LEV, "Trying to send_queued() to dead client %s[%d][%s]", to->name, to->fd, to->from->name); 

      return -1;
   }

   if(forserver && ZipOut(to) && zip_is_data_out(to->serv->zip_out))
   {
      if(DBufLength(&to->sendQ))
         more_data = 1;
      else
      {
         int ldata = (to->serv->burstflags & BFLAGS_SENDBURST);

         msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);

         if(len == -1)
         {
            sendto_realops("Zipout error for %s: (%d) %s", to->name, ldata, msg);
            return dead_link(to, "Zip output error for %s");
         }

         if(len == 0) /* this probably shouldn't happen, but let's sanity check anyway.. */
            return 0;

#ifdef HAVE_ENCRYPTION_ON
         if(IsRC4OUT(to))
         {
            rc4_process_stream_to_buf(to->serv->rc4_out, msg, rc4buf, len);
            msg = rc4buf;
         }         
#endif
         if(!dbuf_put(&to->sendQ, msg, len, forserver))
            return dead_link(to, "Buffer allocation error for %s");
      }
   }

   while (DBufLength(&to->sendQ) > 0) 
   {
      if(forserver)
         msg = dbuf_map(&to->sendQ, &len); /* always returns >= 0 */
      else
      {
         len = dbuf_copy_map(&to->sendQ, clisendbuf, 4096);
         msg = clisendbuf;
      }
      if ((rlen = deliver_it(to, msg, len)) < 0)
	return dead_link(to, "Write error to %s, closing link (%s)");
      dbuf_delete(&to->sendQ, rlen);
      to->lastsq = (DBufLength(&to->sendQ) >> 10);
      if (rlen < len) /* could not write full buffer.. break out of send queue loop */
	break;

      if(more_data && DBufLength(&to->sendQ) == 0)
      {
         int ldata = (to->serv->burstflags & BFLAGS_SENDBURST);

         more_data = 0;

         msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);

         if(len == -1)
         {
            sendto_realops("Zipout error for %s: (%d) %s", to->name, ldata, msg);
            return dead_link(to, "Zip output error for %s");
         }

         if(len == 0) /* this probably shouldn't happen, but let's sanity check anyway.. */
            return 0;

#ifdef HAVE_ENCRYPTION_ON
         if(IsRC4OUT(to))
         {
            rc4_process_stream_to_buf(to->serv->rc4_out, msg, rc4buf, len);
            msg = rc4buf;
         }         
#endif
         if(!dbuf_put(&to->sendQ, msg, len, forserver))
            return dead_link(to, "Buffer allocation error for %s");
      }
   }

   if(forserver && (to->serv->burstflags & BFLAGS_SOBSENT) && IsBurst(to) && DBufLength(&to->sendQ) < 20480) /* 20k */
   {
      if (!(to->serv->burstflags & BFLAGS_BURST))
      {
         to->serv->burstflags &= (~BFLAGS_SOBSENT);
         sendto_one(to, "BURST %d", DBufLength(&to->sendQ));
#ifdef HTM_LOCK_ON_NETBURST
         if (!(to->serv->burstflags & BFLAGS_EOBRECV)) /* hey we're the last to synch.. */
            HTMLOCK = NO;
#endif
     }
  }
   
  return (IsDead(to)) ? -1 : 0;
}

/*
 * BEGIN SEND_* FUNCTIONS
 *
 * A lot of these functions have the same code internally.
 * I suppose that's okay, it just makes it long and ugly.
 *
 * - join_message_buffer
 * - join_prefix_message_buffer
 * these inline functions take a message of known length (prepared by ircsprintf)
 * to create a buffer to send off.
 *
 * - join_message_buffer_copy
 * - join_prefix_message_buffer_copy
 * these inline functions take a message of UNKNOWN length
 * to create a buffer to send off.
 *
 * - sendto_one_client (pre & postfix, "from X to ..." send to to)
 * - sendto_one_client_real ("from X to ..." sent to realto)
 * - sendto_one_client_nopostfix ("from X ..." send to to)
 * - sendto_one_client_nopostfix_copy ("from X string" send to to)
 * - sendto_one_client_prefixed ("from!user@host X to ..." send to to)
 * - sendto_one_client_prefixed_copy ("from!user@host X to string" send to to)
 * - sendto_one_clientonserver ("from X user@server ..." send to to)
 * - sendto_client_numeric ("from 123 to [numeric format] ..." send to to)
 * - send_serv_message_butone ("from X to ..." send to all server butone)
 * - send_channel_message_butone ("from!user@host X #channel ..." send to all on channel butone)
 * - send_channel_message_butone_copy ("from!user@host X #channel string" send to all on channel butone)
 * - send_channelflags_message_butone ( .. to all on channel with flags 'flags' butone)
 * - send_channelflags_message_butone_copy ( .. to all on channel with flags 'flags' butone)
 * - send_channel_message_localremote_butone ( .. X #channel ..." to all on channel locally butone/remotely)
 * - send_channel_message_localremote_butone_copy ( .. X #channel text" to all on channel locally butone/remotely)
 * note that all of the send_*_copy functions prefix the argument with a :, all others DO NOT.
 * this is an exception for _nopostfix_copy!
 */

/*
 * join_message_buffer
 *
 * params:
 * from:	the sender of the message or NULL
 * fromprefix:	prefix char to put in front of 'from' (required of from != NULL)
 * to:		the recipient of the message or NULL
 * toprefix:	prefix char to put in front of 'to' (can be 0 for no prefix)
 * command:	text of the command
 * msg:		preformatted message
 * msglen:	pointer to length of preformatted message, overwritten with total length of message
 *
 * returns:
 * [fromprefix]from COMMAND [toprefix]to msg
 */

static inline char *join_message_buffer(char *from, char fromprefix, char *to, char toprefix, char *command, 
                                        char msgprefix, char *msg, int *msglen)
{
   int pfixlen = 0;
   int fulllen = *msglen;

   if(from)
   {
      sendservbuf[pfixlen++] = fromprefix;
      while(*from)
      {
         sendservbuf[pfixlen++] = *from;
         from++;
      }
      sendservbuf[pfixlen++] = ' ';
   }

   while(*command)
   {
      sendservbuf[pfixlen++] = *command;
      command++;
   }
   sendservbuf[pfixlen++] = ' ';

   if(to)
   {
      if(toprefix)
         sendservbuf[pfixlen++] = toprefix;
      while(*to)
      {
         sendservbuf[pfixlen++] = *to;
         to++;
      }
      sendservbuf[pfixlen++] = ' ';
   }

   if(msgprefix)
      sendservbuf[pfixlen++] = msgprefix;

   if(fulllen)
      memcpy(&sendservbuf[pfixlen], msg, fulllen);
   else
   {
      if(sendservbuf[pfixlen - 1] == ' ')
         pfixlen--;
      sendservbuf[pfixlen] = '\0';
   }

   *msglen = fulllen + pfixlen;
   return sendservbuf;
}

static inline char *join_message_buffer_copy(char *from, char fromprefix, char *to, char toprefix, char *command, 
                                             char msgprefix, char *msg, int *msglen)
{
   int pfixlen = 0;

   if(from)
   {
      sendservbuf[pfixlen++] = fromprefix;
      while(*from)
      {
         sendservbuf[pfixlen++] = *from;
         from++;
      }
      sendservbuf[pfixlen++] = ' ';
   }

   while(*command)
   {
      sendservbuf[pfixlen++] = *command;
      command++;
   }
   sendservbuf[pfixlen++] = ' ';

   if(to)
   {
      if(toprefix)
         sendservbuf[pfixlen++] = toprefix;
      while(*to)
      {
         sendservbuf[pfixlen++] = *to;
         to++;
      }
   }

   if(msg && *msg)
   {
      if(sendservbuf[pfixlen - 1] != ' ')
         sendservbuf[pfixlen++] = ' ';
      if(msgprefix)
         sendservbuf[pfixlen++] = msgprefix;
      while(*msg)
         sendservbuf[pfixlen++] = *msg++;
   }

   *msglen = pfixlen;
   return sendservbuf;
}

static inline char *join_prefix_message_buffer(char *from, char *fromuser, char *fromhost, char fromprefix, 
             char *to, char toprefix, char *command, char msgprefix, char *msg, int *msglen)
{
   int pfixlen = 0;
   int fulllen = *msglen;

   if(from)
   {
      sendservbuf[pfixlen++] = fromprefix;
      while(*from)
      {
         sendservbuf[pfixlen++] = *from;
         from++;
      }
      if(fromuser)
      {
         sendservbuf[pfixlen++] = '!';
         while(*fromuser)      
         {
            sendservbuf[pfixlen++] = *fromuser;
            fromuser++;
         }
      }
      if(fromhost)
      {
         sendservbuf[pfixlen++] = '@';
         while(*fromhost)      
         {
            sendservbuf[pfixlen++] = *fromhost;
            fromhost++;
         }
      }
      sendservbuf[pfixlen++] = ' ';
   }

   while(*command)
   {
      sendservbuf[pfixlen++] = *command;
      command++;
   }
   sendservbuf[pfixlen++] = ' ';

   if(to)
   {
      if(toprefix)
         sendservbuf[pfixlen++] = toprefix;
      while(*to)
      {
         sendservbuf[pfixlen++] = *to;
         to++;
      }
      sendservbuf[pfixlen++] = ' ';
   }

   if(msgprefix)
      sendservbuf[pfixlen++] = msgprefix;

   if(fulllen)
      memcpy(&sendservbuf[pfixlen], msg, fulllen);
   else
   {
      if(sendservbuf[pfixlen - 1] == ' ')
         pfixlen--;
      sendservbuf[pfixlen] = '\0';
   }

   *msglen = fulllen + pfixlen;
   return sendservbuf;
}

static inline char *join_prefix_message_buffer_copy(char *from, char *fromuser, char *fromhost, char fromprefix, 
             char *to, char toprefix, char *command, char msgprefix, char *msg, int *msglen)
{
   int pfixlen = 0;

   if(from)
   {
      sendservbuf[pfixlen++] = fromprefix;
      while(*from)
      {
         sendservbuf[pfixlen++] = *from;
         from++;
      }
      if(fromuser)
      {
         sendservbuf[pfixlen++] = '!';
         while(*fromuser)      
         {
            sendservbuf[pfixlen++] = *fromuser;
            fromuser++;
         }
      }
      if(fromhost)
      {
         sendservbuf[pfixlen++] = '@';
         while(*fromhost)      
         {
            sendservbuf[pfixlen++] = *fromhost;
            fromhost++;
         }
      }
      sendservbuf[pfixlen++] = ' ';
   }

   while(*command)
   {
      sendservbuf[pfixlen++] = *command;
      command++;
   }
   sendservbuf[pfixlen++] = ' ';

   if(to)
   {
      if(toprefix)
         sendservbuf[pfixlen++] = toprefix;
      while(*to)
      {
         sendservbuf[pfixlen++] = *to;
         to++;
      }
   }

   if(msg && *msg)
   {
      if(sendservbuf[pfixlen - 1] != ' ')
         sendservbuf[pfixlen++] = ' ';
      if(msgprefix)
         sendservbuf[pfixlen++] = msgprefix;
      while(*msg)
         sendservbuf[pfixlen++] = *msg++;
   }

   *msglen = pfixlen;
   return sendservbuf;
}

/*
 * sendto_one_client
 *
 * smarter replacement for sendto_one()
 * can send to local or remote servers.
 */

void sendto_one_client(aClient *to, aClient *from, struct msg_pointer *message, char *pattern, ...) 
{
   va_list vl;
   int msglen;
   aClient *realto = to->from ? to->from : to;

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer(from_id, fromprefix, to_id, toprefix, message->token, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is not id capable, or this is a local client of mine! */
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                      to ? to->name : NULL, 0, 
                                      message->message, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
}

/*
 * sendto_one_client_real
 *
 * smarter replacement for sendto_one()
 * can send to local or remote servers.
 * prefixes and postfixes to and from, but actually sends this message to 'realto'
 */

void sendto_one_client_real(aClient *realto, aClient *to, aClient *from, struct msg_pointer *message, char *pattern, ...) 
{
   va_list vl;
   int msglen;

   realto = realto->from ? realto->from : realto;

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer(from_id, fromprefix, to_id, toprefix, message->token, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is not id capable, or this is a local client of mine! */
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                      to ? to->name : NULL, 0, 
                                      message->message, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
}

/*
 * sendto_one_client_nopostfix[_copy]
 *
 * smarter replacement for sendto_one()
 * can send to local or remote servers.
 * donsn't stick 'to' after command like sendto_one_client
 */

void sendto_one_client_nopostfix(aClient *to, aClient *from, struct msg_pointer *message, char *pattern, ...) 
{
   va_list vl;
   int msglen;
   aClient *realto = to->from ? to->from : to;

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer(from_id, fromprefix, NULL, 0, message->token, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is not id capable, or this is a local client of mine! */
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                      NULL, 0, 
                                      message->message, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
}

void sendto_one_client_nopostfix_copy(aClient *to, aClient *from, struct msg_pointer
                                      *message, int doprefix, char *mtxt) 
{
   aClient *realto = to->from ? to->from : to;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char fromprefix = 0;
      int tlen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer_copy(from_id, fromprefix, NULL, 0, message->token, doprefix,
                                     mtxt, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is not id capable, or this is a local client of mine! */
   {
      int tlen;
      char *txt = join_message_buffer_copy(from ? from->name : NULL, ':', 
                                      NULL, 0, 
                                      message->message, doprefix, mtxt, &tlen);

      send_message(realto, txt, tlen);
   }
}

/*
 * sendto_one_client_prefixed
 *
 * smarter replacement for sendto_prefix_one
 * can send to local or remote servers.
 */

void sendto_one_client_prefixed(aClient *to, aClient *from, struct msg_pointer *message, char *pattern, ...) 
{
   va_list vl;
   int msglen;
   aClient *realto = to->from ? to->from : to;

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer(from_id, fromprefix, to_id, toprefix, message->token, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
   else if(IsAServer(realto)) /* realto is not id capable */
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                      to ? to->name : NULL, 0, 
                                      message->message, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is a local client of mine! */
   {
      int tlen = msglen;
      char *txt = join_prefix_message_buffer(from->name, from->user ? from->user->username : NULL, 
                                      from->user ? *from->user->host ? from->user->host : from->sockhost : NULL, ':', 
                                      to ? to->name : NULL, 0, message->message, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
}

/*
 * sendto_one_client_prefixed_copy
 *
 * smarter replacement for sendto_prefix_one
 * can send to local or remote servers.
 * saves time by not mucking with varags/ircsprintf
 */

void sendto_one_client_prefixed_copy(aClient *to, aClient *from, struct msg_pointer *message,
                                     int doprefix, char *pmsg) 
{
   aClient *realto = to->from ? to->from : to;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = 0;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer_copy(from_id, fromprefix, to_id, toprefix, message->token,
                                     doprefix, pmsg, &tlen);

      send_message(realto, txt, tlen);
   }
   else if(IsAServer(realto)) /* realto is not id capable */
   {
      int tlen = 0;
      char *txt = join_message_buffer_copy(from ? from->name : NULL, ':', 
                                      to ? to->name : NULL, 0, 
                                      message->message, doprefix, pmsg, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is a local client of mine! */
   {
      int tlen = 0;
      char *txt = join_prefix_message_buffer_copy(from->name, from->user
                      ? from->user->username : NULL, 
                      from->user ? *from->user->host ? from->user->host : from->sockhost : NULL, 
                      ':', to ? to->name : NULL, 0, message->message, doprefix, pmsg, &tlen);

      send_message(realto, txt, tlen);
   }
}

/*
 * sendto_one_clientonserver
 *
 * sends to nick@server if remote (and not identity)
 * will be phased out once identity is global (maybe?)
 */

void sendto_one_clientonserver(aClient *to, aClient *from, struct msg_pointer *message, char *pattern, ...) 
{
   va_list vl;
   int msglen;
   aClient *realto = to->from ? to->from : to;

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer(from_id, fromprefix, to_id, toprefix, message->token, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is not id capable, or this is a local client of mine! */
   {
      int tlen = msglen;
      char nickserver[HOSTLEN + NICKLEN + 3];
      char *txt;

      ircsprintf(nickserver, "%s@%s", to->name, to->uplink->name);

      txt = join_message_buffer(from ? from->name : NULL, ':', 
                                nickserver, 0, 
                                message->message, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
}

void sendto_one_clientonserver_copy(aClient *to, aClient *from, struct msg_pointer *message, int doprefix, char *pmsg) 
{
   aClient *realto = to->from ? to->from : to;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer_copy(from_id, fromprefix, to_id, toprefix, message->token, doprefix, pmsg, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is not id capable, or this is a local client of mine! */
   {
      int tlen;
      char nickserver[HOSTLEN + NICKLEN + 3];
      char *txt;

      ircsprintf(nickserver, "%s@%s", to->name, to->uplink->name);

      txt = join_message_buffer_copy(from ? from->name : NULL, ':', 
                                nickserver, 0, 
                                message->message, doprefix, pmsg, &tlen);

      send_message(realto, txt, tlen);
   }
}

/* 
 * sendto_client_numeric
 *
 * smart function for sending numerics out to clients.
 * if to and *to->name is \0, nick will be replaced with *
 * this is for numerics in pre-connect phase
 */

void sendto_client_numeric(aClient *to, aClient *from, int numeric, ...) 
{
   va_list vl;
   int msglen;
   aClient *realto = to->from ? to->from : to;
   char *pattern = get_numeric_format(&numeric);
   char numerictext[16];

   make_numeric_string(numerictext, numeric);

   if(*pattern)
   {
      va_start(vl, numeric);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   if(IsAServer(realto) && IsIDCapable(realto) && IsToken1(realto))
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer(from_id, fromprefix, to_id, toprefix, numerictext, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
   else /* realto is not id capable, or local client */
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                      to ? (*to->name ? to->name : "*") : NULL, 0, 
                                      numerictext, 0, sendbuf, &tlen);

      send_message(realto, txt, tlen);
   }
}

/* prefix_buffer
 *
 * take varargs and dump prefixed message into a buffer
 * remote: 1 if client is remote, 0 if local
 * from: the client sending the message
 * prefix: the prefix as specified (parv[0] usually)
 * buffer: the buffer to dump this into (NO BOUNDS CHECKING!)
 * pattern: varargs pattern
 * vl: varargs variable list with one arg taken already
 */
static inline int prefix_buffer(int remote, aClient *from, char *prefix, char *buffer, char *pattern, va_list vl)
{
   char *p;      /* temp pointer */
   int msglen;   /* the length of the message we end up with */
   int sidx = 1; /* start at offset 1 */

   *buffer = ':';

   if(!remote && IsPerson(from))
   {
      int flag = 0;
      anUser *user = from->user;

      for(p = from->name; *p; p++)
         buffer[sidx++] = *p;

      if (user)
      {
         if (*user->username) 
         {
            buffer[sidx++] = '!';
            for(p = user->username; *p; p++)
               buffer[sidx++] = *p;
	 }
         if (*user->host && !MyConnect(from)) 
         {
            buffer[sidx++] = '@';
            for(p = user->host; *p; p++)
               buffer[sidx++] = *p;
            flag = 1;
         }
      }
   
      if (!flag && MyConnect(from) && *user->host) 
      {
         buffer[sidx++] = '@';
         for(p = from->sockhost; *p; p++)
            buffer[sidx++] = *p;
      }
   }
   else
   {
      for(p = prefix; *p; p++)
         buffer[sidx++] = *p;
   }

   msglen = ircvsprintf(&buffer[sidx], pattern + 3, vl);
   msglen += sidx;

   return msglen;
}

static inline int check_fake_direction(aClient *from, aClient *to)
{
   if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) 
   {
      if (IsServer(from)) 
      {
         sendto_ops("Message to %s[%s] dropped from %s (Fake Direction)",
                      to->name, to->from->name, from->name);
         return -1;
      }

      sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->user->username, 
                 to->user->host, from->name, from->user->username, from->user->host, to->from->name);
      sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name, 
                         me.name, to->name, to->user->username, to->user->host, to->from->name);

      to->flags |= FLAGS_KILLED;
      exit_client(NULL, to, &me, "Ghosted client");

      if (IsPerson(from))
         sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name, 
                    to->user->username, to->user->host, to->from);
      return -1;
   }

   return 0;
}

/*
 * send a prefixed, tokenized message to another server.
 * care must be taken to call this function correctly.
 * "to" can be NULL.
 */
void send_serv_message_butone(aClient *one, aClient *from, aClient *to, struct msg_pointer *message, char *pattern, ...)
{
   aClient *cptr;
   va_list vl;
   int msglen, i, j;
   /* so as to not make 20 fdlists here, id servers must be token compatible as well ... */
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) 
   {
      if (!(cptr = local[i]) || (one && cptr == one->from))
         continue;
      if(IsIDCapable(cptr) && IsToken1(cptr))
         id_fdlist.entry[++id_cnt] = i;
      else
         normal_fdlist.entry[++normal_cnt] = i;
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
      return;

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;
   
   if(id_cnt)
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer(from_id, fromprefix, to_id, toprefix, message->token, 0, sendbuf, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                      to ? to->name : NULL, 0, 
                                      message->message, 0, sendbuf, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }
}

/*
 * send a prefixed, tokenized message to another server.
 * care must be taken to call this function correctly.
 * "to" can be NULL.
 */
void send_flag_serv_message_butone(aClient *one, int include, int exclude,
                                   aClient *from, aClient *to, struct msg_pointer *message, 
                                   char *pattern, ...)
{
   aClient *cptr;
   va_list vl;
   int msglen, i, j;
   /* so as to not make 20 fdlists here, id servers must be token compatible as well ... */
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) 
   {
      if (!(cptr = local[i]) || (one && cptr == one->from) ||
	  (include && !((cptr->capabilities & include) == include)) ||
	  (exclude && (cptr->capabilities & exclude)))
         continue;
      if(IsIDCapable(cptr) && IsToken1(cptr))
         id_fdlist.entry[++id_cnt] = i;
      else
         normal_fdlist.entry[++normal_cnt] = i;
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
      return;

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;
   
   if(id_cnt)
   {
      char *from_id = NULL;
      char *to_id = NULL;
      char toprefix = 0, fromprefix = 0;
      int tlen = msglen;
      char *txt;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      if(to)
      {
         if(HasID(to))
         {
            to_id = to->id.string;
            toprefix = '!';
         }
         else
         {
            to_id = to->name;
            toprefix = 0;
         }
      }

      txt = join_message_buffer(from_id, fromprefix, to_id, toprefix, message->token, 0, sendbuf, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                      to ? to->name : NULL, 0, 
                                      message->message, 0, sendbuf, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }
}

/* ALL PARAMETERS MUST BE FILLED except 'one' */
/* send_channel_message_butone[_copy] 
 * sends a message to everyone on channel chptr, local or remote
 * where 'one' does not get the message
 * 'from' sent the message
 * 'chname' is the actual channel name
 * and mtxt is the text of the message
 */
void send_channel_message_butone(aClient *one, aClient *from, aChannel *chptr, char *chname, 
                                 struct msg_pointer *msg, char *pattern, ...) 
{
   chanMember *cm;
   aClient *acptr;
   int i;
   va_list vl;
   int didlocal, msglen;
   char *localtext = NULL;
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   INC_SERIAL();

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   for (cm = chptr->members; cm; cm = cm->nextuser) 
   {
      acptr = cm->cptr;
      if (acptr->from == one)
         continue; /* ...was the one I should skip */
      i = acptr->from->fd;
      if (MyClient(acptr)) 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if(!localtext)
         {
            didlocal = msglen;
            localtext = join_prefix_message_buffer(from->name, from->user ? from->user->username : NULL, 
                                      from->user ? from->user->host : NULL, ':', 
                                      chname, 0, msg->message, ':', sendbuf, &didlocal);
         }

         send_message(acptr, localtext, didlocal);
         sentalong[i] = sent_serial;
      }
      else 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if (sentalong[i] != sent_serial) 
         {
            sentalong[i] = sent_serial;

            if(IsIDCapable(acptr->from) && IsToken1(acptr->from))
               id_fdlist.entry[++id_cnt] = i;
            else
               normal_fdlist.entry[++normal_cnt] = i;
         }
      }
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
   {
      return;
   }

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(id_cnt)
   {
      char *from_id = NULL, *txt;
      char fromprefix = 0;
      int tlen = msglen;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer(from_id, fromprefix, chname, 0, msg->token, ':', sendbuf, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                chname, 0, msg->message, ':', sendbuf, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }

   return;
}

void send_channel_message_butone_copy(aClient *one, aClient *from, aChannel *chptr, char *chname, 
                                 struct msg_pointer *msg, int doprefix, char *mtxt) 
{
   chanMember *cm;
   aClient *acptr;
   int i;
   int didlocal;
   char *localtext = NULL;
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) 
   {
      acptr = cm->cptr;
      if (acptr->from == one)
         continue; /* ...was the one I should skip */
      i = acptr->from->fd;
      if (MyClient(acptr)) 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if(!localtext)
            localtext = join_prefix_message_buffer_copy(from->name, from->user ? from->user->username : NULL, 
                                      from->user ? from->user->host : NULL, ':', 
                                      chname, 0, msg->message, doprefix, mtxt, &didlocal);

         send_message(acptr, localtext, didlocal);
         sentalong[i] = sent_serial;
      }
      else 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if (sentalong[i] != sent_serial) 
         {
            sentalong[i] = sent_serial;

            if(IsIDCapable(acptr->from) && IsToken1(acptr->from))
               id_fdlist.entry[++id_cnt] = i;
            else
               normal_fdlist.entry[++normal_cnt] = i;
         }
      }
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
   {
      return;
   }

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(id_cnt)
   {
      char *from_id = NULL, *txt;
      char fromprefix = 0;
      int tlen;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer_copy(from_id, fromprefix, chname, 0, msg->token, doprefix,
                                     mtxt, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen;
      char *txt = join_message_buffer_copy(from ? from->name : NULL, ':', 
                                chname, 0, msg->message, doprefix, mtxt, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }

   return;
}

/* send_channel_message_localremote_butone[_copy]
 * sends a message to everyone on channel chptr, local or remote
 * where 'one' does not get the message
 * 'from' sent the message
 * 'chname' is the actual channel name
 * and pattern, ... is the text of the message
 */
void send_channel_message_localremote_butone(int dolocal, aClient *one, aClient *from, aChannel *chptr, char *chname, 
     struct msg_pointer *msg, char *pattern, ...) 
{
   chanMember *cm;
   aClient *acptr;
   va_list vl;
   int i;
   int didlocal, msglen;
   char *localtext = NULL;
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   INC_SERIAL();

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   for (cm = chptr->members; cm; cm = cm->nextuser) 
   {
      acptr = cm->cptr;
      if (acptr->from == one)
         continue; /* ...was the one I should skip */
      i = acptr->from->fd;
      if (dolocal && MyClient(acptr)) 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if(!localtext)
         {
            didlocal = msglen;
            localtext = join_prefix_message_buffer(from->name, from->user ? from->user->username : NULL, 
                                      from->user ? from->user->host : NULL, ':', 
                                      chname, 0, msg->message, 0, sendbuf, &didlocal);
         }

         send_message(acptr, localtext, didlocal);
         sentalong[i] = sent_serial;
      }
      else if(!dolocal)
      {
         if(check_fake_direction(from, acptr))
            continue;

         if (sentalong[i] != sent_serial) 
         {
            sentalong[i] = sent_serial;

            if(IsIDCapable(acptr->from) && IsToken1(acptr->from))
               id_fdlist.entry[++id_cnt] = i;
            else
               normal_fdlist.entry[++normal_cnt] = i;
         }
      }
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
   {
      return;
   }

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(id_cnt)
   {
      char *from_id = NULL, *txt;
      char fromprefix = 0;
      int tlen = msglen;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer(from_id, fromprefix, chname, 0, msg->token, 0, sendbuf, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                chname, 0, msg->message, 0, sendbuf, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }

   return;
}

void send_channel_message_localremote_butone_copy(int dolocal, aClient *one, aClient *from, aChannel *chptr, char *chname, 
     struct msg_pointer *msg, int doprefix, char *mtxt) 
{
   chanMember *cm;
   aClient *acptr;
   int i;
   int didlocal;
   char *localtext = NULL;
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) 
   {
      acptr = cm->cptr;
      if (acptr->from == one)
         continue; /* ...was the one I should skip */
      i = acptr->from->fd;
      if (dolocal && MyClient(acptr)) 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if(!localtext)
         {
            localtext = join_prefix_message_buffer_copy(from->name, from->user ? from->user->username : NULL, 
                                      from->user ? from->user->host : NULL, ':', 
                                      chname, 0, msg->message, doprefix, mtxt, &didlocal);
         }

         send_message(acptr, localtext, didlocal);
         sentalong[i] = sent_serial;
      }
      else if(!dolocal)
      {
         if(check_fake_direction(from, acptr))
            continue;

         if (sentalong[i] != sent_serial) 
         {
            sentalong[i] = sent_serial;

            if(IsIDCapable(acptr->from) && IsToken1(acptr->from))
               id_fdlist.entry[++id_cnt] = i;
            else
               normal_fdlist.entry[++normal_cnt] = i;
         }
      }
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
   {
      return;
   }

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(id_cnt)
   {
      char *from_id = NULL, *txt;
      char fromprefix = 0;
      int tlen;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer_copy(from_id, fromprefix, chname, 0, msg->token, doprefix,
            mtxt, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen;
      char *txt = join_message_buffer_copy(from ? from->name : NULL, ':', 
                                chname, 0, msg->message, doprefix, mtxt, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }

   return;
}

/* same as above, but only sends to clients with channelflags 'flags' -- EXACTLY */
void send_channelflags_message_butone_copy(aClient *one, aClient *from, aChannel *chptr, 
                                           int flags, char *chname, struct msg_pointer *msg, 
                                           int doprefix, char *mtxt) 
{
   chanMember *cm;
   aClient *acptr;
   int i;
   int didlocal;
   char *localtext = NULL;
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) 
   {
      if((cm->flags & flags) == 0)
         continue;

      acptr = cm->cptr;
      if (acptr->from == one)
         continue; /* ...was the one I should skip */

      i = acptr->from->fd;
      if (MyClient(acptr)) 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if(!localtext)
            localtext = join_prefix_message_buffer_copy(from->name, from->user ? from->user->username : NULL, 
                                      from->user ? from->user->host : NULL, ':', 
                                      chname, 0, msg->message, doprefix, mtxt, &didlocal);

         send_message(acptr, localtext, didlocal);
         sentalong[i] = sent_serial;
      }
      else 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if (sentalong[i] != sent_serial) 
         {
            sentalong[i] = sent_serial;

            if(IsIDCapable(acptr->from) && IsToken1(acptr->from))
               id_fdlist.entry[++id_cnt] = i;
            else
               normal_fdlist.entry[++normal_cnt] = i;
         }
      }
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
   {
      return;
   }

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(id_cnt)
   {
      char *from_id = NULL, *txt;
      char fromprefix = 0;
      int tlen;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer_copy(from_id, fromprefix, chname, 0, msg->token, doprefix,
            mtxt, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen;
      char *txt = join_message_buffer_copy(from ? from->name : NULL, ':', 
                                chname, 0, msg->message, doprefix, mtxt, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }

   return;
}

/* same as above, but only sends to clients with channelflags 'flags' -- EXACTLY */
void send_channelflags_message_butone(aClient *one, aClient *from, aChannel *chptr, 
                                           int flags, char *chname, struct msg_pointer *msg, char *pattern, ...) 
{
   chanMember *cm;
   aClient *acptr;
   int i, msglen;
   va_list vl;
   int didlocal;
   char *localtext = NULL;
   fdlist id_fdlist, normal_fdlist;
   int id_cnt = 0, normal_cnt = 0;

   INC_SERIAL();

   if(*pattern)
   {
      va_start(vl, pattern);
      msglen = ircvsprintf(sendbuf, pattern, vl);
      va_end(vl);
   }
   else
      msglen = 0;

   for (cm = chptr->members; cm; cm = cm->nextuser) 
   {
      if((cm->flags & flags) == 0)
         continue;

      acptr = cm->cptr;
      if (acptr->from == one)
         continue; /* ...was the one I should skip */

      i = acptr->from->fd;
      if (MyClient(acptr)) 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if(!localtext)
         {
            didlocal = msglen;
            localtext = join_prefix_message_buffer(from->name, from->user ? from->user->username : NULL, 
                                      from->user ? from->user->host : NULL, ':', 
                                      chname, 0, msg->message, ':', sendbuf, &didlocal);
         }

         send_message(acptr, localtext, didlocal);
         sentalong[i] = sent_serial;
      }
      else 
      {
         if(check_fake_direction(from, acptr))
            continue;

         if (sentalong[i] != sent_serial) 
         {
            sentalong[i] = sent_serial;

            if(IsIDCapable(acptr->from) && IsToken1(acptr->from))
               id_fdlist.entry[++id_cnt] = i;
            else
               normal_fdlist.entry[++normal_cnt] = i;
         }
      }
   }

   if(!(id_cnt || normal_cnt)) /* don't bother if we have no servers.. */
   {
      return;
   }

   id_fdlist.last_entry = id_cnt;
   normal_fdlist.last_entry = normal_cnt;

   if(id_cnt)
   {
      char *from_id = NULL, *txt;
      char fromprefix = 0;
      int tlen = msglen;

      if(from)
      {
         if(HasID(from))
         {
            from_id = from->id.string;
            fromprefix = '!';
         }
         else
         {
            from_id = from->name;
            fromprefix = ':';
         }
      }

      txt = join_message_buffer(from_id, fromprefix, chname, 0, msg->token, ':', sendbuf, &tlen);

      sendto_fdlist_preformatted(&id_fdlist, txt, tlen);
   }

   if(normal_cnt)
   {
      int tlen = msglen;
      char *txt = join_message_buffer(from ? from->name : NULL, ':', 
                                chname, 0, msg->message, ':', sendbuf, &tlen);

      sendto_fdlist_preformatted(&normal_fdlist, txt, tlen);
   }

   return;
}

/* END NEW FUNCTIONS */


/*
 * send message to single client
 * should not be used for remote clients!!
 */

void sendto_one(aClient *to, char *pattern, ...) 
{
   va_list vl;
   int len;		/* used for the length of the current message */
   
   va_start(vl, pattern);
   len = ircvsprintf(sendbuf, pattern, vl);
   
   if (to->from)
     to = to->from;

   send_message(to, sendbuf, len);
   va_end(vl);
}

void vsendto_one(aClient *to, char *pattern, va_list vl) {
   int len;   /* used for the length of the current message */
   
   len = ircvsprintf(sendbuf, pattern, vl);
   
   if (to->from)
     to = to->from;

   send_message(to, sendbuf, len);
}

void sendto_channel_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...) 
{
   chanMember *cm;
   aClient *acptr;
   int i;
   int didlocal = 0, didremote = 0;
   va_list vl;
   char *pfix;
   
   va_start(vl, pattern);

   pfix = va_arg(vl, char *);

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) 
   {
      acptr = cm->cptr;
      if (acptr->from == one)
         continue; /* ...was the one I should skip */
      i = acptr->from->fd;
      if (MyClient(acptr)) 
      {
         if(!didlocal)
            didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);

         if(check_fake_direction(from, acptr))
            continue;

         send_message(acptr, sendbuf, didlocal);
         sentalong[i] = sent_serial;
      }
      else 
      {
         /*
          * Now check whether a message has been sent to this remote
          * link already
          */
         if(!didremote)
            didremote = prefix_buffer(1, from, pfix, remotebuf, pattern, vl);

         if(check_fake_direction(from, acptr))
            continue;

         if (sentalong[i] != sent_serial) 
         {
            send_message(acptr, remotebuf, didremote);
            sentalong[i] = sent_serial;
         }
      }
   }

   va_end(vl);
   return;
}

/*
 * sendto_server_butone
 * 
 * Send a message to all connected servers except the client 'one'.
 */
void sendto_serv_butone(aClient *one, char *pattern, ...) {
	int i;
	aClient *cptr;
	int j, k = 0;
	fdlist send_fdlist;
	va_list vl;
	
	va_start(vl, pattern);
   for (i = serv_fdlist.entry[j = 1];
		  j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
      if (!(cptr = local[i]) || (one && cptr == one->from))
		  continue;
      send_fdlist.entry[++k] = i;
   }
   send_fdlist.last_entry = k;
   if (k)
	  vsendto_fdlist(&send_fdlist, pattern, vl);
	va_end(vl);
   return;
}

/*
 * sendto_flag_serv_butone
 * 
 * Send a message to all connected servers except the client 'one'.
 * Only if they include INCLUDE and exclude EXCLUDE
 */
void sendto_flag_serv_butone(aClient *one, int include,
			     int exclude, char *pattern, ...) {
   int i;
   aClient *cptr;
   int j, k = 0;
   fdlist send_fdlist;
   va_list vl;
   
   va_start(vl, pattern);
   for (i = serv_fdlist.entry[j = 1];
	j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
      if (!(cptr = local[i]) ||
	  (include && !((cptr->capabilities & include) == include)) ||
	  (exclude && (cptr->capabilities & exclude)) ||
	  one == cptr)
	continue;
      send_fdlist.entry[++k] = i;
   }
   send_fdlist.last_entry = k;
   if (k)
     vsendto_fdlist(&send_fdlist, pattern, vl);
   va_end(vl);
   return;
}

/*
 * sendto_common_channels()
 * 
 * Sends a message to all people (inclusing user) on local server who are
 * in same channel with user.
 */
void sendto_common_channels(aClient *from, char *pattern, ...)
{
   chanMember *channels, *users;
   aClient *cptr;
   va_list vl;
   char *pfix;
   int msglen = 0;

   va_start(vl, pattern);

   pfix = va_arg(vl, char *);

   INC_SERIAL();

   if(from->fd >= 0)
      sentalong[from->fd] = sent_serial;

   if (from->user)
   {
      for (channels = from->user->channel; channels; channels = channels->nextchan)
      {
         for (users = channels->chptr->members; users; users = users->nextuser) 
         {
            cptr = users->cptr;

            if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial)
               continue;

            sentalong[cptr->fd] = sent_serial;
            if(!msglen)
               msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
            if(check_fake_direction(from, cptr))
               continue;
            send_message(cptr, sendbuf, msglen);
         }
      }
   }

   if(MyConnect(from))
   {
      if(!msglen)
         msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
      send_message(from, sendbuf, msglen);
   }

   va_end(vl);
   return;
}
#ifdef FLUD
void sendto_channel_butlocal(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	int i;
	va_list vl;
	  
	va_start(vl, pattern);

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) {
      acptr = cm->cptr;
      if (acptr->from == one)
		  continue;		/* ...was the one I should skip */
      i = acptr->from->fd;
      if (!MyFludConnect(acptr)) {
			/*
			 * Now check whether a message has been sent to this remote
			 * link already
			 */
			if (sentalong[i] != sent_serial) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[i] = sent_serial;
			}
      }
   }
	va_end(vl);
   return;
}
#endif /* FLUD */

/*
 * sendto_channel_butserv
 * 
 * Send a message to all members of a channel that are connected to this
 * server.
 */
void sendto_channel_butserv(aChannel *chptr, aClient *from, char *pattern, ...)
{
   chanMember  *cm;
   aClient *acptr;
   va_list vl;
   int didlocal = 0;
   char *pfix;

   va_start(vl, pattern);

   pfix = va_arg(vl, char *);

   for (cm = chptr->members; cm; cm = cm->nextuser)
   {
      if (MyConnect(acptr = cm->cptr))
      {
         if(!didlocal)
            didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);

         if(check_fake_direction(from, acptr))
            continue;

         send_message(acptr, sendbuf, didlocal);

         /* vsendto_prefix_one(acptr, from, pattern, vl); */
      }
   }
   va_end(vl);
   return;
}

/*
 * sendto_ssjoin_servs
 * 
 * send to all servers with ssjoin capability (or not)
 * 
 */
void sendto_ssjoin_servs(int ssjoin, aChannel *chptr, aClient *from, char *pattern, ...)
{
	int j, k = 0;
	fdlist      send_fdlist;
	int     i;
	aClient *cptr;
	va_list vl;
	
   if (chptr) {
      if (*chptr->chname == '&')
		  return;
   }
	va_start(vl, pattern);
   for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
      if (!(cptr = local[i]) || 
           (cptr == from) ||
           (ssjoin && !IsSSJoin(cptr)) ||
           (!ssjoin && IsSSJoin(cptr)))
		  continue;

      send_fdlist.entry[++k] = i;
   }
   send_fdlist.last_entry = k;
   if (k)
	  vsendto_fdlist(&send_fdlist, pattern, vl);
	va_end(vl);
   return;
}


/*
 * * send a msg to all ppl on servers/hosts that match a specified mask *
 * (used for enhanced PRIVMSGs) *
 * 
 * addition -- Armin, 8jun90 (gruner@informatik.tu-muenchen.de)
 */

static int match_it(aClient *one, char *mask, int what)
{
   if (what == MATCH_HOST)
	  return (match(mask, one->user->host) == 0);
   else
	  return (match(mask, one->user->server) == 0);
}
/*
 * sendto_match_servs
 * 
 * send to all servers which match the mask at the end of a channel name
 * (if there is a mask present) or to all if no mask.
 */
void sendto_match_servs(aChannel *chptr, aClient *from, char *pattern, ...)
{
	int j, k = 0;
	fdlist      send_fdlist;
	int     i;
	aClient *cptr;
	va_list vl;
	
   if (chptr) {
      if (*chptr->chname == '&')
		  return;
   }
	va_start(vl, pattern);
   for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
      if (!(cptr = local[i]))
		  continue;
      if (cptr == from)
		  continue;
      send_fdlist.entry[++k] = i;
   }
   send_fdlist.last_entry = k;
   if (k)
	  vsendto_fdlist(&send_fdlist, pattern, vl);
	va_end(vl);
   return;
}

/*
 * sendto_match_butone
 * 
 * Send to all clients which match the mask in a way defined on 'what';
 * either by user hostname or user servername.
 */
void sendto_match_butone(aClient *one, aClient *from, char *mask, int what, 
								 char *pattern, ...)
{
	int     i;
	aClient *cptr, *acptr;
	char cansendlocal, cansendglobal;
	va_list vl;
	
	va_start(vl, pattern);
	if (MyConnect(from)) {
		cansendlocal = (OPCanLNotice(from)) ? 1 : 0;
		cansendglobal = (OPCanGNotice(from)) ? 1 : 0;
	} 
	else 
	  cansendlocal = cansendglobal = 1;
   for (i = 0; i <= highest_fd; i++) {
      if (!(cptr = local[i]))
		  continue;		/* that clients are not mine */
      if (cptr == one)		/* must skip the origin !! */
		  continue;
      if (IsServer(cptr)) {
			if (!cansendglobal) continue;
			for (acptr = client; acptr; acptr = acptr->next)
			  if (IsRegisteredUser(acptr)
					&& match_it(acptr, mask, what)
					&& acptr->from == cptr)
				 break;
			/*
			 * a person on that server matches the mask, so we * send *one*
			 * msg to that server ...
			 */
			if (acptr == NULL)
			  continue;
			/*
			 * ... but only if there *IS* a matching person 
			 */
      }
      /*
       * my client, does he match ? 
       */
      else if (!cansendlocal || !(IsRegisteredUser(cptr) &&
											 match_it(cptr, mask, what)))
		  continue;
      vsendto_prefix_one(cptr, from, pattern, vl);
   }
	va_end(vl);
   return;
}
/*
 * sendto_all_butone.
 * 
 * Send a message to all connections except 'one'. The basic wall type
 * message generator.
 */
void
sendto_all_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int     i;
	aClient *cptr;
	va_list vl;
	
	va_start(vl, pattern);
   for (i = 0; i <= highest_fd; i++)
	  if ((cptr = local[i]) && !IsMe(cptr) && one != cptr)
		 vsendto_prefix_one(cptr, from, pattern, vl);
	va_end(vl);
   return;
}
/*
 * sendto_ops_lev
 * 
 * Send to *local* ops only at a certain level... 0 = normal +s 1 = client
 * connect/disconnect   (+c) [IRCOPS ONLY] 2 = bot rejection
 * (+r) 3 = server kills                      (+k)
 */
void sendto_ops_lev(int lev, char *pattern, ...)
{
	aClient *cptr;
	int     i;
	char        nbuf[1024];
	va_list vl;
	
	va_start(vl,pattern);
	for (i = 0; i <= highest_fd; i++)
	  if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr)) {
		  switch (lev) {
			case CCONN_LEV:
			  if (!SendCConnNotice(cptr) || !IsAnOper(cptr))
				 continue;
			  break;
			case REJ_LEV:
			  if (!SendRejNotice(cptr) || !IsAnOper(cptr))
				 continue;
			  break;
			case SKILL_LEV:
			  if (!SendSkillNotice(cptr))
				 continue;
			  break;
			case SPY_LEV:
			  if (!SendSpyNotice(cptr) || !IsAnOper(cptr))
				 continue;
			  break;
			case DCCSEND_LEV:
			  if (!SendDCCNotice(cptr) || !IsAnOper(cptr))
				 continue;
			  break;
			case FLOOD_LEV:
			  if (!SendFloodNotice(cptr) || !IsAnOper(cptr))
				 continue;
			  break;
			case SPAM_LEV:
			  if (!SendSpamNotice(cptr) || !IsAnOper(cptr))
				 continue;
			  break;
			case DEBUG_LEV:
			  if (!SendDebugNotice(cptr) || !IsAnOper(cptr))
				 continue;
			  break;
			  
			default:		/* this is stupid, but oh well */
			  if (!SendServNotice(cptr))
				 continue;
		  }
		  (void) ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ",
								  me.name, cptr->name);
		  (void) strncat(nbuf, pattern,
							  sizeof(nbuf) - strlen(nbuf));
		  vsendto_one(cptr, nbuf, vl);
	  }
	va_end(vl);
	return;
}				
/*
 * sendto_ops
 * 
 * Send to *local* ops only.
 */
void sendto_ops(char *pattern, ...)
{
	aClient *cptr;
	int     i;
	char        nbuf[1024];
	va_list vl;
	
	va_start(vl, pattern);
   for (i = 0; i <= highest_fd; i++)
	  if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) &&
			IsAnOper(cptr) && SendServNotice(cptr)) {
		  (void) ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ",
								  me.name, cptr->name);
		  (void) strncat(nbuf, pattern,
							  sizeof(nbuf) - strlen(nbuf));
		  vsendto_one(cptr, nbuf, vl);
	  }
	va_end(vl);
   return;
}

/*
 * * sendto_ops_butone *      Send message to all operators. * one -
 * client not to send message to * from- client which message is from
 * *NEVER* NULL!!
 */
void sendto_ops_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int     i;
	aClient *cptr;
   va_list vl;
	   
	va_start(vl, pattern);

   INC_SERIAL();

   for (cptr = client; cptr; cptr = cptr->next) {
      if (!SendWallops(cptr))
		  continue;
      /*
       * we want wallops if (MyClient(cptr) && !(IsServer(from) ||
       * IsMe(from))) continue;
       */
      i = cptr->from->fd;	/*
									 * find connection oper is on 
									 */
      if (sentalong[i] == sent_serial)		/*
									 * sent message along it already ? 
									 */
		  continue;
      if (cptr->from == one)
		  continue;		/*
							 * ...was the one I should skip 
							 */
      sentalong[i] = sent_serial;
      vsendto_prefix_one(cptr->from, from, pattern, vl);
   }
	va_end(vl);
   return;
}
/*
 * sendto_wallops_butone 
 * Send message to all operators. 
 * one - client not to send message to 
 * from - client which message is from *NEVER* NULL!!
 */
void sendto_wallops(aClient *from, char *pattern, ...)
{
   int i;
   aClient *cptr;
   va_list vl;
   int vlen;
   char *pfix;

   va_start(vl, pattern);
   pfix = va_arg(vl, char *);
   vlen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
   va_end(vl);

   for(i=0; i <= highest_fd; i++) 
   {
      if((cptr = local[i]) != NULL) 
      {
	 if(!(IsRegistered(cptr) && SendWallops(cptr)))
	    continue;
         send_message(cptr, sendbuf, vlen);
      }
   }
   return;
}

void send_globops(char *pattern, ...)
{
	aClient    *cptr;
	int         i;
	char        nbuf[1024];
   va_list vl;
	
	va_start(vl, pattern);
   for (i = 0; i <= highest_fd; i++)
	  if ((cptr = local[i]) && !IsServer(cptr) && IsAnOper(cptr) &&
			!IsMe(cptr) && SendGlobops(cptr)) {
		  (void) ircsprintf(nbuf, ":%s NOTICE %s :*** Global -- ",
							  me.name, cptr->name);
		  (void) strncat(nbuf, pattern,
							  sizeof(nbuf) - strlen(nbuf));
		  vsendto_one(cptr, nbuf, vl);
	  }
	va_end(vl);
   return;
}
void send_chatops(char *pattern, ...)
{
	aClient    *cptr;
	int         i;
	char        nbuf[1024];
	va_list vl;
	
	va_start(vl, pattern);
   for (i = 0; i <= highest_fd; i++)
	  if ((cptr = local[i]) && !IsServer(cptr) && IsAnOper(cptr) &&
			!IsMe(cptr) && SendChatops(cptr)) {
		  (void) ircsprintf(nbuf, ":%s NOTICE %s :*** ChatOps -- ",
							  me.name, cptr->name);
		  (void) strncat(nbuf, pattern,
							  sizeof(nbuf) - strlen(nbuf));
		  vsendto_one(cptr, nbuf, vl);
	  }
	va_end(vl);
   return;
}

/*
 * to - destination client from - client which message is from
 * 
 * NOTE: NEITHER OF THESE SHOULD *EVER* BE NULL!! -avalon
 */
void sendto_prefix_one(aClient *to, aClient *from, char *pattern, ...)
{
   static char sender[HOSTLEN + NICKLEN + USERLEN + 5];
   static char temp[1024];
   anUser *user;
   char *idx;
   char *par;
   int flag = 0, sidx = 0;
   va_list vl, vl2;

   va_start(vl, pattern);
   vl2 = vl;

   par = va_arg(vl, char *);
   /*
    * Optimize by checking if (from && to) before everything 
    * uhh, there's _always_ going to be a to!
    */
   if (from) 
   {
      if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) 
      {
         if (IsServer(from)) 
         {
            ircvsprintf(temp, pattern, vl2);
	    sendto_ops("Send message (%s) to %s[%s] dropped from %s(Fake Dir)", temp,
                      to->name, to->from->name, from->name);
            va_end(vl);
            return;
         }

         sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->user->username, 
                    to->user->host, from->name, from->user->username, from->user->host, to->from->name);
         sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name, 
                            me.name, to->name, to->user->username, to->user->host, to->from->name);

         to->flags |= FLAGS_KILLED;
         exit_client(NULL, to, &me, "Ghosted client");
         if (IsPerson(from))
            sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name, 
                       to->user->username, to->user->host, to->from);
         va_end(vl);
         return;
      }

      if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) 
      {
         user = from->user;

         for(idx = from->name; *idx; idx++)
            sender[sidx++] = *idx;

         if (user)
         {
            if (*user->username) 
            {
               sender[sidx++] = '!';
               for(idx = user->username; *idx; idx++)
                  sender[sidx++] = *idx;
	    }
            if (*user->host && !MyConnect(from)) 
            {
               sender[sidx++] = '@';
               for(idx = user->host; *idx; idx++)
                  sender[sidx++] = *idx;
               flag = 1;
            }
         }

         /*
          * flag is used instead of index(sender, '@') for speed and
          * also since username/nick may have had a '@' in them.
          * -avalon
          */

         if (!flag && MyConnect(from) && *user->host) 
         {
            sender[sidx++] = '@';
            for(idx = from->sockhost; *idx; idx++)
               sender[sidx++] = *idx;
         }

         sender[sidx] = '\0';
         par = sender;

      }
   }

   temp[0] = ':';
   sidx = 1;

   /* okay, we more or less know that our sendto_prefix crap is going to be :%s <blah>,
    * so it's easy to fix these lame problems...joy */

    for(idx = par; *idx; idx++)
       temp[sidx++] = *idx;
    for(idx = (pattern + 3); *idx; idx++)
       temp[sidx++] = *idx;

    temp[sidx] = '\0'; 

    vsendto_one(to, temp, vl);
    va_end(vl);
}

/* this is an incredibly expensive function. 
 * removed all strcat() calls. - lucas */

void vsendto_prefix_one(aClient *to, aClient *from, char *pattern, va_list vl)
{
   static char sender[HOSTLEN + NICKLEN + USERLEN + 5];
   static char temp[1024];
   anUser *user;
   char *idx;
   char *par;
   int flag = 0, sidx = 0;
   va_list vl2 = vl;
	
   par = va_arg(vl2, char *);
   /*
    * Optimize by checking if (from && to) before everything 
    * uhh, there's _always_ going to be a to!
    */
   if (from) 
   {
      if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) 
      {
         if (IsServer(from)) 
         {
            ircvsprintf(temp, pattern, vl);
	    sendto_ops("Send message (%s) to %s[%s] dropped from %s(Fake Dir)", temp,
                      to->name, to->from->name, from->name);
            return;
         }

         sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->user->username, 
                    to->user->host, from->name, from->user->username, from->user->host, to->from->name);
         sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name, 
                            me.name, to->name, to->user->username, to->user->host, to->from->name);

         to->flags |= FLAGS_KILLED;
         exit_client(NULL, to, &me, "Ghosted client");
         if (IsPerson(from))
            sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name, 
                       to->user->username, to->user->host, to->from);
         return;
      }

      if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) 
      {
         user = from->user;

         for(idx = from->name; *idx; idx++)
            sender[sidx++] = *idx;

         if (user)
         {
            if (*user->username) 
            {
               sender[sidx++] = '!';
               for(idx = user->username; *idx; idx++)
                  sender[sidx++] = *idx;
	    }
            if (*user->host && !MyConnect(from)) 
            {
               sender[sidx++] = '@';
               for(idx = user->host; *idx; idx++)
                  sender[sidx++] = *idx;
               flag = 1;
            }
         }

         /*
          * flag is used instead of index(sender, '@') for speed and
          * also since username/nick may have had a '@' in them.
          * -avalon
          */

         if (!flag && MyConnect(from) && *user->host) 
         {
            sender[sidx++] = '@';
            for(idx = from->sockhost; *idx; idx++)
               sender[sidx++] = *idx;
         }

         sender[sidx] = '\0';
         par = sender;

      }
   }

   temp[0] = ':';
   sidx = 1;

   /* okay, we more or less know that our sendto_prefix crap is going to be :%s <blah>,
    * so it's easy to fix these lame problems...joy */

    for(idx = par; *idx; idx++)
       temp[sidx++] = *idx;
    for(idx = (pattern + 3); *idx; idx++)
       temp[sidx++] = *idx;

    temp[sidx] = '\0'; 

    vsendto_one(to, temp, vl2);
}

void sendto_fdlist(fdlist *listp, char *pattern, ...)
{
	int len, j, fd;
	va_list vl;

	va_start(vl, pattern);
   len = ircvsprintf(sendbuf, pattern, vl);
	
   for (fd = listp->entry[j = 1]; j <= listp->last_entry; fd = listp->entry[++j])
	  send_message(local[fd], sendbuf, len);
	va_end(vl);
}

static void sendto_fdlist_preformatted(fdlist *listp, char *buf, int len)
{
   int j, fd;
	
   for (fd = listp->entry[j = 1]; j <= listp->last_entry; fd = listp->entry[++j])
      send_message(local[fd], buf, len);
}

void vsendto_fdlist(fdlist *listp, char *pattern, va_list vl)
{
	int len, j, fd;
   len = ircvsprintf(sendbuf, pattern, vl);
	
   for (fd = listp->entry[j = 1]; j <= listp->last_entry; fd = listp->entry[++j])
	  send_message(local[fd], sendbuf, len);
}


/*
 * sendto_realops
 * 
 * Send to *local* ops only but NOT +s nonopers.
 * If it's to local ops only and not +s nonopers, then SendServNotice is
 * wrong. Changed to IsAnOper. -mjs
 */
void sendto_realops(char *pattern, ...)
{
	aClient *cptr;
	int     i;
	char        nbuf[1024];
	fdlist     *l;
	int         fd;
	va_list vl;
	  
	va_start(vl, pattern);
   l = &oper_fdlist;
   for (fd = l->entry[i = 1]; i <= l->last_entry; fd = l->entry[++i]) {
      if (!(cptr = local[fd]))
		  continue;
      if (IsAnOper(cptr)) {
			(void) ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s",
									me.name, cptr->name, pattern);
			vsendto_one(cptr, nbuf, vl);
      }
   }
	va_end(vl);
   return;
}

void vsendto_realops(char *pattern, va_list vl)
{
	aClient *cptr;
	int     i;
	char        nbuf[1024];
	fdlist     *l;
	int         fd;

   l = &oper_fdlist;
   for (fd = l->entry[i = 1]; i <= l->last_entry; fd = l->entry[++i]) {
      if (!(cptr = local[fd]))
		  continue;
      if (IsAnOper(cptr)) {
			(void) ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s",
									me.name, cptr->name, pattern);
			vsendto_one(cptr, nbuf, vl);
      }
   }
   return;
}

/*
 * sendto_realops_lev
 * 
 * Send to *local* ops only but NOT +s nonopers at a certain level
 */
void sendto_realops_lev(int lev, char *pattern, ...)
{
	aClient *cptr;
	int     i;
	char        nbuf[1024];
	fdlist     *l;
	int         fd;
	va_list vl;
	
   l = &oper_fdlist;
	va_start(vl, pattern);
   for (fd = l->entry[i = 1]; i <= l->last_entry; fd = l->entry[++i]) {
      if (!(cptr = local[fd]))
		  continue;
      switch (lev) {
		 case CCONN_LEV:
			if (!SendCConnNotice(cptr))
			  continue;
			break;
		 case REJ_LEV:
			if (!SendRejNotice(cptr))
			  continue;
			break;
		 case SKILL_LEV:	/*
								 * This should not be sent, since this
								 * * can go to normal people 
								 */
			if (!SendSkillNotice(cptr))
			  continue;
			break;
		 case SPY_LEV:
			if (!SendSpyNotice(cptr))
			  continue;
			break;
                 case DCCSEND_LEV:
			if (!SendDCCNotice(cptr) || !IsAnOper(cptr))
			  continue;
			break;
		 case FLOOD_LEV:
			if (!SendFloodNotice(cptr))
			  continue;
			break;
		 case SPAM_LEV:
			if (!SendSpamNotice(cptr))
			  continue;
			break;
		 case DEBUG_LEV:
			if (!SendDebugNotice(cptr))
			  continue;
			break;
      }
      (void) ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ",
								me.name, cptr->name);
      (void) strncat(nbuf, pattern,
							sizeof(nbuf) - strlen(nbuf));
      vsendto_one(cptr, nbuf, vl);
   }
	va_end(vl);
   return;
}

/*
 * * ts_warn *      Call sendto_ops, with some flood checking (at most
 * 5 warnings *      every 5 seconds)
 */

void ts_warn(char * pattern, ...)
{
	static ts_val last = 0;
	static int  warnings = 0;
	ts_val now;
	va_list vl;
	
	va_start(vl, pattern);
   /*
    * * if we're running with TS_WARNINGS enabled and someone does *
    * something silly like (remotely) connecting a nonTS server, *
    * we'll get a ton of warnings, so we make sure we don't send * more
    * than 5 every 5 seconds.  -orabidoo
    */
   /*
    * th+hybrid servers always do TS_WARNINGS -Dianora
    */
   now = time(NULL);
   if (now - last < 5) {
      if (++warnings > 5)
	 return;
   }
   else {
      last = now;
      warnings = 0;
   }

   vsendto_realops(pattern, vl);
	va_end(vl);
   return;
}
/*
 * sendto_locops
 */
void sendto_locops(char *pattern, ...)
{
	aClient *cptr;
	int     i;
	char        nbuf[1024];
	fdlist     *l;
	int         fd;
	va_list vl;
	
	va_start(vl, pattern);
   l = &oper_fdlist;
   for (fd = l->entry[i = 1]; i <= l->last_entry; fd = l->entry[++i]) {
      if (!(cptr = local[fd]))
		  continue;
      if (SendGlobops(cptr)) {
			(void) ircsprintf(nbuf, ":%s NOTICE %s :*** LocOps -- ",
									me.name, cptr->name);
			(void) strncat(nbuf, pattern,
								sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
      }
   }
   va_end(vl);
   return;
}
/*
 * sendto_gnotice - send a routing notice to all local +n users.
 */
void sendto_gnotice(char *pattern, ...)
{
	aClient *cptr;
	int     i;
	char        nbuf[1024];
	va_list vl;
	
	va_start(vl, pattern);

   for (i = 0; i <= highest_fd; i++) {
      if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && SendRnotice(cptr)) {
 			
			(void) ircsprintf(nbuf, ":%s NOTICE %s :*** Routing -- ",
									me.name, cptr->name);
			(void) strncat(nbuf, pattern,
								sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
      }
   }
   va_end(vl);
   return;
}

/*
 * sendto_channelops_butone
 *   Send a message to all OPs in channel chptr that
 *   are directly on this server and sends the message
 *   on to the next server if it has any OPs.
 */
void sendto_channelops_butone(aClient *one, aClient *from, aChannel *chptr, 
										char *pattern, ...)
{
	chanMember   *cm;
	aClient *acptr;
	int     i;
	va_list vl;
	
	va_start(vl, pattern);

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) {
      acptr = cm->cptr;
      if (acptr->from == one ||
	  !(cm->flags & CHFL_CHANOP))
	 continue;
      i = acptr->from->fd;
      if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
			vsendto_prefix_one(acptr, from, pattern, vl);
			sentalong[i] = sent_serial;
      }
      else {
			/*
			 * Now check whether a message has been sent to this
			 * *      * remote link already 
			 */
			if (sentalong[i] != sent_serial) {
				vsendto_prefix_one(acptr, from, pattern, vl);

				sentalong[i] = sent_serial;
			}
      }
   }
	va_end(vl);
   return;
}
/*
 * sendto_channelvoice_butone
 *   Send a message to all voiced users in channel chptr that
 *   are directly on this server and sends the message
 *   on to the next server if it has any voiced users.
 */
void sendto_channelvoice_butone(aClient *one, aClient *from, aChannel *chptr, 
										  char *pattern, ...)
{
	chanMember   *cm;
	aClient *acptr;
	int     i;
	va_list vl;
	
	va_start(vl, pattern);

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) {
      acptr = cm->cptr;
      if (acptr->from == one ||
			 !(cm->flags & CHFL_VOICE))
		  continue;
      i = acptr->from->fd;
      if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
			vsendto_prefix_one(acptr, from, pattern, vl);
			sentalong[i] = sent_serial;
      }
      else {
			/*
			 * Now check whether a message has been sent to this
			 * *      * remote link already 
			 */
			if (sentalong[i] != sent_serial) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[i] = sent_serial;
			}
      }
   }
	va_end(vl);
   return;
}
/*
 * sendto_channelvoiceops_butone
 *   Send a message to all OPs or voiced users in channel chptr that
 *   are directly on this server and sends the message
 *   on to the next server if it has any OPs or voiced users.
 */
void sendto_channelvoiceops_butone(aClient *one, aClient *from, aChannel 
				   *chptr, char *pattern, ...)
{
   chanMember   *cm;
   aClient *acptr;
   int     i;
   va_list vl;
	
   va_start(vl, pattern);

   INC_SERIAL();

   for (cm = chptr->members; cm; cm = cm->nextuser) {
      acptr = cm->cptr;
      if (acptr->from == one || !((cm->flags & CHFL_VOICE) || (cm->flags & CHFL_CHANOP)))
         continue;
      i = acptr->from->fd;
      if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
         vsendto_prefix_one(acptr, from, pattern, vl);
         sentalong[i] = sent_serial;
      }
      else /* remote link */
      {
         if (sentalong[i] != sent_serial) 
         {
            vsendto_prefix_one(acptr, from, pattern, vl);
            sentalong[i] = sent_serial;
         }
      }
   }
   return;
}

/*******************************************
 * Flushing functions (empty queues)
 *******************************************/

/*
 * flush_connections
 * Empty only buffers for clients without FLAGS_BLOCKED
 * dump_connections 
 * Unintelligently try to empty all buffers.
 */
void flush_connections(int fd) 
{
   int     i;
   aClient *cptr;
   
   if (fd == me.fd) 
   {
      for (i = highest_fd; i >= 0; i--)
	if ((cptr = local[i]) && !(cptr->flags & FLAGS_BLOCKED) && DBufLength(&cptr->sendQ) > 0)
	  send_queued(cptr);
   }
   else if (fd >= 0 && (cptr = local[fd]) && !(cptr->flags & FLAGS_BLOCKED) && DBufLength(&cptr->sendQ) > 0)
     send_queued(cptr);
}

void dump_connections(int fd) 
{
   int     i;
   aClient *cptr;
   
   if (fd == me.fd) 
   {
      for (i = highest_fd; i >= 0; i--)
	if ((cptr = local[i]) && DBufLength(&cptr->sendQ) > 0)
	  send_queued(cptr);
   }
   else if (fd >= 0 && (cptr = local[fd]) && DBufLength(&cptr->sendQ) > 0)
     send_queued(cptr);
}

/* flush an fdlist intelligently */
void flush_fdlist_connections(fdlist *listp)
{
   int i, fd;
   aClient *cptr;
	
   for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i])
      if ((cptr = local[fd]) && !(cptr->flags & FLAGS_BLOCKED) && DBufLength(&cptr->sendQ) > 0)
         send_queued(cptr);
}


